/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.data;

import java.io.Serializable;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;

import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.sql.SQL;

/**
 * テーブル定義インターフェース
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
public interface ITableDef extends Serializable {
  
  /**
   * ITableDef実装クラスのテーブル定義情報を作成します。
   * @return
   */
  public MetaData<?> createMetaData();

  /**
   * データベーステーブルの定義情報
   * @author littlesoft
   *
   */
  @Retention(RetentionPolicy.RUNTIME) //有効範囲
  public @interface TableDef {
    String owner()   default "";
    String schema()   default "";
    String name()     default "";
  }
  
  /**
   * データベースカラムの定義情報
   * @author littlesoft
   *
   */
  @Retention(RetentionPolicy.RUNTIME) //有効範囲
  public @interface ColumnDef {
    String name()           default "";
    String oldName()        default "";
    String caption()        default "";
    String formatPattern()  default "";
    String defaultValue()   default "";
    int precision()         default -1;
    int scale()             default -1;
    int length()            default -1;
    boolean lenientConvert()  default true;
    Column.Nullable nullable() default Column.Nullable.DEFAULT;
    Column.InputType defaultInputEditor() default Column.InputType.NONE;
    Column.Constraint stringConstraint() default Column.Constraint.DEFAULT;
    String defaultInputStyle() default "";
    Class<? extends Code> optionType() default Code.class;
  }
  
  static public class ColData<T> implements java.lang.Comparable<T>{
    public int classOrdinal;
    public int menberOrdinal;
    public Column column;
    @SuppressWarnings("unchecked")
    @Override
    public int compareTo(T object) {
      ColData<T> o = (ColData<T>)object;
      if (this.classOrdinal == o.classOrdinal) {
        if (this.menberOrdinal == o.menberOrdinal) {
          return 0;
        } else if (this.menberOrdinal > o.menberOrdinal) {
          return 1;
        } else {
          return -1;
        }
      } else if (this.classOrdinal > o.classOrdinal) {
        return 1;
      } else {
        return -1;
      }
    }
  }
  

  /**
   * エンティティクラスをもとにカラムオブジェクトを生成します。
   * @param entityType
   * @param upperCaseDBColumnName
   * @return List<Column>
   */
  @SuppressWarnings("serial")
  static public class Columns extends ArrayList<Column> {
    
    private HashMap<String, Column> nameMap = new HashMap<String, Column>();
    
    public Columns(Table<?> table) {
      for (int i=0; i<table.getColumnCount(); i++) {
        Column col = table.getColumn(i);
        if ("".equals(col.getDBColumnName())) 
          col.setDBColumnName(col.getName());
        add(col);
      }
    }
    
    public Columns(Class<?> entityType) {
      this(entityType, false);
    }
    
    public Columns(Class<?> entityType, boolean upperCaseDBColumnName) {
      if (entityType.equals(Row.class)) return;
      ArrayList<ColData<Object>> cols = new ArrayList<ColData<Object>>();
      HashMap<Class<?>, Integer> classOrdinalMap = new HashMap<Class<?>, Integer>();
      Class<?> clazz = entityType;
      int i = Short.MAX_VALUE;
      while (clazz!=null) {
        classOrdinalMap.put(clazz, i--);
        clazz = clazz.getSuperclass();
      }
      try {
        i = 0;
        for (Field field : entityType.getFields()) {
          if (!Modifier.isPublic(field.getModifiers()) 
              || Modifier.isStatic(field.getModifiers())) continue;
          ColData<Object> coldata = new ColData<Object>();
          coldata.classOrdinal = classOrdinalMap.get(field.getDeclaringClass());
          coldata.menberOrdinal = i++;
          coldata.column = new Column(field, upperCaseDBColumnName);
          cols.add(coldata);
        }
        Method getter = null;
        Method setter = null;
        for (Method method : entityType.getMethods()) {
          getter = null;
          setter = null;
          //getterを抽出
          if (!Modifier.isPublic(method.getModifiers())) continue;
          if (method.getReturnType().equals(void.class)) continue;
          //if (method.getParameterTypes().length>0) continue;
          String name = method.getName();
          if (!name.startsWith("get") && !name.startsWith("is"))
            continue;
          getter = method;
          //setterを探す
          String setterName = "set";
          setterName += (name.startsWith("is")) ? 
            name.substring("is".length()) : name.substring("get".length());
          try {
            setter = entityType.getMethod(setterName, getter.getReturnType());
          } catch (NoSuchMethodException x) {
            //readOnly
          }
          if (Column.classToColtypeMap(getter.getReturnType())==null) {
            //getterの戻り値タイプがマップできない。
            continue;
          }
       //   if (settersetter.getParameterTypes().length!=1) continue;
          ColData<Object> coldata = new ColData<Object>();
          coldata.classOrdinal = classOrdinalMap.get(method.getDeclaringClass());
          coldata.menberOrdinal = i++;
          coldata.column = new Column(getter, setter, upperCaseDBColumnName);
          cols.add(coldata);
        }
        Collections.sort(cols);
        for (ColData<Object> coldata : cols) {
          this.add(coldata.column);          
        }
      } catch (Exception x) {
        throw new RuntimeException(x);
      }
    }

    private static int TYPE_POS = 24;
    private static int TYPE_LEN = 14;
    private static char[] SPACES = "                                ".toCharArray();
    @Override
    public String toString() {
      return toString(true);
    }

    public String toString(boolean includeReadOnly) {
      StringBuffer buff = new StringBuffer();

      for (Column col : this) {
        if (!includeReadOnly && col.isReadOnly()) continue;
        buff.append((col == get(0)) ? " " : ",");
        appendColumnDefString(col, buff);
        buff.append("\n");
      }
      return buff.toString();
    }
    
    public void appendColumnDefString(Column col, StringBuffer buff) {
      try {    
      String name;
      String type;
      int spaces;
      name = col.getDBColumnName();
      buff.append(name);
      //spaces = ((name.length()+2)>TYPE_POS) ? 1 : (TYPE_POS-(name.length()+1));
      spaces = 1;
      buff.append(new String(SPACES,0,spaces));
      type = Column.mapToSQLString(col.getType());
      //buff.append(type);
      switch (col.getType()) {
      case BIGDECIMAL:
        if (col.getPrecision()>0) {
          type += "(" + col.getPrecision();
          if (col.getScale()>-1) {
            type += "," + col.getScale();
          }
          type += ")";
        }
        break;
      case STRING:
        if (col.getLength()>0) {
          type +=  "(" + col.getLength()+ ")";
        }
        break;
      }
      buff.append(type);
      //spaces = (type.length()>TYPE_LEN) ? 1 : (TYPE_LEN-type.length());
      spaces = 1;
      buff.append(new String(SPACES,0,spaces));
      if (col.getNullable()!=null) {
        buff.append( (col.isNullable()) ? " NULL" : " NOT NULL");
      }
      toDefaultString(col, buff);
/*      if (!"".equals(col.getDefaultValue())) {
        switch (col.getType()) {
        case STRING:
        case CLOB:
        case DATE:            
        case TIME:            
        case TIMESTAMP:
          buff.append(" DEFAULT '");
          buff.append(col.getDefaultValue());
          buff.append("'");
          break;
        case BOOLEAN:
        case BYTE:
        case SHORT:
        case INT:
        case LONG:
        case FLOAT:
        case DOUBLE:
        case BIGDECIMAL:
          buff.append(" DEFAULT ");
          buff.append(col.getDefaultValue());
          buff.append("");
          break;
        default:
          break;
        }  
      }*/

      } catch (DataException e) {
        throw new RuntimeException(e);
      }
    }

    static public void toDefaultString(Column col, StringBuffer buff) {
      if (!"".equals(col.getDefaultValue())) {
        switch (col.getType()) {
        case STRING:
        case CLOB:
        case DATE:            
        case TIME:            
        case TIMESTAMP:
          buff.append(" DEFAULT '");
          buff.append(col.getDefaultValue());
          buff.append("'");
          break;
        case BOOLEAN:
        case BYTE:
        case SHORT:
        case INT:
        case LONG:
        case FLOAT:
        case DOUBLE:
        case BIGDECIMAL:
          buff.append(" DEFAULT ");
          buff.append(col.getDefaultValue());
          buff.append("");
          break;
        default:
          break;
        }  
      }      
    }
    
    public boolean add(Column column) {
      nameMap.put(column.getDBColumnName(), column);
      return super.add(column);
    }
    
    public Column getByDBColumnName(String dbColumnName) {
      return nameMap.get(dbColumnName);
    }
    
    
    static public Column[] listToArray(ArrayList<Column> colList) {
      Column[] colArray = new Column[colList.size()];
      int i = 0;
      for (Column col : colList) {
        colArray[i++] = col;
      }
      return colArray;
    }
  }
  
  
  @SuppressWarnings("serial")
  static public class Constraint implements Serializable {
    MetaData<?> meta;
    Column[] columns;
    public Constraint(MetaData<?> meta, Column... columns) {
      this.meta = meta;
      this.columns = columns;
    }
    
    public String getColumnNames() {
      StringBuffer buff = new StringBuffer();
      for (Column col : columns) {
        buff.append( (col==columns[0]) ?  "" : "," );
        buff.append(col.getDBColumnName());
      }
      return buff.toString();
    }
    
    public boolean contains(Column col) {
      for (Column c : columns) {
        if (c==col) return true;
      }
      return false;
    }
    
    public boolean equalsColArray(Constraint constraint) {
      if (this.columns.length!=constraint.columns.length) {
        return false;
      } else {  
        int i = 0;
        for (Column col : constraint.columns){
          if (!this.columns[i++].getDBColumnName().equals(col.getDBColumnName())) {
            return false;
          }
        }
      }
      return true;
    }
  }
  
  @SuppressWarnings("serial")
  static public class PrimaryKey extends Constraint {

    public PrimaryKey(MetaData<?> meta, Column... columns) {
      super(meta, columns);
      //PrimaryKeyのカラムはNOT NULL
      for (Column col : columns) {
        col.setNullable(false);
      }
    }
    
    public String getName() {
      return meta.getDBTableName(true) + "_PK";
    }
    
    public String toDropString() {
      StringBuffer buff = new StringBuffer();
      buff.append("ALTER TABLE ");
      buff.append(meta.getDBTableName(true));
      buff.append(" DROP CONSTRAINT ");
      buff.append(getName());
      return buff.toString();
    }  
    
    @Override
    public String toString() {
      StringBuffer buff = new StringBuffer();
      buff.append("ALTER TABLE ");
      buff.append(meta.getDBTableName(true));
      buff.append(" ADD CONSTRAINT ");
      buff.append(getName());
      buff.append(" PRIMARY KEY (");
      for (Column col : columns) {
        buff.append( (col==columns[0]) ?  "" : "," );
        buff.append(col.getDBColumnName());
      }
      buff.append(")");
      return buff.toString();
    }
  }

  @SuppressWarnings("serial")
  static public class ForeginKey extends Constraint {
    
    Class<? extends ITableDef> referenceTable = null;
    String referenceTableName = "";
    protected String referenceTableColumnNames = "";
    boolean usingIndex = false;
    Index index = null;
    MetaData.FKOption onDelete = null;
    MetaData.FKOption onUpdate = null;
    
    public ForeginKey(MetaData<?> meta, Column... columns) {
      super(meta, columns);
    }
    
    public ForeginKey references(Class<? extends ITableDef> referenceTable) {
      this.referenceTable = referenceTable;
      usingIndex(true);
      return this;
    }
    
    public ForeginKey references(String referenceTableName, boolean usingIndex) {
      this.referenceTableName = referenceTableName;
      usingIndex(usingIndex);
      return this;
    }
    
    public ForeginKey referenceTableColumnNames(ArrayList<String> referenceTableColumnNames) {
      this.referenceTableColumnNames = "";
      for (String name : referenceTableColumnNames) {
        if (!"".equals(this.referenceTableColumnNames)) {
          this.referenceTableColumnNames += ",";
        }
        this.referenceTableColumnNames += name;
      }
      return this;
    }
    
    public ForeginKey usingIndex(boolean usingIndex) {
      if (this.usingIndex == usingIndex) return this;
      if (index != null) {
        meta.indexes.remove(index);
        index = null;
      }
      this.usingIndex = usingIndex;
      if (this.usingIndex) {
        index = new Index(meta, columns);
        meta.getIndexes().add(index);
      }
      return this;
    }
    
    public ForeginKey onDelete(MetaData.FKOption onDelete) {
      this.onDelete = onDelete;
      return this;
    }

    public ForeginKey onDelete(int option) {
      switch(option) {
      case java.sql.DatabaseMetaData.importedKeyCascade:
        this.onDelete = MetaData.FKOption.CASCADE;
        break;
      case java.sql.DatabaseMetaData.importedKeyNoAction:
        this.onDelete = MetaData.FKOption.NO_ACTION;
        break;
      case java.sql.DatabaseMetaData.importedKeyRestrict:
        this.onDelete = MetaData.FKOption.RESTRICT;
        break;
      case java.sql.DatabaseMetaData.importedKeySetDefault:
        this.onDelete = MetaData.FKOption.SET_DEFAULT;
        break;
      case java.sql.DatabaseMetaData.importedKeySetNull:
        this.onDelete = MetaData.FKOption.SET_NULL;
      }
      return this;
    }
    
    public ForeginKey onUpdate(MetaData.FKOption onUpdate) {
      this.onUpdate = onUpdate;
      return this;
    }

    public ForeginKey onUpdate(int option) {
      switch(option) {
      case java.sql.DatabaseMetaData.importedKeyCascade:
        this.onUpdate = MetaData.FKOption.CASCADE;
        break;
      case java.sql.DatabaseMetaData.importedKeyNoAction:
        this.onUpdate = MetaData.FKOption.NO_ACTION;
        break;
      case java.sql.DatabaseMetaData.importedKeyRestrict:
        this.onUpdate = MetaData.FKOption.RESTRICT;
        break;
      case java.sql.DatabaseMetaData.importedKeySetDefault:
        this.onUpdate = MetaData.FKOption.SET_DEFAULT;
        break;
      case java.sql.DatabaseMetaData.importedKeySetNull:
        this.onUpdate = MetaData.FKOption.SET_NULL;
      }
      return this;
    }

    public String getName() {
      StringBuffer buff = new StringBuffer();
      buff.append(getColumnNames());
      buff.append(" ");
      buff.append((referenceTable==null) ? referenceTableName : referenceTable.getSimpleName());
      return meta.getDBTableName(true) + "_FK" + Integer.toHexString(buff.toString().hashCode());
    }

    public String toDropString() {
      StringBuffer buff = new StringBuffer();
      buff.append("ALTER TABLE ");
      buff.append(meta.getDBTableName(true));
      buff.append(" DROP CONSTRAINT ");
      buff.append(getName());
      return buff.toString();
    }    
    @Override
    public String toString() {
      StringBuffer buff = new StringBuffer();
      
      buff.append("ALTER TABLE ");
      buff.append(meta.getDBTableName(true));
      buff.append(" ADD CONSTRAINT ");
      buff.append(getName());
      buff.append(" FOREIGN KEY (");
      buff.append(getColumnNames());
      buff.append(") REFERENCES ");
      if (referenceTable!=null) {
        MetaData<?> refMeta;
        try {
          refMeta = ((ITableDef)referenceTable.newInstance()).createMetaData();
        } catch (Exception e) {
          throw new RuntimeException(e.getLocalizedMessage() + "[" + referenceTable.getSimpleName() + "]");
        }
        referenceTableName = refMeta.getDBTableName(true);
        if (refMeta.getPrimaryKey()==null) {
          throw new RuntimeException("Primary key is not specified in the table [" + refMeta.getDBTableName(true) + "]");
        }
        referenceTableColumnNames = refMeta.getPrimaryKey().getColumnNames();
      }
      buff.append(referenceTableName);
      buff.append(" (");
      buff.append(referenceTableColumnNames);
      buff.append(")");
      if (onDelete!=null 
          && !onDelete.equals(MetaData.FKOption.NO_ACTION) 
          && !onDelete.equals(MetaData.FKOption.RESTRICT)) {
        buff.append(" ON DELETE ");
        buff.append(onDelete.toString().replace('_', ' '));
      }
      if (onUpdate!=null
          && !onUpdate.equals(MetaData.FKOption.NO_ACTION) 
          && !onUpdate.equals(MetaData.FKOption.RESTRICT)) {
        buff.append(" ON UPDATE ");
        buff.append(onUpdate.toString().replace('_', ' '));
      }
      return buff.toString();
    }
  }
  
  @SuppressWarnings("serial")
  static public class Index extends Constraint {
    boolean unique;
    
    public Index(MetaData<?> meta, Column... columns) {
      super(meta, columns);
    }
    
    public Index unique(boolean unique) {
      this.unique = unique;
      return this;
    }
    
    public String getName() {
      StringBuffer buff = new StringBuffer();
      for (Column col : columns) {
        buff.append( (col==columns[0]) ?  "" : "," );
        buff.append(col.getDBColumnName());
      }
      return meta.getDBTableName(true) + "_IX" + Integer.toHexString(buff.toString().hashCode());
    }
    
    public String toDropString() {
      StringBuffer buff = new StringBuffer();
      buff.append("DROP INDEX ");
      buff.append(getName());
      return buff.toString();
    }
    
    @Override
    public String toString() {
      StringBuffer buff = new StringBuffer();
      buff.append("CREATE ");
      if (unique) buff.append("UNIQUE ");
      buff.append("INDEX ");
      buff.append(getName());
      buff.append(" ON ");
      buff.append(meta.getDBTableName(true));
      buff.append("(");
      for (Column col : columns) {
        buff.append( (col==columns[0]) ?  "" : "," );
        buff.append(col.getDBColumnName());
      }
      buff.append(")");
      return buff.toString();
    }
  }

  @SuppressWarnings("serial")
  static public class MetaData<T extends Serializable> implements Serializable {
    public enum FKOption {CASCADE,RESTRICT,NO_ACTION,SET_DEFAULT,SET_NULL};
    protected boolean upperCaseDBColumnName = true;
    protected Class<?> entityType;
    public T $$;
    protected String owner = null;
    protected String dbCatalog = null;
    protected String dbSchema;
    protected String dbTableName;
    private String alias = null;
    private boolean upperCase = true;
    protected Columns columns;
    private PrimaryKey primaryKey;
    private ArrayList<ForeginKey> foreginKeys;
    private ArrayList<Index> indexes;
    
    public MetaData(T...t) {
      initialize((Class<?>) t.getClass().getComponentType());
      fillDummyValue();
    }
    
    public MetaData(Object entity) {
      if (entity instanceof Class<?>) {
        initialize((Class<?>)entity);
      } else {
        initialize(entity.getClass());
      }
    }
    
    public MetaData(Class<? extends T> entityType) {
      initialize(entityType);
    }
    
    protected void initialize(Class<?> type) {
      entityType = type;
      try {
        columns = new Columns(entityType, upperCaseDBColumnName);
        dbSchema = null;
        dbTableName = null;
        if (entityType.isAnnotationPresent(TableDef.class)) {
          TableDef p = (TableDef) entityType.getAnnotation(TableDef.class);
          owner = "".equals(p.owner()) ? null : p.owner();
          dbSchema = "".equals(p.schema()) ? null : p.schema();
          if (dbSchema==null && entityType.getDeclaringClass()!=null) {
            dbSchema = entityType.getDeclaringClass().getSimpleName();
          }
          dbTableName = "".equals(p.name()) ? null : p.name();
        }
        if (dbTableName == null) dbTableName = SQL.camelToUnderscore(this.entityType.getSimpleName(), upperCase);
      } catch (Exception x) {
        throw new RuntimeException(x);
      }
    }
    
    public void fillDummyValue() {
      fillDummyValue(null);
    }
    
    @SuppressWarnings("unchecked")
    public void fillDummyValue(T entity) {
      if (entityType==null) return; 
      try {
        if (entity!=null) {
          $$ = entity;
        } else if ($$ == null){
          $$ = (T) entityType.newInstance(); 
        }
        Serializable dummy = null;
        for (Column col : columns) {
          if (!col.isReadOnly() && col.getEntityValue($$) == null) {
            dummy = Column.createDefaultValue(col, 0);
            col.setEntityValue($$, dummy);
          }
        }
      } catch (Exception x) {
        throw new RuntimeException(x);
      }
    }
    

    
    public Column $column(Serializable $dummyFiled) {
      fillDummyValue();
      try {
        for (Column c : columns) {
          if (c.getEntityValue($$)==$dummyFiled) {
            return c;
          }
        }
      } catch (Exception x) {
        throw new RuntimeException(x);
      }
      return null;
    }

    public Column[] $columns(Serializable... $dummyFileds) {
      Column[] result = new Column[$dummyFileds.length];
      int i = 0;
      for (Serializable f : $dummyFileds) {
        result[i++] = $column(f);
      }
      return result;
    }
    
    public String getDBCatalog() {
      return dbCatalog;
    }

    public void setDBCatalog(String dbCatalog) {
      this.dbCatalog = dbCatalog;
    }
    
    public String getDBSchema() {
      return dbSchema;
    }

    public void setDBSchema(String dbSchema) {
      this.dbSchema = dbSchema;
    }
    
    public String getDBTableName(boolean withSchema) {
      return (withSchema && dbSchema!=null) ? (dbSchema + "." + dbTableName) : dbTableName;
    }

    public void setDBTableName(String dbTableName) {
      this.dbTableName = dbTableName;
    }
    
    public Columns getColumns() {
      return columns;
    }
    
    public PrimaryKey getPrimaryKey() {
      return primaryKey;
    }

    public ArrayList<ForeginKey> getForeginKeys() {
      if (foreginKeys==null) {
        foreginKeys = new ArrayList<ForeginKey>();
      }
      return foreginKeys;
    }

    public ArrayList<Index> getIndexes() {
      if (indexes==null) {
        indexes = new ArrayList<Index>();
      }
      return indexes;
    }
    
    public String getAlias() {
      return alias;
    }

    public void setAlias(String alias) {
      this.alias = alias;
    }
    
    public boolean hasAlias() {
      return (alias != null);
    }
    
    public String getTableOrAliasName() {
      return (alias == null) ? getDBTableName(true) : alias;
    }
    
    public String getTableAsAlias() {
      return (alias == null) ? getDBTableName(true) : getDBTableName(true) + " AS " + alias;
    }
    
    public PrimaryKey addPrimaryKey(Serializable...dummyFields) {
      primaryKey = new PrimaryKey(this, $columns(dummyFields));
      return primaryKey;
    }
    
    public PrimaryKey addPrimaryKey(ArrayList<Column> colList) {
      primaryKey = new PrimaryKey(this, Columns.listToArray(colList));
      return primaryKey;
    }
    
    public ForeginKey addForeginKey(Serializable...dummyFields) {
      getForeginKeys().add(new ForeginKey(this, $columns(dummyFields)));
      return getForeginKeys().get(foreginKeys.size()-1);
    }
    
    public ForeginKey addForeginKey(ArrayList<Column> colList) {
      getForeginKeys().add(new ForeginKey(this, Columns.listToArray(colList)));
      return getForeginKeys().get(foreginKeys.size()-1);
    }
    
    public Index addIndex(Serializable...dummyFields) {
      Index index = new Index(this, $columns(dummyFields));
      //主キー、外部キーにより暗黙で作成されるインデックスと重複する定義はADDせずスキップする
      if (primaryKey!=null && primaryKey.equalsColArray(index)) return index;
      //for (ForeginKey fk : getForeginKeys()) {
      //  if (fk.equalsColArray(index)) return index;
      //}
      getIndexes().add(index);
      return getIndexes().get(indexes.size()-1);
    }
    
    public Index addIndex(ArrayList<Column> colList) {
      Index index = new Index(this, Columns.listToArray(colList));
      //主キー、外部キーにより暗黙で作成されるインデックスと重複する定義はADDせずスキップする
      if (primaryKey!=null && primaryKey.equalsColArray(index)) return index;
      //for (ForeginKey fk : getForeginKeys()) {
      //  if (fk.equalsColArray(index)) return index;
      //}
      getIndexes().add(index);
      return getIndexes().get(indexes.size()-1);
    }

    public String toCreateAllSQL() {
      StringBuffer buff = new StringBuffer();
      buff.append(toCreateTableSQL());
      if (getPrimaryKey()!=null) {
        buff.append(getPrimaryKey().toString());
        buff.append(";\n");
      }
      
      for (ForeginKey fk : getForeginKeys()) {
        buff.append(fk.toString());
        buff.append(";\n");
      }
      
      for (Index index : getIndexes()) {
        buff.append(index.toString());
        buff.append(";\n");
      }
      return buff.toString();
    }

    public String toDropTableSQL() {
      return "DROP TABLE " + getDBTableName(true);
    }
    
    public String toCreateTableSQL() {
      return toCreateTableSQL(true, null);
    }
    
    public String toCreateTableSQL(boolean withSchema, String option) {
      StringBuffer buff = new StringBuffer();
      buff.append("CREATE ");
      buff.append( (option!=null) ? option : "");
      buff.append(" TABLE ");
      buff.append(getDBTableName(withSchema));
      buff.append(" (\n");
      buff.append(toStringColumns());
      buff.append(")\n");
      return buff.toString();
    }

    public String toStringColumns() {
      return columns.toString(false);
    }    
    public Column getColumnByDBName(String dbColumnName) {
      return getColumns().getByDBColumnName(dbColumnName);
    }

    public String getOwner() {
      return owner;
    }

    public void setOwner(String owner) {
      this.owner = owner;
    }
    
  }
}
