/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.data;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.text.DecimalFormat;
import java.text.ParsePosition;
import java.text.SimpleDateFormat;
import java.util.HashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.json.JSONException;

import jp.littlesoft.data.Column.Type;
import jp.littlesoft.data.event.TableEvent;
import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.data.exception.InvalidColumnException;
import jp.littlesoft.data.exception.InvalidValueException;
import jp.littlesoft.soja.LSJson;
import jp.littlesoft.util.Const;
import jp.littlesoft.util.DataUtil;
import jp.littlesoft.util.IndexedMap;
import jp.littlesoft.util.StringUtil;

/**
 * データ行クラス。<br/><br/>
 *
 * Table オブジェクト内部で、フィールドに対応するデータ値を記録する為に使用されています。<br/>
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
public class Row<E extends Serializable> implements Serializable {

  private static final long serialVersionUID = 4946934585870174155L;
  /**
   * LONGBINARYを読み込む際のバッファサイズ<br/>
   */
  private static final int BUFFER_SIZE = 1024;
  
  /**
   * Columnの集合を保持する内部変数<br/>
   */
  private Table<E> table = null;
  
  /**
   * 「この行はアプリケーション内で削除される」という予約
   */
  private boolean deletionReserved = false;
 
  /** 
   * record Idを保持する内部変数です。<br/>
   */
  private RowID rowID = null;
  
  /**
   * 行の状態を保持する内部変数です。<br/>
   */
  protected Status status = Status.UNKNOWN;
  
  /**
   * 行項目の値を保持する配列です。<br/>
   */
  protected Serializable[] values = null;
  
  /**
   * 行が編集されている間、編集前の値を一時的に保持するRowオブジェクト<br/>
   */
  @SuppressWarnings("unchecked")
  private Row oldRow = null;
  
  /**
   * ロード時の値を保持するRowオブジェクト<br/>
   */
  @SuppressWarnings("unchecked")
  private Row originalRow = null;
  
  /**
   * エンティティ型が指定された行オブジェクトか？
   */
  private boolean typedRow = true;
  /**
   * カーソル行のエンティティオブジェクトへのショートカット.<br/>
   * row.getEntity();<br>
   * で取得できるオブジェクトと同じ内容です。
   */
  public E entity = null;
  
  /**
   * エンティティ型で定義されていない値を保持するMap
   */
  private HashMap<Integer, Serializable> undefinedValues = null;
  
  protected Row() {}
  
  /**
   * Row オブジェクトを生成します。<br/>
   * @param table この行を格納するテーブルオブジェクト
   */
  public Row(Table<E> table) {
    this.table = table;
    if (this.table!=null) {
      this.rowID = new Row.RowID(++this.table.rowCounter);
      initializeValues();
    }
  }
  
  /**
   * Row オブジェクトを生成します。<br/>
   * @param table この行を格納するテーブルオブジェクト
   */
  public Row(Table<E> table, E entity) {
    this.table = table;
    this.rowID = new Row.RowID(++this.table.rowCounter);
    this.entity = entity;
  }

  protected void initializeValues() {
    if (isTypedRow()) {
      //Entity : javaクラスで型が定義された行の場合
      try {
        if (entity==null) {
          entity = (E)table.getEntityType().newInstance();
        } else {
          Serializable temp = (E)table.getEntityType().newInstance();
          for (Column col : table.getColumns().values()){
            if (col.isReadOnly()) continue;
            col.setEntityValue(entity, col.getEntityValue(temp));
          }
        }
      } catch (Exception x) {
        throw new RuntimeException(x);
      }
    } else {
      //型なしの行なので配列で管理
      values = new Serializable[this.table.getColumnCount()];
      entity = null;
    }
  }
  

  public boolean isTypedRow() {
    return (typedRow && table!=null && 
        table.getEntityType()!=null && 
        !table.getEntityType().isAnnotation() &&
        !table.getEntityType().isArray() &&
        !table.getEntityType().isPrimitive() &&
        !table.getEntityType().equals(Object.class) &&
        !table.getEntityType().equals(Serializable.class) &&
        !table.getEntityType().equals(Row.class));
  }
  
  
  

  /**
   * Row オブジェクトを生成します。<br/>
   * @param table この行を格納するテーブルオブジェクト
   * @param status この行を初期状態(Row.Status.CREATINGまたはLOADEDを指定してください。
   */
  public Row(Table<E> table, Row.Status status) {
    this(table);
    this.setStatus(status);
  }
  
  /**
   * entityを格納するRow オブジェクトを生成します。<br/>
   * @param table table この行を格納するテーブルオブジェクト
   * @param status この行を初期状態(Row.Status.CREATINGまたはLOADEDを指定してください。
   * @param entity エンティティ
   */
  public Row(Table<E> table, Row.Status status, E entity) {
    this();
    this.table = table;
    this.rowID = new Row.RowID(++this.table.rowCounter);
    this.entity = entity;
    this.setStatus(status);
  }
  
  
  /**
   * Row オブジェクトの状態を示す列挙型。<br/><br/>
   * SQLDataResolverはこのStatusに基づき、データベーステーブルの更新SQLを生成または適用します。<br/>
   * @see Row
   */
  public enum Status {
    /**
     * Rowオブジェクトを生成したディフォルトの状態です。<br/>
     * この状態では、行に対する設定値の検証機能やイベント通知機能は動作しません。
     */
    UNKNOWN,

    /**
     * この行がデータベースあるいはファイルから読み込まれ、何の変更も行われていないことを表します。<br/>
     */
    LOADED,
    
    /**
     * この行はアプリケーションで新規作成され、内容を編集であることを表します。
     */
    CREATING,

    /**
     * この行はアプリケーションプログラムで新規作成され、編集が完了したことを表します。<br/>
     * SQLDataResolver、WebSQLTableは、この状態の行をINSERTの対象にします。
     */
    CREATED,
    
    /**
     * この行はデータベース等からロードされた後、アプリケーションプログラムで内容を編集であることを表します。
     */
    UPDATING,
    
    /**
     * この行はデータベース等からロードされた後、アプリケーションプログラムで内容が変更され、その編集が完了したことを表します。<br/>
     * SQLDataResolver、WebSQLTableは、この状態の行をUPDATEの対象にします。
     */
    UPDATED,

    /**
     * この行はデータベース等からロードされた後、アプリケーションプログラムで削除されたものであることを表します。
     * SQLDataResolver、WebSQLTableは、この状態の行をDELETEの対象にします。
     */
    DELETED
  }
  
  /**
   * Rowオブジェクトを識別する為のID値クラス。<br/>
   * 
   * @author LittleSoft Corporation
   */
  public final static class RowID implements Serializable {

    private static final long serialVersionUID = 1L;
    
    int value;
    /**
     * 引数 value で指定されたID値で RowID オブジェクトを生成します。<br/>
     * 
     * @param value ID値
     */
    public RowID(int value) {
      this.value = value;
    }
    
    @Override
    public boolean equals(Object obj) {
      if(!(obj instanceof RowID)) { return false; } 
      return (((RowID)obj).value == this.value);
    }
    
    @Override
    public String toString() {
      return "" + value;
    }
    
    @Override
    public int hashCode() {
      return String.valueOf(value).hashCode();
    }

  }
  
    
  /**
   * 変更前の Row オブジェクトを取得します。<br/>
   * 
   * @return 変更前の Row オブジェクト
   */
  @SuppressWarnings("unchecked")
  public Row getOriginalRow() {
    return this.originalRow;
  }
  
  /**
   * 保持している項目の値をクリアします。
   */
  public void clear() {
    if (isTypedRow()) {
      try {
        for (int i=0 ; i<getColumnCount(); i++) {
          internalSetValue(i, null);
        }        
      } catch (DataException x) {
        throw new RuntimeException(x);
      }
    } else {
      values = new Serializable[getColumnCount()];
    }
  }
  
   /**
   * Row オブジェクトの状態を設定します。<br/>
   * 
   * @param status Row オブジェクトの状態
   * @see Row.Status
   */
  public void setStatus(Status status) {
    switch(status) {
    case LOADED:
      this.originalRow = null;
      this.oldRow = null;
      break;
    case CREATED:
    case UPDATED:
      this.oldRow = null;
      break;  
    case DELETED:
      saveLoadedRow();
      this.oldRow = null;
      this.deletionReserved = false;
      break;
    case UPDATING:
      saveLoadedRow();
      saveOldRow();
      break;
    case CREATING:
      if (this.status==Row.Status.UNKNOWN) {
        this.setDefault();
        this.status = status;
      }
      saveOldRow();
      break;
    }
    this.status = status;
  }
  
  /**
   * Row オブジェクトの状態を取得します。<br/>
   * 
   * @return Row オブジェクトの状態
   * @see Row.Status
   */
  public Row.Status getStatus() {
    return this.status;
  }
  
  /**
   * Row オブジェクトのRowIDを取得します。<br/>
   * 
   * @return RecordID 
   */
  public RowID getRowID() {
    return this.rowID;
  }
  
  /**
   * Row オブジェクトに JDBC ResultSetの内容を設定します。<br/>
   * 
   * @param resultSet JDBC ResultSet
   * @param resultSetColumnNames 列名の配列
   * @throws java.sql.SQLException SQL例外発生時
   * @throws DataException 例外発生時
   */
  public synchronized void setValueFromResultSet(java.sql.ResultSet resultSet, String[] resultSetColumnNames)
      throws java.sql.SQLException, DataException {
    int resultSetColCount = resultSetColumnNames.length;
    IndexedMap<String, Column> columns = this.table.columns;
    for (int pos = 1; pos <= resultSetColCount; pos++) {
      final String resultSetColumnName = resultSetColumnNames[pos - 1];
      int myIndex = table.indexOfDBColumnName(resultSetColumnName);
      Column.Type type = columns.getValue(myIndex).getType();
      switch (type) {
      case STRING:
        String value = resultSet.getString(pos);
        if (value == null) {
          setNull(myIndex);
        } else {
          setString(myIndex, DataUtil.replaceFromDB(value));
        }
        break;
      case BINARY:
        setBinary(myIndex, resultSet.getBytes(pos));
        break;
      case LONGBINARY:
        byte[] longBinary = null;
        InputStream in = resultSet.getBinaryStream(pos);
        if (in != null) {
          BufferedInputStream bin = new BufferedInputStream(in, BUFFER_SIZE);
          ByteArrayOutputStream baos = new ByteArrayOutputStream();
          BufferedOutputStream out = new BufferedOutputStream(baos, BUFFER_SIZE);
          try {
            byte[] buf = new byte[BUFFER_SIZE];
            int count;
            while ((count = bin.read(buf)) >= 0) {
              out.write(buf, 0, count);
            }
            out.flush();
            longBinary = baos.toByteArray();
          } catch (Throwable ex) {
            throw new DataException(ex);
          } finally {
            try {
              out.close();
            } catch (IOException ex) {
            }
            ;
            try {
              bin.close();
            } catch (IOException ex) {
            }
            ;
            try {
              in.close();
            } catch (IOException ex) {
            }
            ;
          }
        }
        setBinary(myIndex, longBinary);
        break;
      case BLOB:
        java.sql.Blob blob = resultSet.getBlob(pos);
        if (blob == null) {
          setBinary(myIndex, null);
        } else {
          long blobLen = blob.length();
          if (blobLen > (long) Integer.MAX_VALUE) {
            throw new InvalidColumnException("DATA_SIZE_ERROR",
                new String[] { columns.get(resultSetColumnName)
                    .getCaption() }, columns.get(resultSetColumnName));
          }
          setBinary(myIndex, blob.getBytes(1, (int) blobLen));
        }
        break;
      case CLOB:
        java.sql.Clob clob = resultSet.getClob(pos);
        if (clob == null) {
          setString(myIndex, null);
        } else {
          long clobLen = clob.length();
          if (clobLen > (long) Integer.MAX_VALUE) {
            throw new InvalidColumnException("DATA_SIZE_ERROR",
                new String[] { columns.get(resultSetColumnName).getCaption() }, 
                columns.get(resultSetColumnName));
          }
          setString(myIndex, DataUtil.replaceFromDB(clob.getSubString(1L,
              (int) clobLen)));
        }
        break;
      case BYTE:
        setByte(myIndex, resultSet.getByte(pos));
        break;
      case SHORT:
        setShort(myIndex, resultSet.getShort(pos));
        break;
      case LONG:
        setLong(myIndex, resultSet.getLong(pos));
        break;
      case INT:
        setInt(myIndex, resultSet.getInt(pos));
        break;
      case FLOAT:
        setFloat(myIndex, resultSet.getFloat(pos));
        break;
      case DOUBLE:
        setDouble(myIndex, resultSet.getDouble(pos));
        break;
      case BOOLEAN:
        setBoolean(myIndex, resultSet.getBoolean(pos));
        break;
      case BIGDECIMAL:
        setBigDecimal(myIndex, resultSet.getBigDecimal(pos));
        break;
      case DATE:
        setDate(myIndex, resultSet.getDate(pos));
        break;
      case TIME:
        setTime(myIndex, resultSet.getTime(pos));
        break;
      case TIMESTAMP:
        setTimestamp(myIndex, resultSet.getTimestamp(pos));
        break;
      case JSONOBJECT:
        try {
          setObject(myIndex, LSJson.deserialize(resultSet.getString(pos)));
        } catch (JSONException e) {
          throw new DataException(e);
        }
       
        break;
      case UNKNOWN:
      default:
        String _caption = columns.get(resultSetColumnName).getCaption();
        throw new InvalidColumnException("INVALID_DATAFIELD_TYPE2",
            new String[] { _caption,
                type.toString() }, columns.get(resultSetColumnName));
      } // end of switch
      // intなどのprimitive型でnullの場合はnullを設定する。
      if (resultSet.wasNull())
        setNull(myIndex);
    } // end of for
  }
  
  /**
   * 互換性の無いデータ型操作を行った場合の例外送出を行います。<br/><br/>
   * 
   * 互換性の無いデータ型操作を行った場合に、このメソッドが呼ばれます。<br/>
   * 「Msg.INVALID_OPER_NOT_COMPATIBLE」(パラメタはindex)という エラーメッセージの例外を発生させます。<br/>
   * 
   * @param index 列インデックス
   * @throws InvalidColumnException 例外発生時
   */
  protected void handleCompatiException(int index) throws InvalidColumnException {
    Column _column;
    String _caption;
    try {
      _column = getColumn(index);
      _caption = _column.getCaption();
    } catch (Throwable t) {
      _column = null;
      _caption = Const.NULLSTRING;
    }

    throw new InvalidColumnException("INVALID_OPER_NOT_COMPATIBLE",
        new String[] { _caption }, _column);
  }
  
  /**
   * Row オブジェクトの引数 columnName で指定された列にjava.lang.String値を設定します。<br/>
   * 
   * @param columnName 列名
   * @param value java.lang.String型のデータ値
   * @throws DataException 例外発生時
   */
  public synchronized void setValueAsString(String columnName, String value) throws DataException {
    getColumn(columnName);
    setValueAsString(indexOfColumn(columnName), value);
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にjava.lang.String値を設定します。<br/>
   * 
   * @param index 列インデックス
   * @param value java.lang.String型のデータ値
   * @throws DataException 例外発生時
   */
  public synchronized void setValueAsString(int index, String value) throws DataException {
    if (isNull(index) && StringUtil.isNullString(value)) 
      return;
    if (!isNull(index) && StringUtil.isNotNullString(value)
         && getValueAsString(index).equals(value)) {
      return;
    }
    setValue(index, this.parseObject(getColumn(index), value));
  }
  
  /**
   * Row オブジェクトの引数 columnName で指定された列にjava.lang.Object値を設定します。<br/>
   * 
   * @param columnName 列名
   * @param value java.lang.Object型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setValue(String columnName, Serializable value) throws DataException, InvalidValueException {
    getColumn(columnName);
    setValue(indexOfColumn(columnName), value);
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にjava.lang.Object値を設定します。<br/>
   * 
   * @param index 列インデックス
   * @param value java.lang.Object型のデータ値
   * @throws DataException 例外発生時
   */
  public synchronized void setValue(int index, Serializable value) throws DataException {
    Column col = getColumn(index);
    if (value==null) {
      setNull(index);
      return;
    }
    checkColumnType(getColumn(index), value);   
    switch (col.getType()) {
    case STRING:
    case CLOB:
      setString(index, (String) value);
      break;
    case BINARY:
    case LONGBINARY:
    case BLOB:
      setBinary(index, (byte[]) value);
      break;
    case BYTE:
      setByte(index, ((Number) value).byteValue());
      break;
    case SHORT:
      setShort(index, ((Number) value).shortValue());
      break;
    case LONG:
      setLong(index, ((Number) value).longValue());
      break;
    case INT:
      setInt(index, ((Number) value).intValue());
      break;
    case FLOAT:
      setFloat(index, ((Number) value).floatValue());
      break;
    case DOUBLE:
      setDouble(index, ((Number) value).doubleValue());
      break;
    case BOOLEAN:
      setBoolean(index, ((Boolean) value).booleanValue());
      break;
    case BIGDECIMAL:
      setBigDecimal(index, (java.math.BigDecimal) value);
      break;
    case DATE:
      if (value instanceof java.sql.Timestamp) {
        setDate(index, new java.sql.Date(((java.sql.Timestamp) value).getTime()));
      } else if (value instanceof java.util.Date) {
        long time = ((java.util.Date) value).getTime();
        setDate(index, new java.sql.Date(time));
      } else {
        setDate(index, (java.sql.Date) value);
      }
      break;
    case TIME:
      if (value instanceof java.sql.Timestamp) {
        setTime(index,
            new java.sql.Time(((java.sql.Timestamp) value).getTime()));
      } else {
        setTime(index, (java.sql.Time) value);
      }
      break;
    case TIMESTAMP:
      if (value instanceof java.sql.Date) {
        setTimestamp(index, new java.sql.Timestamp(((java.sql.Date) value).getTime()));
      } else if (value instanceof java.sql.Time) {
        setTimestamp(index, new java.sql.Timestamp(((java.sql.Time) value).getTime()));
      } else {
        setTimestamp(index, (java.sql.Timestamp) value);
      }
      break;
    case JSONOBJECT:
      setObject(index, value);
      break;
    case UNKNOWN:
    default:
      String _caption = getColumn(index).getCaption();
      throw new InvalidColumnException("INVALID_DATAFIELD_TYPE2",
          new String[] { _caption,
              col.getType().toString() }, getColumn(index));
    }
  }
  
  /**
   * カラムのデータ型を決定するメソッド<br/>
   * value型をもとにデータ型を決定し、columnのTypeプロパティを設定しています。
   * @param column Column オブジェクト
   * @param value フィールド値
   */
  private void checkColumnType(Column column, Serializable value) {
    
    if (!column.getType().equals(Column.Type.UNKNOWN)) return;
    
    if (value instanceof String)
      column.setType(Column.Type.STRING);
    else if (value instanceof byte[])
      column.setType(Column.Type.BINARY);
    else if (value instanceof Byte)
      column.setType(Column.Type.BYTE);
    else if (value instanceof Short)
      column.setType(Column.Type.SHORT);
    else if (value instanceof Long)
      column.setType(Column.Type.LONG);
    else if (value instanceof Integer)
      column.setType(Column.Type.INT);
    else if (value instanceof Float)
      column.setType(Column.Type.FLOAT);
    else if (value instanceof Double)
      column.setType(Column.Type.DOUBLE);
    else if (value instanceof Boolean)
      column.setType(Column.Type.BOOLEAN);
    else if (value instanceof java.math.BigDecimal)
      column.setType(Column.Type.BIGDECIMAL);
    else if (value instanceof java.sql.Date)
      column.setType(Column.Type.DATE);
    else if (value instanceof java.sql.Time)
      column.setType(Column.Type.TIME);
    else if (value instanceof java.sql.Timestamp)
      column.setType(Column.Type.TIMESTAMP);
    else
      column.setType(Column.Type.JSONOBJECT);
  }
  
  /**
   * Row オブジェクト設定されたjava.lang.String値に対して、トリム処理を行います。<br/>
   * 
   * @param index
   *          列インデックス
   * @param value
   *          トリム処理を行うデータ値
   * @return トリム処理後のデータ値
   * @throws DataException
   *           例外発生時
   */
  private String trimString(int index, String value) throws DataException {
    if (value == null)
      return null;
    switch (getColumn(index).getTrimType()) {
    case BOTH:
      return value.trim();
    case LEFT:
      return StringUtil.trimLeft(value);
    case RIGHT:
      return StringUtil.trimRight(value);
    case NOTRIM:
    default:
      return value;
    }
  }

  
  /**
   * Row オブジェクトの引数 columnName で指定された列にjava.lang.String値を設定します。<br/>
   * 
   * @param columnName 列名
   * @param value java.lang.String型のデータ値
   * @throws DataException 例外発生時
   */
  public void setString(String columnName, String value) throws DataException {
    setString(indexOfColumn(columnName), value);
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にjava.lang.String値を設定します。<br/>
   * 
   * @param index 列インデックス
   * @param value java.lang.String型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setString(int index, String value) throws DataException, InvalidValueException {
    //Column col = table.getColumn(index);
    //value = abstractTable.checkValidation(col, value);
    checkValidation(index, value);
    value = trimString(index, value);
    if (StringUtil.isNullString(value)) {
      value = null;
    }
    switch (getColumn(index).getType()) {
    case STRING:
    case CLOB:
      internalSetValue(index, value);
      break;
    default:
      handleCompatiException(index);
    }
    //ChangeEvent fieldHandleEvent = new ChangeEvent(abstractTable, col, newValue, null);
    //abstractTable.fireAfterSetValue(fieldHandleEvent);    
  }
  
  /**
   * Row オブジェクトの引数 columnName で指定された列にバイナリ値（byte[]）を設定します。<br/>
   * 
   * @param columnName 列名
   * @param value バイナリ値（byte[]）
   * @throws DataException 例外発生時
   */
  public void setBinary(String columnName, byte[] value) throws DataException {
    getColumn(columnName);
    setBinary(indexOfColumn(columnName), value);
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にバイナリ値（byte[]）を設定します。<br/>
   * 
   * @param index 列インデックス
   * @param value バイナリ値（byte[]）
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setBinary(int index, byte[] value) throws DataException, InvalidValueException {
    //Column col = table.getColumn(index);
    //final Object newValue = checkValidation(col, value);
    //cursorRow.setBinary(index, (byte[]) newValue);
    //ChangeEvent fieldHandleEvent = new ChangeEvent(this, col, newValue, null);
    //fireAfterSetValue(fieldHandleEvent);
    checkValidation(index, value);
    switch (getColumn(index).getType()) {
    case BINARY:
    case LONGBINARY:
    case BLOB:
      internalSetValue(index, value);
      break;
    default:
      handleCompatiException(index);
    }
  }
  
  /**
   * Row オブジェクトの引数 columnName で指定された列にbyte値を設定します。<br/>
   * 
   * @param columnName 列名
   * @param value byte型のデータ値
   * @throws DataException 例外発生時
   */
  public synchronized void setByte(String columnName, Byte value) throws DataException {
    setByte(indexOfColumn(columnName), value);
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にbyte値を設定します。<br/>
   * 
   * @param index 列インデックス
   * @param value byte型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setByte(int index, Byte value) throws DataException, InvalidValueException {
    checkValidation(index, value);
    switch (getColumn(index).getType()) {
    case BYTE:
      internalSetValue(index, value);
      break;
    case SHORT:
      internalSetValue(index, new Short(value));
      break;
    case INT:
      internalSetValue(index, new Integer(value));
      break;
    case LONG:
      internalSetValue(index, new Long(value));
      break;
    case FLOAT:
      internalSetValue(index, new Float(value));
      break;
    case DOUBLE:
      internalSetValue(index, new Double(value));
      break;
    case BIGDECIMAL:
      internalSetValue(index, new java.math.BigDecimal(value));
      break;
    default:
      handleCompatiException(index);
    }
  }
  
  /**
   * Row オブジェクトの引数 columnName で指定された列にshort値を設定します。<br/>
   * 
   * @param columnName 列名
   * @param value short型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setShort(String columnName, Short value) throws DataException, InvalidValueException {
    setShort(indexOfColumn(columnName), value);
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にshort値を設定します。<br/>
   * 
   * @param index 列インデックス
   * @param value short型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setShort(int index, Short value) throws DataException, InvalidValueException {
    checkValidation(index, value);
    switch (getColumn(index).getType()) {
    case SHORT:
      internalSetValue(index, value);
      break;
    case INT:
      internalSetValue(index, new Integer(value));
      break;
    case LONG:
      internalSetValue(index, new Long(value));
      break;
    case FLOAT:
      internalSetValue(index, new Float(value));
      break;
    case DOUBLE:
      internalSetValue(index, new Double(value));
      break;
    case BIGDECIMAL:
      internalSetValue(index, new java.math.BigDecimal(value));
      break;
    default:
      handleCompatiException(index);
    }
  }
  
  /**
   * Row オブジェクトの引数 columnName で指定された列にint値を設定します。<br/>
   * 
   * @param columnName 列名
   * @param value int型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setInt(String columnName, Integer value) throws DataException, InvalidValueException {
    setInt(indexOfColumn(columnName), value);
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にint値を設定します。<br/>
   * 
   * @param index 列インデックス
   * @param value int型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setInt(int index, Integer value) throws DataException, InvalidValueException {
    checkValidation(index, value);
    switch (getColumn(index).getType()) {
    case INT:
      internalSetValue(index, value);
      break;
    case LONG:
      internalSetValue(index, new Long(value));
      break;
    case FLOAT:
      internalSetValue(index, new Float(value));
      break;
    case DOUBLE:
      internalSetValue(index, new Double(value));
      break;
    case BIGDECIMAL:
      internalSetValue(index, new java.math.BigDecimal(value));
      break;
    default:
      handleCompatiException(index);
    }
  }
  
  /**
   * Row オブジェクトの引数 columnName で指定された列にlong値を設定します。<br/>
   * 
   * @param columnName 列名
   * @param value long型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setLong(String columnName, Long value) throws DataException, InvalidValueException {
    setLong(indexOfColumn(columnName), value);    
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にlong値を設定します。<br/>
   * 
   * @param index 列インデックス
   * @param value long型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setLong(int index, Long value) throws DataException, InvalidValueException {
    checkValidation(index, value);
    switch (getColumn(index).getType()) {
    case LONG:
      internalSetValue(index, value);
      break;
    case FLOAT:
      internalSetValue(index, new Float(value));
      break;
    case DOUBLE:
      internalSetValue(index, new Double(value));
      break;
    case BIGDECIMAL:
      internalSetValue(index, new java.math.BigDecimal(value));
    default:
      handleCompatiException(index);
    }
  }
  
  /**
   * Row オブジェクトの引数 columnName で指定された列にfloat値を設定します。<br/>
   * 
   * @param columnName 列名
   * @param value float型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setFloat(String columnName, Float value) throws DataException, InvalidValueException {
    setFloat(indexOfColumn(columnName), value);    
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にfloat値を設定します。<br/>
   * 
   * @param index 列インデックス
   * @param value float型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setFloat(int index, Float value) throws DataException, InvalidValueException {
    checkValidation(index, value);
    switch (getColumn(index).getType()) {
    case FLOAT:
      internalSetValue(index, value);
      break;
    case DOUBLE:
      internalSetValue(index, new Double(value));
      break;
    case BIGDECIMAL:
      internalSetValue(index, new java.math.BigDecimal(value));
      break;
    default:
      handleCompatiException(index);
    }
  }
  
  /**
   * Row オブジェクトの引数 columnName で指定された列にdouble値を設定します。<br/>
   * 
   * @param columnName 列名
   * @param value double型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setDouble(String columnName, Double value) throws DataException, InvalidValueException {
    setDouble(indexOfColumn(columnName), value);       
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にdouble値を設定します。<br/>
   * 
   * @param index 列インデックス
   * @param value double型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setDouble(int index, Double value) throws DataException, InvalidValueException {
    checkValidation(index, value);
    switch (getColumn(index).getType()) {
    case DOUBLE:
      internalSetValue(index, value);
      break;
    case BIGDECIMAL:
      internalSetValue(index, new java.math.BigDecimal(value));
      break;
    default:
      handleCompatiException(index);
    }
  }
  
  /**
   * Row オブジェクトの引数 columnName で指定された列にboolean値を設定します。<br/>
   * 
   * @param columnName 列名
   * @param value boolean型のデータ値
   * @throws DataException 例外発生時
   */
  public synchronized void setBoolean(String columnName, Boolean value) throws DataException {
    setBoolean(indexOfColumn(columnName), value);          
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にboolean値を設定します。<br/>
   * 
   * @param index 列インデックス
   * @param value javaboolean型のデータ値
   * @throws DataException 例外発生時
   */
  public synchronized void setBoolean(int index, Boolean value) throws DataException {
    checkValidation(index, (value ? Boolean.TRUE : Boolean.FALSE));
    switch (getColumn(index).getType()) {
    case BOOLEAN:
      internalSetValue(index, value);
      break;
    default:
      handleCompatiException(index);
    }
  }
  
  private Serializable internalGetValue(int index) {
    if (isTypedRow()) {
      try {
        return getColumn(index).getEntityValue(entity);
      } catch (Column.NoAccessibleObjectException x) {
        if (undefinedValues != null && undefinedValues.containsKey(index) ) {
          return undefinedValues.get(index);
        } else {
          return null;
        }
      }
    } else {
      return values[index];
    }    
  }
  
  private void internalSetValue(int index, Serializable value) throws DataException {
    if (isTypedRow()) {
      try {
        getColumn(index).setEntityValue(entity, value);
      } catch (NullPointerException n) {
        if (value == null) return;
        throw new DataException(getColumn(index).getAccessibleObjectName() + " did a null value. ");
      } catch (Column.NoAccessibleObjectException x) {
        if (undefinedValues==null) {
          undefinedValues = new HashMap<Integer, Serializable>();
        }
        undefinedValues.put(index, value);
      }
    } else {
      values[index] = value;
    }
  }
  /**
   * Row オブジェクトの引数 columnName で指定された列にjava.math.BigDecimal値を設定します。<br/>
   * 
   * @param columnName 列名
   * @param value java.math.BigDecimal型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setBigDecimal(String columnName, java.math.BigDecimal value) throws DataException, InvalidValueException {
    setBigDecimal(indexOfColumn(columnName), value);        
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にjava.math.BigDecimal値を設定します。<br/>
   * 
   * @param index 列インデックス
   * @param value java.math.BigDecimal型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setBigDecimal(int index, java.math.BigDecimal value) throws DataException, InvalidValueException {
    checkValidation(index, value);
    switch (getColumn(index).getType()) {
    case BIGDECIMAL:
      //this.values[index] = value;
      internalSetValue(index, value);
      break;
    case DOUBLE:
      //this.values[index] = value.doubleValue();
      internalSetValue(index, value.doubleValue());
      break;
    case FLOAT:
      //this.values[index] = value.floatValue();
      internalSetValue(index, value.floatValue());
      break;
    case INT:
      //this.values[index] = value.intValue();
      internalSetValue(index, value.intValue());
      break;
    case LONG:
      //this.values[index] = value.longValue();
      internalSetValue(index, value.longValue());
      break;
    case BYTE:
      //this.values[index] = value.byteValue();
      internalSetValue(index, value.byteValue());
      break;
    case SHORT:
      //this.values[index] = value.shortValue();
      internalSetValue(index, value.shortValue());
      break;
    default:
      handleCompatiException(index);
    }
  }
  
  /**
   * Row オブジェクトの引数 columnName で指定された列にjava.sql.Date値を設定します。<br/>
   * 
   * @param columnName 列名
   * @param value java.sql.Date型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setDate(String columnName, java.sql.Date value) throws DataException, InvalidValueException {
    setDate(indexOfColumn(columnName), value);     
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にjava.sql.Date値を設定します。<br/>
   * 
   * @param index 列インデックス
   * @param value java.sql.Date型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setDate(int index, java.sql.Date value) throws DataException, InvalidValueException {
    checkValidation(index, value);
    switch (getColumn(index).getType()) {
    case DATE:
      internalSetValue(index, value);
      break;
    case TIMESTAMP:
      internalSetValue(index, new java.sql.Timestamp(value.getTime()));
      break;
    case TIME:
      internalSetValue(index, new java.sql.Time(value.getTime()));
      break;
    default:
      handleCompatiException(index);
    }
  }
  
  /**
   * Row オブジェクトの引数 columnName で指定された列にjava.sql.Time値を設定します。<br/>
   * 
   * @param columnName 列名
   * @param value java.sql.Time型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setTime(String columnName, java.sql.Time value) throws DataException, InvalidValueException {
    setTime(indexOfColumn(columnName), value);        
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にjava.sql.Time値を設定します。<br/>
   * 
   * @param index 列インデックス
   * @param value java.sql.Time型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setTime(int index, java.sql.Time value) throws DataException, InvalidValueException {
    checkValidation(index, value);
    switch (getColumn(index).getType()) {
    case TIME:
      internalSetValue(index, value);
      break;
    case TIMESTAMP:
      internalSetValue(index, new java.sql.Timestamp(value.getTime()));
      break;
    case DATE:
      internalSetValue(index, new java.sql.Date(value.getTime()));
      break;
    default:
      handleCompatiException(index);
    }
  }
  
  /**
   * Row オブジェクトの引数 columnName で指定された列にjava.sql.Timestamp値を設定します。<br/>
   * 
   * @param columnName 列名
   * @param value java.sql.Timestamp型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setTimestamp(String columnName, java.sql.Timestamp value) throws DataException, InvalidValueException {
    setTimestamp(indexOfColumn(columnName), value);    
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にjava.sql.Timestamp値を設定します。<br/>
   * 
   * @param index 列インデックス
   * @param value java.sql.Timestamp型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setTimestamp(int index, java.sql.Timestamp value) throws DataException, InvalidValueException {
    checkValidation(index, value);
    switch (getColumn(index).getType()) {
    case TIMESTAMP:
      internalSetValue(index, value);
      break;
    default:
      handleCompatiException(index);
    }
  }
  
  /**
   * Row オブジェクトの引数 columnName で指定された列にjava.lang.Object値を設定します。<br/>
   * 
   * @param columnName 列名
   * @param value java.lang.Object型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setObject(String columnName, Serializable value) throws DataException, InvalidValueException {
    setObject(indexOfColumn(columnName), value);    
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にjava.lang.Object値を設定します。<br/>
   * 
   * @param index 列インデックス
   * @param value java.lang.Object型のデータ値
   * @throws DataException 例外発生時
   * @throws InvalidValueException 例外発生時
   */
  public synchronized void setObject(int index, Serializable value) throws DataException, InvalidValueException {
    checkValidation(index, value);
    switch (getColumn(index).getType()) {
    case JSONOBJECT:
      internalSetValue(index, value);
      break;
    default:
      handleCompatiException(index);
    }
  }
  
  /**
   * Row オブジェクトの引数 columnName で指定された列にnull値を設定します。<br/>
   * 
   * @param columnName 列名
   * @throws InvalidColumnException 例外発生時
   */
  public synchronized void setNull(String columnName) throws DataException {
    setNull(indexOfColumn(columnName));        
  }
  
  /**
   * Row オブジェクトの引数 index で指定された列にnull値を設定します。<br/>
   * 
   * @param index 列インデックス
   * @throws DataException 
   */
  public synchronized void setNull(int index) throws DataException {
    checkValidation(index, null);
    internalSetValue(index, null);
  }
  
  /**
   * Row オブジェクトより引数 columnName で指定された列のjava.lang.String値を取得します。<br/><br/>
   * 
   * <em>※注意：</em>
   * <ul>
   *   <li>型が数値または日付の場合はDataUtilで暗黙の書式化が行われます。</li>
   *   <li>値がnullの場合は、nullではなくjp.littlesoft.util.Const.NULLSTRINGを返します。</li>
   *   <li>型がjava.lang.Stringの場合は getString(columnName) メソッドを利用した方がごく僅かですが高速です。</li>
   * </ul><br/>
   * 
   * @param columnName 列名
   * @return java.lang.String型のデータ値
   * @throws DataException 例外発生時
   */
  public String getValueAsString(String columnName) throws DataException {
    getColumn(columnName);
    return getValueAsString(indexOfColumn(columnName));    
  }
  
  /**
   * Row オブジェクトより引数 index で指定された列のjava.lang.String値を取得します。<br/><br/>
   * 
   * <em>※注意：</em>
   * <ul>
   *   <li>型が数値または日付の場合はDataUtilで暗黙の書式化が行われます。</li>
   *   <li>値がnullの場合は、nullではなくjp.littlesoft.util.Const.NULLSTRINGを返します。</li>
   *   <li>型がjava.lang.Stringの場合は getString(index) メソッドを利用した方がごく僅かですが高速です。</li>
   * </ul><br/>
   * 
   * @param index 列インデックス
   * @return java.lang.String型のデータ値
   * @throws DataException 例外発生時
   */
  public String getValueAsString(int index) throws DataException {
    if (isNull(index)) {
      if (getColumn(index).getType()==Type.BOOLEAN) {
        // boolean型のnullは"false"を返す。
        return new java.lang.Boolean(false).toString();
      } else {
        // nullの場合には "" を返す。
        return Const.NULLSTRING;
      }
    }
    switch (getColumn(index).getType()) {
      case DATE:
      case TIME:
      case TIMESTAMP:
        return this.formatDate(getColumn(index), (java.util.Date)getValue(index));
      default:
        return this.formatObject(getColumn(index), getValue(index));
    }
  }
  
  /**
   * Row オブジェクトより引数 columnName で指定された列のjava.lang.Object値を取得します。<br/>
   * 
   * @param columnName 列名
   * @return java.lang.Object型のデータ値
   * @throws DataException 例外発生時
   */
  public Serializable getValue(String columnName) throws DataException {
    getColumn(columnName);
    return getValue(indexOfColumn(columnName));
  }
  
  /**
   * Row オブジェクトより引数 index で指定された列のjava.lang.Object値を取得します。<br/>
   * 
   * @param index 列インデックス
   * @return java.lang.Object型のデータ値
   * @throws DataException 例外発生時
   */
  public Serializable getValue(int index) throws DataException {
    return this.internalGetValue(index);
  }
  
  /**
   * Row オブジェクトより引数 columnName で指定された列のjava.lang.String値を取得します。<br/>
   * 
   * @param columnName 列名
   * @return java.lang.String型のデータ値
   * @throws DataException 例外発生時
   */
  public String getString(String columnName) throws DataException {
    getColumn(columnName);
    return getString(indexOfColumn(columnName));
  }
  
  /**
   * Row オブジェクトより引数 index で指定された列のjava.lang.String値を取得します。<br/>
   * 
   * @param index 列インデックス
   * @return java.lang.String型のデータ値
   * @throws DataException 例外発生時
   */
  public String getString(int index) throws DataException {
    if (isNull(index))
      return null;
    switch (getColumn(index).getType()) {
    case STRING:
    case CLOB:
      return (String) this.internalGetValue(index);
    default:
      handleCompatiException(index);
    }
    return null;
  }
  
  /**
   * Row オブジェクトより引数 columnName で指定された列のバイナリ値（byte[]）を取得します。<br/>
   * 
   * @param columnName 列名
   * @return バイナリ値（byte[]）
   * @throws DataException 例外発生時
   */
  public byte[] getBinary(String columnName) throws DataException {
    getColumn(columnName);
    return getBinary(indexOfColumn(columnName));
  }
  
  /**
   * Row オブジェクトより引数 index で指定された列のバイナリ値（byte[]）を取得します。<br/>
   * 
   * @param index 列インデックス
   * @return バイナリ値（byte[]）
   * @throws DataException 例外発生時
   */
  public byte[] getBinary(int index) throws DataException {
    if (isNull(index))
      return null;
    switch (getColumn(index).getType()) {
    case BINARY:
    case LONGBINARY:
    case BLOB:
      return (byte[]) this.internalGetValue(index);
    default:
      handleCompatiException(index);
    }
    return null;
  }
  
  /**
   * Row オブジェクトより引数 columnName で指定された列のbyte値を取得します。<br/>
   * 
   * @param columnName 列名
   * @return byte型のデータ値
   * @throws DataException 例外発生時
   */
  public byte getByte(String columnName) throws DataException {
    getColumn(columnName);
    return getByte(indexOfColumn(columnName));
  }
  
  /**
   * Row オブジェクトより引数 index で指定された列のbyte値を取得します。<br/>
   * 
   * @param index 列インデックス
   * @return byte型のデータ値
   * @throws DataException 例外発生時
   */
  public byte getByte(int index) throws DataException {
    if (isNull(index))
      return 0;
    switch (getColumn(index).getType()) {
    case BYTE:
      return (Byte) this.internalGetValue(index);
    default:
      handleCompatiException(index);
    }
    return 0;
  }
  
  /**
   * Row オブジェクトより引数 columnName で指定された列のshort値を取得します。<br/>
   * 
   * @param columnName 列名
   * @return short型のデータ値
   * @throws DataException 例外発生時
   */
  public short getShort(String columnName) throws DataException {
    getColumn(columnName);
    return getShort(indexOfColumn(columnName));
  }
  
  /**
   * Row オブジェクトより引数 index で指定された列のshort値を取得します。<br/>
   * 
   * @param index 列インデックス
   * @return short型のデータ値
   * @throws DataException 例外発生時
   */
  public short getShort(int index) throws DataException {
    if (isNull(index))
      return 0;
    switch (getColumn(index).getType()) {
    case BYTE:
      return (Byte) this.internalGetValue(index);
    case SHORT:
      return (Short) this.internalGetValue(index);
    default:
      handleCompatiException(index);
    }
    return 0;
  }
  
  /**
   * Row オブジェクトより引数 columnName で指定された列のint値を取得します。<br/>
   * 
   * @param columnName 列名
   * @return int型のデータ値
   * @throws DataException 例外発生時
   */
  public int getInt(String columnName) throws DataException {
    getColumn(columnName);
    return getInt(indexOfColumn(columnName));
  }
  
  /**
   * Row オブジェクトより引数 index で指定された列のint値を取得します。<br/>
   * 
   * @param index 列インデックス
   * @return int型のデータ値
   * @throws DataException 例外発生時
   */
  public int getInt(int index) throws DataException {
    if (isNull(index))
      return 0;
    switch (getColumn(index).getType()) {
    case BYTE:
      return (Byte) this.internalGetValue(index);
    case SHORT:
      return (Short) this.internalGetValue(index);
    case INT:
      return (Integer) this.internalGetValue(index);
    default:
      handleCompatiException(index);
    }
    return 0;
  }
  
  /**
   * Row オブジェクトより引数 columnName で指定された列のlong値を取得します。<br/>
   * 
   * @param columnName 列名
   * @return long型のデータ値
   * @throws DataException 例外発生時
   */
  public long getLong(String columnName) throws DataException {
    getColumn(columnName);
    return getLong(indexOfColumn(columnName));    
  }
  
  /**
   * Row オブジェクトより引数 index で指定された列のlong値を取得します。<br/>
   * 
   * @param index 列インデックス
   * @return long型のデータ値
   * @throws DataException 例外発生時
   */
  public long getLong(int index) throws DataException {
    if (isNull(index))
      return 0;
    switch (getColumn(index).getType()) {
    case BYTE:
      return (Byte) this.internalGetValue(index);
    case SHORT:
      return (Short) this.internalGetValue(index);
    case INT:
      return (Integer) this.internalGetValue(index);
    case LONG:
      return (Long) this.internalGetValue(index);
    default:
      handleCompatiException(index);
    }
    return 0;
  }
  
  /**
   * Row オブジェクトより引数 columnName で指定された列のfloat値を取得します。<br/>
   * 
   * @param columnName 列名
   * @return float型のデータ値
   * @throws DataException 例外発生時
   */
  public float getFloat(String columnName) throws DataException {
    getColumn(columnName);
    return getFloat(indexOfColumn(columnName));    
  }
  
  /**
   * Row オブジェクトより引数 index で指定された列のfloat値を取得します。<br/>
   * 
   * @param index 列インデックス
   * @return float型のデータ値
   * @throws DataException 例外発生時
   */
  public float getFloat(int index) throws DataException {
    if (isNull(index))
      return 0;
    switch (getColumn(index).getType()) {
    case BYTE:
      return (Byte) this.internalGetValue(index);
    case SHORT:
      return (Short) this.internalGetValue(index);
    case INT:
      return (Integer) this.internalGetValue(index);
    case LONG:
      return (Long) this.internalGetValue(index);
    case FLOAT:
      return (Float) this.internalGetValue(index);
    default:
      handleCompatiException(index);
    }
    return 0;
  }
  
  /**
   * Row オブジェクトより引数 columnName で指定された列のdouble値を取得します。<br/>
   * 
   * @param columnName 列名
   * @return double型のデータ値
   * @throws DataException 例外発生時
   */
  public double getDouble(String columnName) throws DataException {
    getColumn(columnName);
    return getDouble(indexOfColumn(columnName));    
  }
  
  /**
   * Row オブジェクトより引数 index で指定された列のdouble値を取得します。<br/>
   * 
   * @param index 列インデックス
   * @return double型のデータ値
   * @throws DataException 例外発生時
   */
  public double getDouble(int index) throws DataException {
    if (isNull(index))
      return 0;
    switch (getColumn(index).getType()) {
    case BYTE:
      return (Byte) this.internalGetValue(index);
    case SHORT:
      return (Short) this.internalGetValue(index);
    case INT:
      return (Integer) this.internalGetValue(index);
    case LONG:
      return (Long) this.internalGetValue(index);
    case FLOAT:
      return (Float) this.internalGetValue(index);
    case DOUBLE:
      return (Double) this.internalGetValue(index);
    case BIGDECIMAL:
      // scale値も考慮する。
      java.math.BigDecimal bg = this.getBigDecimal(index);
      return bg.doubleValue();
    default:
      handleCompatiException(index);
    }
    return 0;
  }
  
  /**
   * Row オブジェクトより引数 columnName で指定された列のjava.math.BigDecimal値を取得します。<br/>
   * 
   * @param columnName 列名
   * @return java.math.BigDecimal型のデータ値
   * @throws DataException 例外発生時
   */
  public java.math.BigDecimal getBigDecimal(String columnName) throws DataException {
    getColumn(columnName);
    return getBigDecimal(indexOfColumn(columnName));    
  }
  
  /**
   * Row オブジェクトより引数 index で指定された列のjava.math.BigDecimal値を取得します。<br/>
   * 
   * @param index 列インデックス
   * @return java.math.BigDecimal型のデータ値
   * @throws DataException 例外発生時
   */
  public java.math.BigDecimal getBigDecimal(int index) throws DataException {
    if (isNull(index))
      return null;
    switch (getColumn(index).getType()) {
    case BYTE:
      return new java.math.BigDecimal((Byte) this.internalGetValue(index));
    case SHORT:
      return new java.math.BigDecimal((Short) this.internalGetValue(index));
    case INT:
      return new java.math.BigDecimal((Integer) this.internalGetValue(index));
    case LONG:
      return new java.math.BigDecimal((Long) this.internalGetValue(index));
    case FLOAT:
      // 文字型にすると誤差が混入しにくい
      return new java.math.BigDecimal(((Float) this.internalGetValue(index)).toString());
    case DOUBLE:
      // 文字型にすると誤差が混入しにくい
      return new java.math.BigDecimal(((Double) this.internalGetValue(index)).toString());
    case BIGDECIMAL:
      // ---- scale値を考慮する
      java.math.BigDecimal b = (java.math.BigDecimal) (this.internalGetValue(index));
      if (b != null && getColumn(index).getScale() >= 0) {
        return b.setScale(getColumn(index).getScale(),
            java.math.BigDecimal.ROUND_HALF_UP);
      } else {
        return b;
      }
    default:
      handleCompatiException(index);
    }
    return null;
  }
  
  /**
   * Row オブジェクトより引数 columnName で指定された列のboolean値を取得します。<br/>
   * 
   * @param columnName 列名
   * @return boolean型のデータ値
   * @throws DataException 例外発生時
   */
  public boolean getBoolean(String columnName) throws DataException {
    getColumn(columnName);
    return getBoolean(indexOfColumn(columnName));    
  }
  
  /**
   * Row オブジェクトより引数 index で指定された列のboolean値を取得します。<br/>
   * 
   * @param index 列インデックス
   * @return boolean型のデータ値
   * @throws DataException 例外発生時
   */
  public boolean getBoolean(int index) throws DataException {
    if (isNull(index))
      return false;
    switch (getColumn(index).getType()) {
    case BOOLEAN:
      return (Boolean) this.internalGetValue(index);
    default:
      handleCompatiException(index);
    }
    return false;
  }
  
  /**
   * Row オブジェクトより引数 columnName で指定された列のjava.sql.Date値を取得します。<br/>
   * 
   * @param columnName 列名
   * @return java.sql.Date型のデータ値
   * @throws DataException 例外発生時
   */
  public java.sql.Date getDate(String columnName) throws DataException {
    getColumn(columnName);
    return getDate(indexOfColumn(columnName));    
  }
  
  /**
   * Row オブジェクトより引数 index で指定された列のjava.sql.Date値を取得します。<br/>
   * 
   * @param index 列インデックス
   * @return java.sql.Date型のデータ値
   * @throws DataException 例外発生時
   */
  public java.sql.Date getDate(int index) throws DataException {
    if (isNull(index))
      return null;
    switch (getColumn(index).getType()) {
    case DATE:
      return (java.sql.Date) this.internalGetValue(index);
    default:
      handleCompatiException(index);
    }
    return null;
  }
  
  /**
   * Row オブジェクトより引数 columnName で指定された列のjava.sql.Time値を取得します。<br/>
   * 
   * @param columnName 列名
   * @return java.sql.Time型のデータ値
   * @throws DataException 例外発生時
   */
  public java.sql.Time getTime(String columnName) throws DataException {
    getColumn(columnName);
    return getTime(indexOfColumn(columnName));    
  }
  
  /**
   * Row オブジェクトより引数 index で指定された列のjava.sql.Time値を取得します。<br/>
   * 
   * @param index 列インデックス
   * @return java.sql.Time型のデータ値
   * @throws DataException 例外発生時
   */
  public java.sql.Time getTime(int index) throws DataException {
    if (isNull(index))
      return null;
    Serializable obj = this.internalGetValue(index);
    switch (getColumn(index).getType()) {
    case TIME:
      return (java.sql.Time) obj;
    case DATE:
      return new java.sql.Time(((java.sql.Date) obj).getTime());
    case TIMESTAMP:
      return new java.sql.Time(((java.sql.Timestamp) obj).getTime());
    default:
      handleCompatiException(index);
    }
    return null;
  }
  
  /**
   * Row オブジェクトより引数 columnName で指定された列のjava.sql.Timestamp値を取得します。<br/>
   * 
   * @param columnName 列名
   * @return java.sql.Timestamp型のデータ値
   * @throws DataException 例外発生時
   */
  public java.sql.Timestamp getTimestamp(String columnName) throws DataException {
    getColumn(columnName);
    return getTimestamp(indexOfColumn(columnName));    
  }
  
  /**
   * Row オブジェクトより引数 index で指定された列のjava.sql.Timestamp値を取得します。<br/>
   * 
   * @param index 列インデックス
   * @return java.sql.Timestamp型のデータ値
   * @throws DataException 例外発生時
   */
  public java.sql.Timestamp getTimestamp(int index) throws DataException {
    if (isNull(index))
      return null;
    Serializable obj = this.internalGetValue(index);
    switch (getColumn(index).getType()) {
    case TIMESTAMP:
      return (java.sql.Timestamp) obj;
    case DATE:
      return new java.sql.Timestamp(((java.sql.Date) obj).getTime());
    case TIME:
      return new java.sql.Timestamp(((java.sql.Time) obj).getTime());
    default:
      handleCompatiException(index);
    }
    return null;
  }
  
  /**
   * Row オブジェクトより引数 columnName で指定された列のjava.lang.Object値を取得します。<br/>
   * 
   * @param columnName 列名
   * @return java.lang.Object型のデータ値
   * @throws DataException 例外発生時
   */
  public Serializable getObject(String columnName) throws DataException {
    getColumn(columnName);
    return getObject(indexOfColumn(columnName));    
  }
  
  /**
   * Row オブジェクトより引数 index で指定された列のjava.lang.Object値を取得します。<br/>
   * 
   * @param index 列インデックス
   * @return java.lang.Object型のデータ値
   * @throws DataException 例外発生時
   */
  public Serializable getObject(int index) throws DataException {
    if (isNull(index))
      return null;
    switch (getColumn(index).getType()) {
    case JSONOBJECT:
      return this.internalGetValue(index);
    default:
      handleCompatiException(index);
    }
    return null;
  }
  
  /**
   * データエンティティオブジェクトを取得する.<br/>
   * @return
   */
  public E getEntity() {
    if (!isTypedRow()) throw new RuntimeException("Entity type is not specified.");
    return entity;
  }
  

  
  /**
   * 引数 columnName で指定された列がNULLであるか検査します。<br/>
   * 
   * @param columnName 列名
   * @return True：NULLである、Talse：NULLではない
   * @throws DataException 例外発生時
   */
  public boolean isNull(String columnName) throws DataException {
    getColumn(columnName);
    return isNull(indexOfColumn(columnName));    
  }
  
  /**
   * 引数 index で指定された列がNULLであるか検査します。<br/>
   * 
   * @param index 列インデックス
   * @return True：NULLである、Talse：NULLではない
   */
  public boolean isNull(int index) {
    return (this.internalGetValue(index) == null);
  }
  
  /**
   * Rowのすべて列がNULLであるか検査します。<br/>
   * 
   * @return True：すべて列がNULL、Talse：NULLではない列を含む
   */
  public boolean isNullAll() {
    for (int i = 0; i < getColumnCount(); i++) {
      if (!isNull(i))
        return false;
    }
    return true;
  }
  
  
  /**
   * valueをColumnに設定されたformatPatternに基づきフォーマットします。<br/>
   * <br/>
   * 
   * 詳細はjava.textパッケージのドキュメントを参照してください。<br/>
   * 
   * Date, Time, Timestampの場合には、formatDate()メソッドを使用してください。<br/>
   * 
   * 数値型のカラムがゼロサプレスする設定になっていた場合、フォーマットを試みて、その結果が"0"だったら ""を返します。<br/>
   * ゼロサプレスしない設定であれば、フォーマットした結果をそのまま返します。<br/>
   * 
   * 
   * @param column 対象のColumnオブジェクト（ここからフォーマットパターンを取得）
   * @param value フォーマットするオブジェクト
   * @return フォーマットした文字列
   * @throws DataException フォーマットに失敗した場合
   */
  String formatObject(Column column, Serializable value) throws DataException {
    if (value == null)
      return Const.NULLSTRING;
    if (value instanceof String)
      return (String) value;

    if (value instanceof java.lang.Number) {
      try {
        // ---- FormatPatternなしのBigDecimalに対応
        String formatPattern = column.getFormatPattern();

        if (StringUtil.isNullString(formatPattern)) {
          String s = (value instanceof BigDecimal) ?
              ((BigDecimal) value).toPlainString() : value.toString();
          // Doubleの場合は、"0.0"となる
          if (column.isZeroSuppress()) {
            if ("0".equals(s) || "0.0".equals(s)) {
              return Const.NULLSTRING;
            } else if (column.getType().equals(Column.Type.BIGDECIMAL)
                && column.getScale() > 1) {
              java.math.BigDecimal bd = java.math.BigDecimal.valueOf(0L, column
                  .getScale());
              if (bd.toString().equals(s)) {
                return Const.NULLSTRING;
              }
            }
          }
          return s;
        } else {
          DecimalFormat decimalFormat = new DecimalFormat(formatPattern);
          String s = decimalFormat.format(value);
          if (column.isZeroSuppress()) {
            // -0.00のようになる可能性がある。
            String s2;
            if (s.length() > 1 && s.charAt(0) == '-') {
              s2 = s.substring(1);
            } else {
              s2 = s;
            }
            if (decimalFormat.format(0L).equals(s2)) {
              return Const.NULLSTRING;
            }
          }
          return s;
        }
      } catch (Exception x) {
        String _caption = column.getCaption();
        throw new InvalidColumnException("FORMAT_ERROR_NUMBER",
            new String[] { _caption, value.toString() }, column);
      }
    } else {
      try {
        return value.toString();
      } catch (Exception ex) {
      }
      return Const.NULLSTRING;
    }
  }
  
  /**
   * dateをColumnに設定されたformatPatternに基づきフォーマットします。<br/><br/>
   * 
   * 詳細はjava.textパッケージのドキュメントを参照してください。<br/>
   * 
   * このメソッドを呼ぶ前にcolumnがnullでないことを確認してから、このメソッドを呼び出してください。<br/>
   * nullの場合には事前に "" を返すように処理を行ってください。<br/>
   * <br/>
   * 
   * <em>※注意：</em>
   * ナノ秒部分はjava.text.SimpleDateFormatが対応していないためフォーマットすることはできません。<br/>
   * 
   * @param column 対象のColumnオブジェクト（ここからフォーマットパターンを取得）
   * @param date フォーマットするDateオブジェクト
   * @return フォーマットした文字列
   * @throws InvalidColumnException フォーマットに失敗した場合
   * @see jp.littlesoft.data.InternalRecordHandler#getValueAsString(int)
   */
  private String formatDate(Column column, java.util.Date date)
      throws InvalidColumnException {
    try {
      String formatPattern = column.getFormatPattern();
      // 本当は dateFormat, simpleDateFormat, wareki それぞれでsynchronizedしたいが、
      SimpleDateFormat simpleDateFormat = new SimpleDateFormat();
      if (DataUtil.XWareki.isWarekiFormatPattern(formatPattern)) {
        return new DataUtil.XWareki().format(date, formatPattern);
      } else {
        if (StringUtil.isNullString(formatPattern)) {
          // フォーマットパターンがnullの場合は、Columnの型に合わせたデフォルトフォーマットパターンを適用する。
          switch (column.getType()) {
          case DATE:
            simpleDateFormat.applyPattern(Column.DEFAULT_FORMAT_PATTERN_DATE);
            break;
          case TIME:
            simpleDateFormat.applyPattern(Column.DEFAULT_FORMAT_PATTERN_TIME);
            break;
          case TIMESTAMP:
            simpleDateFormat
                .applyPattern(Column.DEFAULT_FORMAT_PATTERN_TIMESTAMP);
            break;
          default:
            String _caption = column.getCaption();
            String _value = (date == null) ? "null" : date.toString();
            throw new InvalidColumnException("FORMAT_ERROR_DATE",
                new String[] { _caption, _value }, column);
          }
        } else {
          simpleDateFormat.applyPattern(formatPattern);
        }
        return simpleDateFormat.format(date);
      }

    } catch (InvalidColumnException x) {
      throw x;
    } catch (Exception x) {
      String _caption = column.getCaption();
      String _value = (date == null) ? "null" : date.toString();
      throw new InvalidColumnException("FORMAT_ERROR_DATE", new String[] {
          _caption, _value }, column);
    }
  }
  
  /**
   * 文字列をColumnに設定されたformatPatternに基づきパースします。<br/><br/>
   * 
   * 詳細はjava.textパッケージのドキュメントを参照してください。<br/>
   * 
   * <em>※注意：</em>
   * ColumnがBigDecimal型の場合 double の精度を超える数値は設定できません。<br/>
   * doubleの精度内に丸められて設定されます。<br/>
   * 
   * @param column 対象のColumnオブジェクト（ここからフォーマットパターンを取得）
   * @param value パースする文字列
   * @return パースしたオブジェクト
   * @throws InvalidColumnException パースに失敗した場合
   */
  Serializable parseObject(Column column, String value) throws InvalidColumnException {
    if (value == null || Const.NULLSTRING.equals(value)) {
      if (column.getType()==Type.BOOLEAN) {
        return new java.lang.Boolean(false);
      } else {
        return null;
      }
    }
    if (column.getType().equals(Column.Type.UNKNOWN)) {
      column.setType(Column.Type.STRING);
    }
    String formatPattern = column.getFormatPattern();
    switch (column.getType()) {
    case STRING:
    case CLOB:
    case JSONOBJECT:
      return value;
    case BYTE:
    case SHORT:
    case LONG:
    case INT:
    case FLOAT:
    case DOUBLE:
      value = DataUtil.normalizeNumeric(value);
      DataUtil.checkDecimalLength(column, value);
      if (StringUtil.isNullString(formatPattern)) {
        try {
          switch (column.getType()) {
          case BYTE:
            return Byte.valueOf(value);
          case SHORT:
            return Short.valueOf(value);
          case INT:
            return Integer.valueOf(value);
          case LONG:
            return Long.valueOf(value);
          case FLOAT:
            return Float.valueOf(value);
          case DOUBLE:
            return Double.valueOf(value);
          default:
            throw new Exception(); // ここに来ることはないはず
          }
        } catch (Exception x) {
          String _caption = column.getCaption();
          throw new InvalidColumnException(
              "PARSE_ERROR_INVALID_NUMBERFMT2",
              new String[] { _caption, value }, column);
        }
      } else {
        return parseNumber(column, formatPattern, value);
      }
    case BIGDECIMAL:
      value = DataUtil.normalizeNumeric(value);
      DataUtil.checkDecimalLength(column, value);
      java.math.BigDecimal dec;
      if (StringUtil.isNullString(formatPattern)) {
        try {
          dec = new java.math.BigDecimal(value);
        } catch (Exception x) {
          String _caption = column.getCaption();
          throw new InvalidColumnException("PARSE_ERROR_INVALID_NUMBER",
              new String[] { _caption, value }, column);
        }
      } else {
        Number num = parseNumber(column, formatPattern, value);
        dec = new java.math.BigDecimal(num.toString());
        if (column.getScale() != Const.NOT_ASSIGNED) {
          dec = dec.setScale(column.getScale(),
              java.math.BigDecimal.ROUND_HALF_EVEN);
        }
      }
      return dec;
    case DATE:
    case TIME:
    case TIMESTAMP:
      value = DataUtil.normalizeNumeric(value);
      java.util.Date dt;
      try {
        if (DataUtil.XWareki.isWarekiFormatPattern(formatPattern)) {
          dt = new DataUtil.XWareki().parse(value, formatPattern);
        } else {
          SimpleDateFormat simpleDateFormat = new SimpleDateFormat();
          //入力日付の妥当性まで厳密にチェックする
          simpleDateFormat.setLenient(false);
          if (StringUtil.isNullString(formatPattern)) {
            // フォーマットパターンがnullの場合は、Columnの型に合わせたデフォルトフォーマットパターンを適用する。
            switch (column.getType()) {
            case DATE:
              simpleDateFormat.applyPattern(Column.DEFAULT_FORMAT_PATTERN_DATE);
              break;
            case TIME:
              simpleDateFormat.applyPattern(Column.DEFAULT_FORMAT_PATTERN_TIME);
              break;
            case TIMESTAMP:
              simpleDateFormat
                  .applyPattern(Column.DEFAULT_FORMAT_PATTERN_TIMESTAMP);
              break;
            }
          } else {
            simpleDateFormat.applyPattern(formatPattern);
          }
          
          // SimpleDateFormat#parse で0埋めが行われる前に形式チェック
          String regex = column.getParseRegex();
          if (StringUtil.isNotNullString(regex)) {
            Pattern pattern = Pattern.compile(regex);
            Matcher matcher = pattern.matcher(value);
            if (!matcher.matches()) {
              throw new Exception();
            }
          }
          
          dt = simpleDateFormat.parse(value);
        }

        switch (column.getType()) {
        case DATE:
          return new java.sql.Date(dt.getTime());
        case TIME:
          return new java.sql.Time(dt.getTime());
        case TIMESTAMP:
          return new java.sql.Timestamp(dt.getTime());
        }
      } catch (Exception x) {
        String _caption = column.getCaption();
        throw new InvalidColumnException("PARSE_ERROR_INVALID_DATE",
            new String[] { _caption, value }, column);
      }
      return dt;
    case BOOLEAN:
      return Boolean.valueOf(value);
    case UNKNOWN:
    case BINARY:
    case LONGBINARY:
    case BLOB:
      String _caption = column.getCaption();
      throw new InvalidColumnException("PARSE_ERROR_INVALID_TYPE",
          new String[] { _caption, column.getType().toString(), value },
          column);
    default:
      String _caption2 = column.getCaption();
      String[] errParams = { _caption2, column.getType().toString() };
      throw new InvalidColumnException("INVALID_DATAFIELD_TYPE2",
          errParams, column);
    }
  }
  
  private Number parseNumber(Column df, String formatPattern, String value)
      throws InvalidColumnException {
    DecimalFormat fmt = new DecimalFormat(formatPattern);
    fmt.setParseBigDecimal(true);
    ParsePosition parsePosition = new ParsePosition(0);
    Number answer = fmt.parse(value, parsePosition);
    if (parsePosition.getIndex() < value.length()) {
      String _caption = df.getCaption();
      throw new InvalidColumnException(
          "PARSE_ERROR_INVALID_NUMBERFMT2", new String[] { _caption, value }, df);
    }
    return answer;
  }
  
  /**
   * 行の削除予約の状態を調べます。<br>
   * @deprecated
   * @return True：予約されている False：予約されていない
   */
  public boolean isDeletionDeserved() {
    return deletionReserved;
  }
  
  /**
   * 行の削除予約の状態を調べます。<br>
   * 
   * @return True：予約されている False：予約されていない
   */
  public boolean isDeletionReserved() {
    return deletionReserved;
  }
  /**
   * 行の削除予約を設定します。<br>
   * 
   * @param deletionDeserved True：予約する False：予約しない
   * @deprecated
   */
  public void setDeletionDeserved(boolean deletionDeserved) {
    this.deletionReserved = deletionDeserved;
  }

  /**
   * 行の削除予約を設定します。<br>
   * 
   * @param deletionDeserved True：予約する False：予約しない
   */
  public void setDeletionReserved(boolean deletionReserved) {
    this.deletionReserved = deletionReserved;
  }
  /**
   * この行オブジェクトが格納されているテーブルオブジェクトを取得します。
   * @return AbstractTable
   */
  public Table<E> getTable() {
    return table;
  }
  
  
  protected Column getColumn(String name) throws DataException{
    if (table!=null) {
      return table.getColumn(name);
    }
    return null;
  }
  
  protected Column getColumn(int index) {
    if (table!=null) {
      return table.getColumn(index);
    }
    return null;    
  }
  
  protected int indexOfColumn(String name) throws InvalidColumnException{
    if (table!=null) {
      int result = table.indexOfColumn(name);
      if (result==-1) throw new InvalidColumnException("INVALID_DATAFIELD_NAME", new Serializable[]{name}, null );
      return result;
    }
    return -1;     
  }
  
  protected int getColumnCount()  {
    if (table!=null) {
      return table.getColumnCount();
    }
    return 0;    
  }
  
  /**
   * この行が編集中であるか？<br/>
   * 行の状態がRow.Status.CREATINまたはRow.Status.UPDATINGの時、「編集中」と判断されます。
   * @return この行が編集中であるか？
   */
  public boolean isEditing() {
    return (this.status == Row.Status.CREATING || this.status == Row.Status.UPDATING);
  }
  
  /**
   * 行オブジェクトに設定された全てのデータ値(１行分)の検証を行い、設定内容を確定します。<br/><br/>
   * 
   * 行の状態が編集中（CREATINGまたはUPDATING）の場合、当メソッドの処理対象となります。
   * 左記以外の行に対しては、何の処理も行われません。<br/>
   * また、削除予約された行（DeletionDeserved:true)の行も処理対象外です。<br/><br/>
   * 
   * このメソッドでは、NULL不可のカラムに値が設定されているか？を検証し、
   * 問題が無ければ、行の状態が以下の様に変更されます。<br/>
   * ・CREATINGの場合、CREATEDとなる。<br/>
   * ・UPDATINGの場合、UPDATEDとなる。<br/>
   * 
   * 
   */
  @SuppressWarnings("unchecked")
  public void validate()  throws DataException {
    
    //削除予約された行、編集中（CREATING、UPDATING）でない行は対象外
    if (this.isDeletionReserved() || !this.isEditing()) {
      return;
    }
    
    if (table!=null) {
      try {
        this.table.fireRowValidating(new TableEvent(this.table, this, this.table.indexOfRow(this)));
      } catch (Exception x) {
        throw new DataException(x);
      }
    }
    
    if (isTypedRow()) {
      validate(entity);
    }
    
    String nulls = Const.NULLSTRING;
    for (int i = 0; i < getColumnCount(); i++) {
      // ヌル値チェックを実行
      if (!getColumn(i).isNullable() && isNull(i)) {
        if (!nulls.equals(Const.NULLSTRING)) nulls += ", ";
        nulls += getColumn(i).getCaption();
      }
    }
    if (!nulls.equals(Const.NULLSTRING)){
      throw new InvalidValueException("INVALID_NULL_VALUE",
          new String[]{nulls}, this, null, null);
    }
    
    switch(this.status) {
    case UPDATING:
      //変更された値(this)とロード時の値(originalRow)を比較し、変化がある場合、カラムの更新フラグをセットする
      for (int i = 0; i < getColumnCount(); i++) {
        if (!getColumn(i).isFromDataBase()) continue;
        if ((!this.isNull(i) && originalRow.isNull(i))
            || (this.isNull(i) && !originalRow.isNull(i))
            || (!this.isNull(i) && !originalRow.isNull(i)
              && !this.getValue(i).equals(originalRow.getValue(i)))) {
          getColumn(i).setUpdated(true);
        }
      }
      setStatus(Status.UPDATED);
      break;
    case CREATING:
      setStatus(Status.CREATED);
      break;
    }
  }
  
  private void validate(E entity) throws DataException {
    if (!isTypedRow() || entity != this.entity) {
      throw new DataException("Invalid operation.");
    }
    for (int i=0; i<table.getColumnCount(); i++) {
      Serializable newValue =  internalGetValue(i);
      if (newValue != null && newValue instanceof BigDecimal && table.getColumn(i).getScale()>0) {
        ((BigDecimal)newValue).setScale(table.getColumn(i).getScale(), BigDecimal.ROUND_HALF_UP);
      }
      checkValidation(i, newValue);
    }
    //validate();
  }
  
  /**
   * 行オブジェクトに対する編集（値の設定）を元に戻します。<br/><br/>
   * 
   * [パラメータ:trueの場合]<br/>
   * 行をデータベース等から読み込まれた時点の内容に復元します。<br/>
   * 行が新規作成されたものである場合は、作成された時点の内容に復元します。<br/>
   * 削除された行(DELETED)も元の状態に復元させることが出来ますが、テーブルオブジェクトの可視行に
   * 含めるにはテーブルオブジェクト側での操作が必要です。
   * <br/>
   * 
   * [パラメータ:falseの場合]<br/>
   * 行を前回validateした直後の内容に復元します。<br/>
   * このオプションは、行の状態がRow.Status.UPDATING,CREATING（行の編集中)の場合のみ有効で、
   * UPDATED,CREATED,DELETED(編集完了)の場合は、何の効果もありません。
   * validateした回数が１回以下の場合は、[パラメータ:true]の結果と同じになります。<br/><br/>
   * 
   * いずれも、状態がUNKNOWNの場合には何の効果もありません。
   * 
   * @param initial
   */
  @SuppressWarnings("unchecked")
  public void cancelUpdates(boolean initial) {
    if (initial) {
      if (status == Row.Status.CREATING || status == Row.Status.CREATED) {
        status = Row.Status.CREATED;
        initializeValues();
        setDefault();
      } else if (originalRow!=null) {
        this.status = originalRow.status;
        recover(originalRow);
        this.originalRow = null;
      }
    } else {
      if (this.oldRow!=null) {
        this.status = oldRow.status;
        recover(oldRow);
      } else if (status == Row.Status.CREATING || status == Row.Status.CREATED) {
        status = Row.Status.CREATING;
        initializeValues();
        setDefault();        
      }
    }
    this.oldRow = null;
    if (table!=null) {
      table.fireDataChanged(new TableEvent(table, this, table.indexOfRow(this)), TableEvent.Operation.DONE_CANCEL);
    }
  }
  
  /**
   * このRowオブジェクトまたは含まれるエンティティオブジェクトを編集モードにします.<br/>
   * 編集が終了したら、validate()で内容を検証・確定してください。
   * @param forUpdate
   * @return
   */
  public void startEditing() {
    if (isDeletionReserved()) setDeletionReserved(false);
    if (this.status == Status.LOADED || this.status == Status.UPDATED) {
      setStatus(Status.UPDATING);
    } else if (this.status == Status.CREATED) {
      setStatus(Status.CREATING);
    }
  }
  
  private void recover(Row<E> saveRow) {
    try {
      for (int i=0; i<table.getColumnCount(); i++) {
        internalSetValue(i, saveRow.values[i]);
      }
    } catch(DataException x) {
      throw new RuntimeException(x);
    }
  }

  /**
   * カラムに設定されたデータ値を検証する内部メソッド。<br/>
   * カラム値設定メソッドから必ず呼び出され、行の状態(Row.Status)を編集中(CREATING,UPDATING)に変更します。
   * 行の状態がStatus.UNKNOWNの場合は、何も処理されません。<br/>
   * 
   * @param index 列インデックス
   * @param value java.lang.Object型のデータ値
   * @throws DataException 例外発生時
   */
  private void checkValidation(int index, Serializable value) throws DataException {
    if (this.status==Status.UNKNOWN || this.isDeletionReserved()) return;
    Column col = getColumn(index);
    //カラムレベルの制約検証
    col.getValidator().execute(this, col, value);
    if (!col.isFromDataBase()) return;
    startEditing();
  }
  
  private void setDefault() {
    for (int i=0; i < getColumnCount(); i++) {
      String defaultValue = getColumn(i).getDefaultValue();
      if (StringUtil.isNotNullString(defaultValue)) {
        try {this.setValueAsString(i, defaultValue);} catch (DataException x) {}
      }
    }    
  }
  
  @SuppressWarnings("unchecked")
  private void saveOldRow() {
    oldRow = new Row();
    saveValues(oldRow);
  }
  
  @SuppressWarnings("unchecked")
  private void saveLoadedRow(){
    if (this.originalRow == null) {
      originalRow = new Row();
      saveValues(originalRow);
    }
  }
  
  @SuppressWarnings("unchecked")
  private void saveValues(Row row) {
    row.table = this.table;
    row.typedRow = false;
    row.rowID = this.rowID;
    row.status = this.status;
    row.values = new Serializable[this.table.getColumnCount()];
    for (int i = 0; i<row.values.length; i++) {
      row.values[i]  = internalGetValue(i);
      if (row.values[i] instanceof Cloneable) {
        try {
          Method clone = row.values[i].getClass().getMethod("clone");
          row.values[i] = (Serializable) clone.invoke(row.values[i]);
        } catch (Exception x) {
          throw new RuntimeException(x);
        }
      }
    }
  }

  
}

