/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.data;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.Serializable;
import java.lang.reflect.Field;
import java.util.*;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import jp.littlesoft.data.Column.Type;
import jp.littlesoft.data.event.ChangeEvent;
import jp.littlesoft.data.event.ChangeListener;
import jp.littlesoft.data.event.TableEvent;
import jp.littlesoft.data.event.TableListener;
import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.util.Const;
import jp.littlesoft.util.IndexedMap;

/**
 * 表データクラス。<br/><br/>
 *
 * 表構造のデータを格納する為のオブジェクトです。<br/>
 * 単純な２次元配列と異なる点は、データ項目の型や、その他の属性をColumnクラスによって定義できる点です。<br/>
 * また、データ項目値の設定、取得の他、行の追加、削除、並び替え、フィルタなど数多くのメソッドが用意されています。<br/><br/>
 * 
 * @see Column
 * @see Params
 * @author  LittleSoft Corporation
 * @version 1.3.0
 */
public class Table<E extends Serializable> implements Cloneable, Serializable {
  
  /**
   * SortKeyクラス。<br/>
   * 
   * @author LittleSoft Corporation
   */
  public final static class SortKey implements Serializable {
    private static final long serialVersionUID = 1L;
    String columnName;
    boolean ascending = true;
    /**
     * 引数 columnName で指定された列名で SortKey オブジェクトを生成します。<br/>
     * 
     * @param columnName 列名
     */
    public SortKey(String columnName) {
      this.columnName = columnName;
    }

    public SortKey(String columnName, boolean ascending) {
      this.columnName = columnName;
      this.ascending = ascending;
    }

    public String getColumnName() {
      return columnName;
    }

    public boolean isAscending() {
      return ascending;
    }

  }

  private static final long serialVersionUID = 1L;
  
  private static final transient Logger logger = LogManager.getLogger(Table.class.getName());
  
  /**
   * 文字列のバイト数をチェックする際に使用するエンコードを表す定数です。<br/><br/>
   * 
   * デフォルト: "Windows-31J" <br/>
   */
  private String charsetForCheckingLength = "Windows-31J";
  
  private String name;
  
  /**
   * Column オブジェクトの集合です。<br/>
   */
  protected final IndexedMap<String,Column> columns = new IndexedMap<String,Column>();

  /**
   * 表で可視となる行オブジェクトのIDを管理するVector
   */
  private final ArrayList<Row<E>> visibleRows = new ArrayList<Row<E>>();
  
  /**
   * 全ての行オブジェクトが登録されるMAP
   */
  private final LinkedHashMap<Row.RowID,Row<E>> rows = new LinkedHashMap<Row.RowID,Row<E>>();

  //表に登録されたカウント
  int rowCounter = -1;
  //表で発生するイベントを監視するオブジェクト
  private Vector<ChangeListener> changeListeners;

  private Vector<TableListener> tableListeners;

  /**
   * Table オブジェクトの状態を表します。<br/><br/>
   * 
   * True：アクティブ状態、False：非アクティブ状態<br/>
   */
  protected boolean active;
  
  //行カーソルが有効範囲内にあるか？
  private boolean _inBounds = false;
  //行カーソルの位置
  private int cursorRowIndex = -1;

  /**
   * 表に付属するパラメータを格納する Params オブジェクトです。
   */
  protected Params<E> _params = null;
  /**
   * パラメータとして使用するエンティティオブジェクト.<br/>
   * getParams().getEntity()でアクセスできるオブジェクトと同じものです.<br/>
   * エンティティの型を指定しないテーブルオブジェクトでは、nullです。
   */
  public E params = null;
  
  //表にデータを供給してくれるオブジェクト
  private DataProvider dataProvider = null;
  //表のデータをDBや外部ファイルに書き出してくれるオブジェクト
  private DataResolver dataResolver = null;
  //フィルターを適用する
  private boolean applyFilter = false;

  //列の見出し(Columnクラスのcaptionプロパティ)に自動設定するリソースクラスの配列
  @SuppressWarnings("unchecked")
  private Class[] captionResource = null;

  private SortKey[] sortKeys = null;

  private ArrayList<SortKey> sortKeyHistory = null;

  //ページ取得の際の全件数
  private int allRowCount = -1;
  
  //setMaxRowsで設定した件数よりも実際に発行したSQLで取得できるレコードが存在するか否か
  private boolean hasMoreRow = false;
  
  //データ値を格納するクラスを指定します。Rowオブジェクトで値格納には、このクラスのインスタンスが使われます。
  private Class<E> entityType = null;

  private boolean upperCaseDBColumnName = true;

  /**
   * エンティティのダミーオブジェクト.<br/>
   * 
   * SQLオブジェクトを利用して、SQL文を構築する際に利用します。<br/>
   * [ex.]
   * <pre>
   * employee.select().
   *   where(employee.$$.deptCode).is(new Long(25)).z.
   *   orderBy(employee.$$.name)....</pre>
   * ※エンティティの型を指定しない場合は、nullです。<br>
   */
   public E $$ = null;
   
   /**
    * カーソル行のエンティティオブジェクトへのショートカット.<br/>
    * table.getEntity();<br>
    * で取得できるオブジェクトと同じ内容です。
    */
   public E entity = null;
  
  /**
   * Table オブジェクトを生成します.<br/>
   */
//  public Table() {
//  }
  
  /**
   * Table オブジェクトを生成します.<br/><br/>
   * 
   * 以下のように、エンティティオブジェクトの型を指定してテーブルオブジェクトに作成します.<br/>
   * Table<Employee> employee = new Table<Employee>();<br/><br/>
   * 
   * これによりgetEntity()メソッドで、エンティティオブジェクトへアクセスできるようになります。
   * 
   */
  @SuppressWarnings("unchecked")
  public Table(E... e) {
    //if (e.length==0) return;
    try {
      _params = new Params<E>(e);
      setEntityType((Class<E>) e.getClass().getComponentType());
    } catch (Exception x) {
      throw new RuntimeException(x);
    }
  }
  
  private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
    in.defaultReadObject();
    if (!isTypedRow()) return;
    List<Column> entityCols = new ITableDef.Columns(entityType, this.upperCaseDBColumnName);
    Column col = null;
    try {
      for (Column newCol : entityCols) {
        col = getColumn(newCol.getName());
        col.entityField  = newCol.entityField;
        col.entityGetter = newCol.entityGetter;
        col.entitySetter = newCol.entitySetter;
      }
    } catch (Exception x) { new IOException(x); }
  }
  
  public boolean isTypedRow() {
    return (entityType!=null && 
           !entityType.isAnnotation() &&
           !entityType.isArray() &&
           !entityType.isPrimitive() &&
           !entityType.equals(Object.class) &&
           !entityType.equals(Serializable.class) &&
           !entityType.equals(Row.class));
  }
  


  /**
   * Table オブジェクトをアクティブ化します。<br/>
   * 
   * @throws DataException 例外発生時
   */
  public void activate() throws DataException {
    setActive(true);
  }

  /**
   * Table オブジェクトへの変更を監視する ChangeListener オブジェクトを登録します。<br/>
   * 
   * @param l ChangeListener オブジェクト
   */
  @SuppressWarnings("unchecked")
  public synchronized void addChangeListener(ChangeListener l) {
    Vector<ChangeListener> v = 
      changeListeners == null ? 
        new Vector<ChangeListener>(2) : (Vector<ChangeListener>) changeListeners.clone();
    if (!v.contains(l)) {
      v.addElement(l);
      changeListeners = v;
    }
  }  
  
  /**
   * カラムの追加（作成したオブジェクトを登録）します。<br/>
   *
   * @param  column  Column オブジェクト
   * @throws  DataException  例外発生時
   */
  public synchronized void addColumn(Column column) throws DataException {
    checkDeactive();
    columns.put(column.getName(), column);
  };
  
  /**
   * カラムの追加（名前で作成）します。<br/>
   * 
   *
   * @param   fieldName  DataField名
   * @return  Column オブジェクト
   * @throws  DataException  例外発生時
   */
  public synchronized Column addColumn(String fieldName) throws DataException {
    Column col = new Column(fieldName);
    addColumn(col);
    return col;    
  }

  /**
   * カラムの追加（名前とタイプを指定して作成）します。<br/>
   *
   * @param   fieldName DataField名
   * @param   type タイプ
   * @return  Column オブジェクト
   * @throws  DataException  例外発生時
   */
  public synchronized Column addColumn(String fieldName, Type type) throws DataException {
    Column col = new Column(fieldName, type);
    addColumn(col);
    return col;
  }

  /**
   * カラムの追加（名前配列で作成）します。<br/>
   * 
   * @param names カラム名の配列
   * @throws DataException  例外発生時
   */
  public synchronized void addColumns(String[] names) throws DataException {
    for (int i=0; i<names.length; i++) {
      addColumn(names[i]);
    }
  }

  /**
   * TableListenerの登録します。<br/>
   *
   * @param  l  TableListenerインターフェースを実装したオブジェクト
   */
  @SuppressWarnings("unchecked")
  public synchronized void addTableListener(TableListener l) {
    Vector<TableListener> v = tableListeners == null ? 
        new Vector<TableListener>(2) : (Vector<TableListener>) tableListeners.clone();
    if (!v.contains(l)) {
      v.addElement(l);
      tableListeners = v;
    }
  }

  /**
   * 行を追加します。<br/><br/>
   *
   * 新規の行を追加します。<br/>
   * 追加された行は最終行となります。
   * また、追加処理の完了後、行カーソルは追加された行となります。<br/><br/>
   * 
   * 非アクティブの状態でこのメソッドが呼ばれた場合は自動的にアクティブにします。<br/><br/>
   * 
   * 追加した行をデータベースに反映する前には、Row.validate() メソッドで変更を確定させる必要があります。<br/>
   *
   * @throws  DataException  例外発生時
   */
  public synchronized Row<E> appendRow() throws DataException {
    return insertRow(visibleRows.size());
  }

  /**
   * 行を追加します。<br/><br/>
   * 
   * 追加したエンティティをデータベースに反映する場合は、Row.validate(entity) メソッドで変更を確定させる必要があります。<br/><br/>
   * 
   * @param entity エンティティオブジェクト
   * @throws DataException
   */
  public synchronized Row<E> appendRow(E entity) throws DataException {
    return insertRow(visibleRows.size(), entity);
  }

  /**
   * Tableにフィルターを適用する指示を行います。<br/><br/>
   * 
   * 行データに対するフィルターを適用します。このメソッドを呼び出すと、TableListener objectに対しrowFilteringが通知されます。<br/>
   * rowFilteringはTableが所有する全ての行データに対して呼ばれます。<br/>
   * rowFilteringの戻り値としてtrueが返されるとその行データは表示対象となります。<br/><br/>
   * 
   * TableListener側では以下の様なコード実装する必要があります。
   * <pre>
   *   public boolean rowFiltering(TableEvent e,  Row row) {
   *     try {
   *       if (e.getSource().equals(rsList)) {
   *         //部門コードが"1000"だったら表示対象とします。
   *         return "1000".equals(row.getValueAsString("BUMON_CODE"));
   *       } else {
   *         return true;
   *       }
   *     } catch (Exception x) {
   *       x.printStackTrace();
   *     }
   *     return true;
   *   }
   * </pre>
   * 
   * <em>※注意：</em>
   * applyFilterを使用しフィルタリングされた行の並び順は、行が登録された順番になります。<br/>
   * 並び順を指定して行をフィルタリングしたい場合は、applySortFilterを使用してください。</p>
   *
   * @throws DataException フィルター適用時に発生した例外
   */
  public synchronized void applyFilter() throws DataException {
    applyFilter = true;
    internalApplyFilter(true, true);
  }
  
  /**
   * Tableにフィルターを適用する指示を行うと同時にその結果をソートします。<br/><br/>
   * 
   * フィルター結果を一定の順序で並べ替える際は当メソッドを使用してください。<br/>
   *
   * @param  sortKeys  並び替えを行うソートキー配列
   * @throws DataException フィルター適用時に発生した例外
   */
  public synchronized void applySortFilter(SortKey[] sortKeys) throws DataException {
    applyFilter = true;
    this.sortRows(sortKeys);
  }

  /**
   * 行に対する変更、行追加をキャンセルします。<br/><br/>
   * 
   * [initialパラメータ:trueの場合]<br/>
   * 全ての行をデータベース等から読み込まれた時点の内容に復元します。<br/>
   * 行が新規作成されたものである場合は、作成された時点の内容に復元します。<br/>
   * 削除された行(DELETED)も元の状態に復元させることが出来ます。
   * <br/>
   * 
   * [initialパラメータ:falseの場合]<br/>
   * 全ての行を前回validateした直後の内容に復元します。<br/>
   * このオプションは、行の状態がRow.Status.UPDATING,CREATING（行の編集中)の場合のみ有効で、
   * UPDATED,CREATED,DELETED(編集完了)の場合は、何の効果もありません。
   * validateした回数が１回以下の場合は、[パラメータ:true]の結果と同じになります。<br/><br/>
   * 
   * @param initial true:
   * @throws  DataException  例外発生時
   */
  public synchronized void cancelUpdates(boolean initial) throws DataException {
    for (Row<E> row : rows.values()) {
      row.cancelUpdates(initial);
    }    
    Row<E> curRow = null;
    if (visibleRows.size() != 0) curRow = getRow();

    if (initial) {
      // 可視行を全て削除
      visibleRows.clear();
      cursorRowIndex = -1;
      doInternalApplyFilter();
    }

    if (visibleRows.size() > 0) {
      // 先頭行へもどす。
      if (curRow != null && (visibleRows.indexOf(curRow) > (visibleRows.size() - 1)) && !(visibleRows.indexOf(curRow) != -1))
        internalMoveTo(visibleRows.indexOf(curRow));
      else
        internalMoveTo(0);
    }
  }
  

  /**
   * Table オブジェクトの状態チェックし、アクティブでなかったらアクティブ化します。<br/>
   *
   * @throws DataException 例外発生時
   */
  protected void checkActive() throws DataException {
    if (!isActive())
      setActive(true);
  }
  
  /**
   * アクティブ時にしてはいけない事をした場合の例外処理を実装します。<br/><br/>
   *
   * アクティブではないことを前提にした処理を始める前にこのメソッドが呼ばれます。 <br/>
   * このメソッドではアクティブかどうかを調べ、アクティブなら例外が発生します。 <br/>
   *
   * @throws DataException アクティブ時にしてはいけない事をした場合
   */
  protected void checkDeactive() throws DataException {
    if (isActive())
      throw new DataException(null, null, "INVALID_OPER_DATA_IS_ACTIVE",
          new String[]{this.getName()});
  }

  /**
   * 指定したColumnの重複値チェックを行います。<br/><br/>
   *
   * Column名配列で指定された行の値をチェックし、重複する値をもつ行のインデックスを返します。<br/>
   * 重複が検出された時点でチェックを中止するので、全てを一度にチェックする事はできません。<br/>
   *
   * @param   columnNameArray  チェックするColumn名配列
   * @return  重複する値をもつ行のインデックス配列（重複が検出されない場合nullが返される）<br/>
   * [0],[1]に重複があった行インデックスが格納されます。
   * @throws  DataException  例外発生時
   */
  public int[] checkDuplicate(String[] columnNameArray) throws DataException {
  	  return checkDuplicate(columnNameArray, 0);
  }
  
  /**
   * 指定したColumnの重複値チェックを行います。<br/><br/>
   *
   * Column名配列で指定された行の値をチェックし、重複する値をもつ行のインデックスを返します。<br/>
   * 重複が検出された時点でチェックを中止するので、全てを一度にチェックする事はできません。<br/><br/>
   *
   * 全ての重複行を取得するには、rowIndexを変えながら、繰り返し呼び出してください。
   * <pre>
   *   String[] columnNames = new String[]{"COL1"};
   *   int rowIndex = 0;
   *   int[] results;
   *   while (true) {
   *     results = dataTable.checkDuplicate(columnNames, rowIndex);
   *     if (results == null) break;
   *     System.out.println(results[0] + " & " + results[1] + " rows are duplicated.");
   *     rowIndex = results[0] + 1; 
   *   };
   * </pre>
   * 
   * @param columnNameArray  チェックするColumn名配列
   * @param rowIndex 重複値チェックを開始する行インデックス
   * @return  重複する値をもつ行のインデックス配列（重複が検出されない場合nullが返される）<br/>
   * [0],[1]に重複があった行インデックスが格納されます。
   * @throws  DataException  例外発生時
   */
  public int[] checkDuplicate(String[] columnNameArray, int rowIndex) throws DataException {
    if (columnNameArray == null || columnNameArray.length == 0)
      return null;
    validateAllRows();
    int[] colpos = new int[columnNameArray.length];
    for (int i = 0; i < columnNameArray.length; i++) {
      colpos[i] = indexOfColumn(columnNameArray[i]);
    }

    Map<String, Integer> values = new HashMap<String, Integer>();
    for (int row = rowIndex; row < getRowCount(); row++) {
      StringBuffer checkKey = new StringBuffer();
      for (int i = 0; i < columnNameArray.length; i++) {
        checkKey.append(getRow(row).getValueAsString(colpos[i])).append('\t');
      }
      String checkKeyStr = checkKey.toString();
      Integer dupRow = (values.get(checkKeyStr));
      if (dupRow == null)
        values.put(checkKeyStr, new Integer(row));
      else {
        int[] dupRows = new int[2];
        dupRows[0] = dupRow.intValue();
        dupRows[1] = row;
        return dupRows;
      }
    }
    return null;
  }

  /**
   * 追加行が存在するかどうかを調べます。<br/>
   *
   * @param   toLast  True：最終行まで調べる、False：最終行まで調べない
   * @return  True：追加行が存在する、False：存在しない
   * @throws  DataException  例外発生時
   */
  public boolean checkMoreRow(boolean toLast) throws DataException{
    if (dataProvider!=null) {
      return dataProvider.checkMoreRow(this, toLast);
    } else {
      return false;
    }
  }


  /**
   * すべての行を削除します。<br/><br/>
   * 
   * 当オブジェクトが所有する全ての行データを削除します。<br/>
   * 
   * @throws  DataException  例外発生時
   */
  public synchronized void deleteAllRows() throws DataException {
    if (!moveTo(0))
      return;
    while (inBounds()) {
      deleteRow();
    }
  }

  
  /**
   * 現在行を削除します。<br/><br/>
   * 
   * 現在行カーソルが位置している行を削除します。<br/>
   * 同時に行カーソルは次の行へ送られます。最終行が削除された場合、行カーソルは新たな最終行にとどまります。<br/>
   *
   * @throws  DataException  例外発生時
   */
  public synchronized void deleteRow() throws DataException {
    Row<E> cursorRow = getRow();
    fireRowDeleting(new TableEvent<E>(this, cursorRow, cursorRowIndex));
    Row.RowID rowid = this.getRowID(cursorRowIndex);
    Row<E> delRow = internalGet(rowid);
    switch(delRow.getStatus()) {
    case UNKNOWN:
    case CREATING:
    case CREATED:
      internalRemove(rowid);
      break;
    default:
      delRow.setStatus(Row.Status.DELETED);
    }
    visibleRows.remove(cursorRowIndex);
    fireDataChanged(new TableEvent<E>(this, cursorRow, cursorRowIndex), TableEvent.Operation.ROW_DELETING);
    cursorRowIndex--;
    //if (visibleRows.size() == 0)
    //  cursorRow.clear();
    if (visibleRows.size() == (cursorRowIndex + 1))
      cursorRowIndex--;
    next();
    // 削除処理が完了したので、DONE_DELETEイベントを発生させる。
    fireDataChanged(new TableEvent<E>(this, cursorRow, cursorRowIndex), TableEvent.Operation.DONE_DELETE);
  }





  /**
   * 行データを指定された順序で並び替えます。<br/><br/>
   * 
   * Column名配列の先頭から第１キーとなります。<br/>
   * 降順、昇順を指定するboolean配列はColumn名配列と長さが一致してなければなりません。<br/><br/>
   *
   * <em>※注意：</em>
   * Unicode のコード順に並び替えが行われます。<br/><br/>
   * 
   * isApplyFilter() メソッドが true の場合は、フィルターも適用されます。<br/>
   *
   * @param  sortKeyColNameArray  並び替えを行うColumn名配列
   * @param  ascendingArray       Column名配列に対応した降順、昇順を指定します。（True：昇順、False：降順）
   * @throws  DataException  例外発生時
   */
  private void doInternalApplyFilter() throws DataException {
    Row<E> compRow = null;
    boolean _sort = (sortKeys == null || sortKeys.length == 0) ? false : true;
    int[] colIndexes;
    Column.Type[] colTypes;
    if (_sort) {
      colIndexes = new int[sortKeys.length];
      colTypes = new Column.Type[sortKeys.length];
      for (int i = 0; i < sortKeys.length; i++) {
        colIndexes[i] = this.columns.indexOfKey(sortKeys[i].columnName);
        if (colIndexes[i] == -1)
          throw new DataException(null, null, "INVALID_SORT_COLUMNNAME",
            new String[]{this.getName(), sortKeys[i].columnName});
        colTypes[i] = this.columns.getValue(colIndexes[i]).getType();
      }
    } else {
      colIndexes = null;
      colTypes = null;
    }

    for (Row<E> newRow : this.allRows()) {
      if (newRow.getStatus() == Row.Status.DELETED)
        continue;
      if (this.isApplyFilter() && !fireRowFiltering(new TableEvent<E>(this, newRow, -1)))
        continue;
      if (_sort) {
        doSort(0, (visibleRows.size() - 1), newRow, compRow, colIndexes, colTypes, sortKeys);
      } else {
        this.visibleRows.add(newRow);        
      }
    }
  }

  /**
   * データのロードを行います。<br/><br/>
   *
   * このメソッドの処理は空です。<br/>
   *
   * @throws DataException  例外発生時
   */
  protected void doLoadData() throws DataException {}

  /**
   * DataFieldのセットアップを行います。<br/><br/>
   *
   * このメソッドの処理は空です。<br/>
   *
   * @param  fields  ColumnMap
   * @throws DataException  例外発生時
   */
  protected void doSetupColumn(IndexedMap<String,Column> fields) throws DataException {}

  @SuppressWarnings("unchecked")
  private void doSort (
    int start,
    int end,
    Row newRow,
    Row compRow,
    int[] colIndexes,
    Column.Type[] colTypes,
    SortKey[] ascendings)
    throws DataException {
    // handleRowsインデックスリスト並び替えながら新しい行データのインデックスをhandleRowsへ登録していきます。
    int result = 0;
    int target = 0;
    if (this.visibleRows.size() == 0) //最初の１件目
      visibleRows.add(newRow);
    else {
      target = start + ((end - start + 1) / 2); //handleRowsの中間位置にいる行データと新しい行データを比較します。
      compRow = doSortSetCompRows(target);
      result = doSortCompAll(newRow, compRow, colIndexes, colTypes, ascendings);

      if (result == 0) {
        visibleRows.add(target, newRow); //等しい場合はtarget位置に挿入。
      } else if (result < 0) { //新しいデータを現在データ前方に挿入するケース。
        if (target == 0) {
          visibleRows.add(target, newRow);
        } else {
          compRow = doSortSetCompRows(target - 1); //target位置とその手前にあるデータの間に挿入可能か？
          result = doSortCompAll(newRow, compRow, colIndexes, colTypes, ascendings);
          if (result == 0 || result > 0) //挿入可能！
            visibleRows.add(target, newRow);
          else //挿入不可、可能な場所を再び探しに行く
            doSort(start, (target - 1), newRow, compRow, colIndexes, colTypes, ascendings);
        }
      } else { //新しいデータを現在のデータの後ろに挿入するケース。
        if (target == (visibleRows.size() - 1))
          visibleRows.add(newRow); //探している場所が一番最後なので、そのまま追加
        else
          doSort((target + 1), end, newRow, compRow, colIndexes, colTypes, ascendings); //現在データの後方の挿入場所を再び探しに行く。
      }
    }
  }

  private int doSortComp(Row<E> newRow, Row<E> compRow, int colIndex, Column.Type type, SortKey asc)
    throws DataException {
      // 昇順／降順の指定にいる値の大小結果を逆転させます。
    if (asc.ascending)
      return doSortInternalComp(newRow, compRow, colIndex, type);
    else
      return doSortInternalComp(newRow, compRow, colIndex, type) * -1;
  }

  private int doSortCompAll(Row<E> newRow, Row<E> compRow, int[] colIndexes, Column.Type[] colTypes, SortKey[] ascendings)
    throws DataException {
    int result = 0;
    for (int i = 0; i < colIndexes.length; i++) {
      result = doSortComp(newRow, compRow, colIndexes[i], colTypes[i], ascendings[i]);
      if (result != 0)
        return result;
    }
    return result;
  }

  /**
   * 2つの列値を比較します。
   * 
   * @param newRow 新しい行オブジェクト
   * @param compRow 比較対象行オブジェクト
   * @param colIndex 列インデックス
   * @param type Column オブジェクトの型
   * @return -1：新しい行が小さい、0：等しい、1：新しい行が大きい
   * @throws DataException 例外発生時
   */
  private int doSortInternalComp(Row<E> newRow, Row<E> compRow, int colIndex, Column.Type type)
    throws DataException {

    if (newRow.isNull(colIndex) && compRow.isNull(colIndex))
      return 0; //等しい
    else if (newRow.isNull(colIndex))
      return -1; //新しい行が小さい
    else if (compRow.isNull(colIndex))
      return 1; //新しい行が大きい

    switch (type) {
      case STRING :
        // ここにはnullは来ないはず
        return newRow.getString(colIndex).compareTo(compRow.getString(colIndex));
      case BIGDECIMAL :
        //---- BigDecimal対応 (nullは事前に考慮済みなので、ここでは考慮しなくてよい)
        return newRow.getBigDecimal(colIndex).compareTo(compRow.getBigDecimal(colIndex));
      case BYTE :
      case SHORT :
      case LONG :
      case INT :
      case FLOAT :
      case DOUBLE :
        {
          if (newRow.getDouble(colIndex) == compRow.getDouble(colIndex))
            return 0;
          else if (newRow.getDouble(colIndex) < compRow.getDouble(colIndex))
            return -1;
          else
            return 1;
        }
      case BOOLEAN :
        {
          if (newRow.getBoolean(colIndex) == compRow.getBoolean(colIndex))
            return 0;
          else if (!newRow.getBoolean(colIndex))
            return -1;
          else
            return 1;
        }
      case DATE :
        {
          if (newRow.getDate(colIndex).getTime() == compRow.getDate(colIndex).getTime())
            return 0;
          else if (newRow.getDate(colIndex).getTime() < compRow.getDate(colIndex).getTime())
            return -1;
          else
            return 1;
        }
      case TIME :
        {
          if (newRow.getTime(colIndex).getTime() == compRow.getTime(colIndex).getTime())
            return 0;
          else if (newRow.getTime(colIndex).getTime() < compRow.getTime(colIndex).getTime())
            return -1;
          else
            return 1;
        }
      case TIMESTAMP :
        {
          //----  JDK1.2/1.3では秒までしか取れないが、後でナノ秒を取得して比較しているのでこれでよい。
          java.sql.Timestamp newTimestamp = newRow.getTimestamp(colIndex);
          java.sql.Timestamp compTimestamp = compRow.getTimestamp(colIndex);
          if (newTimestamp.getTime() < compTimestamp.getTime())
            return -1;
          else if (newTimestamp.getTime() > compTimestamp.getTime())
            return 1;
          else {
            // 秒までは同じ
            if (newTimestamp.getNanos() == compTimestamp.getNanos())
              return 0;
            else if (newTimestamp.getNanos() < compTimestamp.getNanos())
              return -1;
            else
              return 1;
          }
        }
      case UNKNOWN :
      case BINARY :
      case LONGBINARY :
      case BLOB :
      case JSONOBJECT :
      default :
        throw new DataException(null, null, "INVALID_DATAFIELD_TYPE",
            new String[]{type.toString()});
    }
  }

  private Row<E> doSortSetCompRows(int index) throws DataException {
    return this.internalGet(this.getRowID(index));
  }

  /**
   * 項目値更新直後イベントを送出します。<br/>
   * 
   * @param e ChangeEvent オブジェクト
   */
  protected void fireAfterSetValue(ChangeEvent e) {

    if (changeListeners != null) {
      Vector<ChangeListener> listeners = changeListeners;
      int count = listeners.size();
      for (int i = 0; i < count; i++) {
        ((ChangeListener) listeners.elementAt(i)).afterSetValue(e);
      }
    }
  }

  /**
   * 項目値更新直前イベントを送出します。<br/>
   * 
   * @param e ChangeEvent オブジェクト
   * @throws java.lang.Exception 例外発生時
   */
  protected void fireBeforeSetValue(ChangeEvent e) throws java.lang.Exception {
    if (changeListeners != null) {
      Vector<ChangeListener> listeners = changeListeners;
      int count = listeners.size();
      for (int i = 0; i < count; i++) {
        ((ChangeListener) listeners.elementAt(i)).beforeSetValue(e);
      }
    }
  }

  /**
   * DataChangeイベントの送出をします。<br/>
   *
   * @param  e          TableEvent
   * @param  operation  オペレーション（TableEventの定数）
   */
  public void fireDataChanged(TableEvent<E> e, TableEvent.Operation operation) {
    try {
      setEntityShortcut(getRow());
    } catch (Exception x) {
      setEntityShortcut(null);
    }
    if (tableListeners != null) {
      Vector<TableListener> listeners = tableListeners;
      int count = listeners.size();
      for (int i = 0; i < count; i++) {
        ((TableListener) listeners.elementAt(i)).dataChanged(e, operation);
      }
    }
  }

  /**
   * アクティブ化イベントを送出します。<br/>
   * 
   * @param e ChangeEvent オブジェクト
   * @throws java.lang.Exception 例外発生時
   */
  protected void fireOnActivate(ChangeEvent e) throws java.lang.Exception {
    if (changeListeners != null) {
      Vector<ChangeListener> listeners = changeListeners;
      int count = listeners.size();
      for (int i = 0; i < count; i++) {
        ((ChangeListener) listeners.elementAt(i)).onActivate(e);
      }
    }
    if (visibleRows.size() > 0) {
      moveTo(0);
    }
  }

  /**
   * フィールド構築イベントを送出します。<br/>
   * 
   * @param e ChangeEvent オブジェクト
   */
  protected void fireOnColumnSetup(ChangeEvent e) {
    if (changeListeners != null) {
      Vector<ChangeListener> listeners = changeListeners;
      int count = listeners.size();
      for (int i = 0; i < count; i++) {
        ((ChangeListener) listeners.elementAt(i)).onColumnSetup(e);
      }
    }
  }

  /**
   * 非アクティブ化イベントを送出します。<br/>
   * 
   * @param e ChangeEvent オブジェクト
   */
  protected void fireOnPassivate(ChangeEvent e) {
    if (changeListeners != null) {
      Vector<ChangeListener> listeners = changeListeners;
      int count = listeners.size();
      for (int i = 0; i < count; i++) {
        ((ChangeListener) listeners.elementAt(i)).onPassivate(e);
      }
    }
  }

  /**
   * rowDeletingイベントを送出します。<br/>
   *
   * @param  e  TableEvent
   * @throws  java.lang.Exception  例外発生時
   */
  public void fireRowDeleting(TableEvent<E> e) throws DataException {
    try {
      if (tableListeners != null) {
        Vector<TableListener> listeners = tableListeners;
        int count = listeners.size();
        for (int i = 0; i < count; i++) {
          ((TableListener) listeners.elementAt(i)).rowDeleting(e);
        }
      }
    } catch (Exception ex) {
      throw new DataException(ex);
    }
  }

  /**
   * rowFilteringイベントを送出します。<br/>
   *
   * @param  e  イベントオブジェクト
   * @return boolean True：表示行に含める、False：表示行に含めない。
   */
  public boolean fireRowFiltering(TableEvent<E> e) {
    if (tableListeners != null) {
      Vector<TableListener> listeners = tableListeners;
      int count = listeners.size();
      for (int i = 0; i < count; i++) {
        if (!((TableListener) listeners.elementAt(i)).rowFiltering(e))
          return false;
      }
    }
    return true;
  }
  
  /**
   * rowValidatingイベントを送出します。<br/>
   *
   * @param  e  TableEvent
   * @throws  java.lang.Exception  例外発生時
   */
  public void fireRowValidating(TableEvent<E> e) throws java.lang.Exception {
    if (tableListeners != null) {
      Vector<TableListener> listeners = tableListeners;
      int count = listeners.size();
      for (int i = 0; i < count; i++) {
        ((TableListener) listeners.elementAt(i)).rowValidating(e);
      }
    }
  }

  /**
   * 行カーソルを先頭に移動します。<br/>
   *
   * @return  True：行の移動に成功した、False：失敗した
   * @throws  DataException  例外発生時
   */
  public synchronized boolean first() throws DataException {
    return moveTo(0);
  }

  /**
   * 内部にある全ての行オブジェクトのコレクションを取得します。<br/>
   * 削除された行、フィルターで除外された行など、全ての行オブジェクトが含まれます。
   * @return  java.util.Collection<Row>
   */
  public final java.util.Collection<Row<E>> allRows(){
    return rows.values();
  }
  
  /**
   * 内部にある全てのエンティティオブジェクトのコレクションを取得します。<br/>
   * 削除されたもの行、フィルターで除外されたものなど、全てのエンティティオブジェクトが含まれます。
   * @return  java.util.Collection<Row>
   */
  public final java.util.Collection<E> allEntities(){
    ArrayList<E> result = new ArrayList<E>();
    for (Row<E> row : rows.values()) {
      result.add(row.getEntity());
    }
    return result;
  }
  
  /**
   * 可視行オブジェクトのコレクションを取得します。<br/>
   * 
   * @return java.util.Collection<Row>
   */
  public final java.util.Collection<Row<E>> visibleRows() {
    return this.visibleRows;
  }
  
  /**
   * 可視行のエンティティオブジェクトのコレクションを取得します。<br/>
   * 
   * @return java.util.Collection<Row>
   */
  public final EntityList<E> visibleEntities() {
    return new EntityList<E>(this);
  }
  
  
  
  /**
   * 列の見出し(Columnクラスのcaptionプロパティ)に設定するリソースクラスの配列を取得します。 <br/>
   * 
   * @return リソースクラスの配列
   */
  @SuppressWarnings("unchecked")
  public Class[] getCaptionResource() {
    return captionResource;
  }

  public Column getColumn(Object $dummyFiled) {
    try {
      for (Column c : columns.values()) {
        if (c.getEntityValue($$)==$dummyFiled) {
          return c;
        }
      }
    } catch (Exception x) {
      throw new RuntimeException(x);
    }
    return null;
  }
  /**
   * 引数 index で指定された列インデックスから Column オブジェクトを取得します。<br/>
   *
   * @param index 列インデックス
   * @return Column オブジェクト
   **/
  public Column getColumn(int index) {
    return columns.getValue(index);
  }

  /**
   * 引数 columnName で指定されたカラム名から Column オブジェクトを取得します。<br/>
   *
   * @param columnName カラム名
   * @return Column オブジェクト
   * @throws DataException 指定されたcolumnNameが存在しない場合
   */
  public Column getColumn(String columnName) throws DataException {
    if (!columns.containsKey(columnName)) {
      throw new DataException(null, null, "INVALID_DATAFIELD_NAME", new String[]{columnName});      
    }
    return columns.get(columnName);
  }

  /**
   * 引数 columnName で指定されたカラム名から Column オブジェクトを取得します。<br/>
   * 
   * カラムが存在しない場合、NULLを返します。
   * 
   * @param columnName カラム名
   * @param returnNull Dummy
   * @return Column オブジェクト
   */
  public Column getColumn(String columnName, boolean returnNull){
    if (columns.containsKey(columnName)) {
      return columns.get(columnName);
    } else {
      return null;
    }
  }
  
  /**
   * Column オブジェクトを配列で取得します。<br/>
   *  
   * @return Column オブジェクトの配列
   */
  public synchronized Column[] getColumnArray() {
    Column[] cols = new Column[columns.size()];
    return columns.toArray(cols);
  }
  

  /**
   * フィールド数を取得します。<br/>
   *
   * @return フィールド数
   */
  public int getColumnCount() {
    return columns.size();
  }

  /**
   * Row オブジェクト構造を定義するカラム集合を取得します。<br/>
   *
   * @return カラム集合
   */
  final public IndexedMap<String, Column> getColumns() {
    return columns;
  }

  /**
   * 表にデータを供給する DataProvider オブジェクトを取得します。<br/>
   * 
   * @return DataProvider オブジェクト
   */
  public DataProvider getDataProvider() {
    return dataProvider;
  }

  /**
   * 表のデータをDBや外部ファイルに書き出す DataResolver オブジェクトを取得します。<br/>
   * 
   * @return DataResolver オブジェクト
   */
  public DataResolver getDataResolver() {
    return dataResolver;
  }

  
  /**
   * Table オブジェクトの名称を取得します。<br/>
   * 
   * @return Table オブジェクトの名称
   */
  public String getName() {
    return name;
  }
  
  /**
   * Nullable属性の配列を取得します。<br/>
   * 
   * @return 各フィールドのNullable属性を格納した配列
   * @throws DataException 例外発生時
   */
  public boolean[] getNullables() throws DataException {
    boolean[] nullables = new boolean[this.getColumnCount()];
    for (int i=0; i<nullables.length; i++) {
      nullables[i] = this.getColumn(i).isNullable();
    }
    return nullables;
  }
  /**
   * パラメータセットを取得します。<br/><br/>
   *
   * @return  パラメータセット
   */
  public Params<E> getParams() {
    return _params;
  }

  /**
   * カーソル行の Row オブジェクトを取得します。<br/>
   * 
   * @return Row オブジェクト
   */
  public Row<E> getRow() throws DataException {
    return getRow(cursorRowIndex);
  }

  /**
   * 引数 rowIndex で指定された Row オブジェクトを取得します。<br/>
   * 
   * @param rowIndex 行インデックス
   * @return Row オブジェクト
   * @throws  DataException  例外発生時
   */
  public Row<E> getRow(int rowIndex) throws DataException {
    return this.visibleRows.get(rowIndex);
  }

  /**
   * カーソル行のエンティティオブジェクトを取得します。<br/>
   * ※このメソッドでエンティティオブジェクトを変更しても、その内容を永続化する事はできません。<br/>
   * 永続化する場合は、getEntityForEdit()を使用してエンティティオブジェクトを取得してください。
   * @return エンティティオブジェクト
   * @throws DataException 
   */
  public E getEntity() throws DataException {
    return (E) getRow().getEntity();
  }

  /**
   * 引数 rowIndex にカーソルを移動し、エンティティオブジェクトを取得します。<br/>
   * ※このメソッドでエンティティオブジェクトを変更しても、その内容を永続化する事はできません。<br/>
   * 永続化する場合は、getEntityForEdit(rowIndex)を使用してエンティティオブジェクトを取得してください。
   * @param rowIndex
   * @return エンティティオブジェクト
   * @throws DataException
   */
  public E getEntity(int rowIndex) throws DataException {
    return (E) getRow(rowIndex).getEntity();
  }
  
  /**
   * カーソル行のエンティティオブジェクトを編集・更新可能な状態にします。<br/>
   * 
   * @param entity
   * @throws DataException
   */
  public E startEditing() throws DataException {
    getRow().startEditing();
    return getRow().entity;
  }
  
  /**
   * 指定されたエンティティオブジェクトを編集・更新可能な状態にします。<br/>
   * 
   * @param entity
   * @throws DataException
   */
  public void startEditing(E entity) throws DataException {
    int index = indexOf(entity);
    if (index==-1) {
      throw new DataException("This table does not contain the specified entity.");
    }
    getRow(index).startEditing();
  }
  
  /**
   * カーソルのエンティティオブジェクトをチェックし、問題がなければ値を受け入れます.
   */
  public void validate() throws DataException {
    if (!getRow().status.equals(Row.Status.CREATING) &&
        !getRow().status.equals(Row.Status.UPDATING)) {
      throw new DataException("Specified entity in the state is not updatable.");
    }
    validate(getRow().getEntity());
  }
  
  /**
   * 内容が編集・更新されたエンティティオブジェクトをチェックし、問題がなければ値を受け入れます.
   * @param entity
   * @throws DataException
   */
  public void validate(E entity) throws DataException {
    int index = indexOf(entity);
    if (index==-1) {
      appendRow(entity).validate();
    } else {
      Row<E> row = getRow(index);
      if (!row.status.equals(Row.Status.CREATING) &&
          !row.status.equals(Row.Status.UPDATING)) {
        throw new DataException("Specified entity in the state is not updatable.");
      }
      row.validate();
    }
  }

  
  /**
   * エンティティオブジェクトの行インデックスを取得します.<br>
   * 存在しない場合は-1を返します。
   * @param entity
   * @return
   */
  public int indexOf(E entity) {
    try {
      if (getRowCount()<1) return -1;
      if (getRow().getEntity()==entity) return getRowIndex();
      int i = 0;
      for (Row<E> row : visibleRows) {
        if (row.getEntity()==entity) {
          return i;
        }
        i++;
      }
      return -1;
    } catch (DataException x) {
      return -1;
    }
  }
  /**
   * Rowオブジェクトの行インデックスを取得します.<br>
   * 存在しない場合は-1を返します。
   * @param entity
   * @return
   */
  public int indexOf(Row<E> row) {
    try {
      if (getRowCount()<1) return -1;
      if (getRow()==row) return getRowIndex();
      int i = 0;
      for (Row<E> _row : visibleRows) {
        if (_row==row) {
          return i;
        }
        i++;
      }
      return -1;
    } catch (DataException x) {
      return -1;
    }
  }

  /**
   * 行数を取得します。<br/><br/>
   *
   * 行カーソルの位置づけ対象となるデータ行の数です。不可視の行、削除された行は含まれません。<br/>
   *
   * @return  行数
   */
  public int getRowCount() {
    return visibleRows.size();
  }
  
  /**
   * 全体の行数を取得します。<br/><br/>
   * 
   * テーブルに分割格納されている行数ではありません。<br/>
   * 
   * @return 全体の行数
   */
  public int getAllRowCount() {
    return this.allRowCount;
  }

  /**
   * 全体の行数を設定します。<br/>
   * 
   * @param count  全体の行数
   */
  public void setAllRowCount(int count) {
    this.allRowCount = count;
  }

  /**
   * 行インデックスを指定して、RowIDを取得します。<br/><br/>
   *
   * 各行には行インデックスとは異なる絶対的なRowIDをもちます。この値をキーに内部テーブルに登録されています。<br/>
   * このメソッドは上記のRowIDを取得します。<br/>
   *
   * @param   rowIndex  行インデックス
   * @return  内部RecordId
   * @throws  DataException 指定された行が存在しない
   */
  public Row.RowID getRowID(int rowIndex) throws DataException {
    try {
      return visibleRows.get(rowIndex).getRowID();      
    } catch (Throwable t) {
      throw new DataException(null, null, "RECORD_NOT_FOUND",
          new String[]{this.getName(), "index:" + rowIndex});
    }
  }
  
  /**
   * 現在の行カーソル位置を取得します。<br/>
   *
   * @return  現在行インデックス
   */
  public int getRowIndex() {
    return cursorRowIndex;
  }

  /**
   * ソート用のフィールド名配列を取得します。<br/>
   *
   * @return ソート用のフィールド名配列
   */
  public SortKey[] getSortKeys() {
    return this.sortKeys;
  }
  

  /**
   * 行カーソルが有効範囲内にあるかどうか調べます。<br/>
   *
   * @return  True：有効範囲内、False：有効範囲外
   */
  public boolean inBounds() {
    return this._inBounds;
  }
  /**
   * 指定した列オブジェクトのインデックスを取得します。<br/>
   * @param column
   * @return 列インデックス
   */
  public int indexOfColumn(Column column) {
    return columns.indexOfValue(column);
  }
  
  /**
   * 引数 columnName で指定されたフィールド名から列インデックスを取得します。<br/>
   *
   * @param columnName  フィールド名
   * @return 列インデックス（存在しない場合は -1 ）
   */
  public int indexOfColumn(String columnName) {
    return columns.indexOfKey(columnName);
  }
  
  /**
   * 指定した列のデータベース項目名(DBColumnName)でインデックスを取得します。<br/>
   * データベース項目名(DBColumnName)が指定されていない場合、通常の列名(Name)を検索します。
   * @param name
   * @return 列インデックス
   */
  public int indexOfDBColumnName(String name) {
    int i = 0;
    for (Column col : columns.values()) {
      if (col.getDBColumnName().equals(name) || col.getName().equals(name)) {
        return i;
      }
      i++;
    }
    return -1;
  }
  
  /**
   * 指定した行オブジェクトのインデックスを取得します。<br/>
   *
   * @param  row  対象の行
   * @return  現在行インデックス
   */
  public int indexOfRow(Row<E> row) {
    return this.visibleRows.indexOf(row);
  }
  
  /**
   * 行を挿入します。<br/><br/>
   *
   * 新規の行を指定された場所に挿入します。<br/>
   * 指定されたインデックスが負値の場合、先頭に挿入されます。
   * また、現在保持されている行数以上のインデックスが指定された場合、追加扱いになります。<br/>
   * 挿入処理の完了後、行カーソルは挿入した行になります。<br/><br/>
   * 
   * ColumnにDefaultValueが設定されていた場合、新規行にDefaultValueが設定されます。<br/><br/>
   * 
   * 非アクティブの状態でこのメソッドが呼ばれた場合は自動的にアクティブにします。<br/><br/>
   * 
   * 追加した行をデータベースに反映する前には、Row.validate() メソッドで変更を確定させる必要があります。<br/><br/>
   *
   * @param index 挿入場所を示すインデックス（先頭行は0）
   * @throws  DataException  例外発生時
   */
  public synchronized Row<E> insertRow(int index) throws DataException {
    return insertRow(index, null);
  }
  
  /**
   * エンティティオブジェクトを行を挿入します。<br/><br/>
   * 
   * 追加したエンティティをデータベースに反映する場合は、Row.validate(entity) メソッドで変更を確定させる必要があります。<br/><br/>
   * 
   * @param index  挿入場所を示すインデックス（先頭行は0）
   * @param entity エンティティオブジェクト
   * @throws DataException
   */
  @SuppressWarnings("unchecked")
  public synchronized Row<E> insertRow(int index, E entity) throws DataException {
    checkActive();  // activeでなかったらactiveにします。
    checkEntity(entity);
    Row cursorRow = null;
    if (isTypedRow()) {
      if (entity!=null) {
        cursorRow = new Row<E>(this, Row.Status.CREATING, entity);
      } else {
        try {
          cursorRow = new Row<E>(this, Row.Status.CREATING, entityType.newInstance());
        } catch (Exception e) {
          throw new DataException(e);
        }
      }
    } else {
      cursorRow = new Row(this, Row.Status.CREATING);
    }
    internalPut(cursorRow);
    if (index < 0)
      index = 0;
    if (index >= visibleRows.size()) {
      visibleRows.add(cursorRow);
      cursorRowIndex = visibleRows.size() - 1;
    } else {
      visibleRows.add(index, cursorRow);
      cursorRowIndex = index;
    }
    fireDataChanged(new TableEvent(this, cursorRow, cursorRowIndex), TableEvent.Operation.START_APPEND);
    return cursorRow;
  }

  private void checkEntity(Object entity) throws DataException {
    if (entity!=null && !entity.getClass().equals(entityType)) {
      throw new DataException("invalid entity type");
    }
  }
  
  
  /**
   * 
   * @param  doneFilter  
   * @param  visible     
   * @throws DataException 例外発生時
   */
  private void internalApplyFilter(boolean doneFilter, boolean visible) throws DataException {

    this.checkActive();

    if (allRows().size() == 0)
      return;

    Row<E> curRow = null;
    if (visibleRows.size() != 0) curRow = getRow();

    // 可視行を全て削除
    visibleRows.clear();
    //cursorRow = new Row(this);
    cursorRowIndex = -1;

    if (visible) {
      doInternalApplyFilter();
    }

    if (visibleRows.size() > 0) {
      // 先頭行へもどす。
      if (curRow != null && (visibleRows.indexOf(curRow) > (visibleRows.size() - 1)) && !(visibleRows.indexOf(curRow) != -1))
        internalMoveTo(visibleRows.indexOf(curRow));
      else
        internalMoveTo(0);
    }
    fireDataChanged(new TableEvent<E>(this, getRow(), getRowIndex()), TableEvent.Operation.DONE_MOVE);

    if (doneFilter)
      fireDataChanged(new TableEvent<E>(this, getRow(), getRowIndex()), TableEvent.Operation.DONE_FILTER);

  }
  
  
  /**
   * deActiveの時の後始末を行います。<br/>
   *
   * @throws  DataException  例外発生時
   */
  protected void internalFree() throws DataException {
    rows.clear();
    visibleRows.clear();
    cursorRowIndex = -1;
    setEntityShortcut(null);
    if (dataProvider!=null) {
      dataProvider.free();
    }
  }
  
  /**
   * RowIDを指定して行オブジェクトを取得します。<br/><br/>
   *
   * @param   rowID  行ID
   * @return  Row　オブジェクト
   * @throws  DataException  例外発生時
   */
  private Row<E> internalGet(Row.RowID rowID) throws DataException{
    final Row<E> row = rows.get(rowID);        
    if (row==null) {
      throw new DataException(null, null, "INTERNAL_RECORD_NOT_FOUND",
          new Object[] {"", rowID});
    }
    return row;
  }
  
  
  
  /**
   * アクティブ時の内部設定を実装します。<br/><br/>
   *
   * このメソッドの処理は空です。<br/>
   */
  protected void internalOnActivate() throws DataException {}
  
  /**
   * 行オブジェクトを内部テーブルに登録します。<br/>
   *
   * @param   row  Row オブジェクト
   * @return  行ID
   */
  public Row.RowID internalPut(Row<E> row) {
    rows.put(row.getRowID(), row);
    return row.getRowID();
  }
  
  /**
   * 自分が使う行テーブルの登録します。<br/><br/>
   * 
   * 継承先doLoadDataにてこれを呼び出してください。<br/>
   * 
   * @param row 行オブジェクト
   * @param visible 登録と同時に可視とするか？
   * @param insert  先頭に挿入するか？
   * @return 取り扱い行へ追加できたかどうか
   * @throws DataException 例外発生時
   */
  public boolean internalPut(Row<E> row, boolean visible, boolean insert) throws DataException {
    internalPut(row);
    if (visible) {
      if (insert)
        visibleRows.add(0, row);
      else 
        visibleRows.add(row);
    }
    return true;
  }
  
  /**
   * RowIDを指定して内部テーブルから行オブジェクトを削除します。<br/>
   *
   * @param  rowID  行ID
   * @throws  DataException  例外発生時
   */
  public void internalRemove(Row.RowID rowID) throws DataException {
    Row<E> record = rows.get(rowID);
    if (record==null)
      throw new DataException(null, null, "INVALID_OPER_NODATA", new String[]{""});
    rows.remove(rowID);
  }
  
  
  
  /**
   * 引数 index で指定されたインデックスへ行カーソルを移動します。<br/><br/>
   *
   * 行移動イベント送出を伴わない為、TableListenerオブジェクトに通知する事なく、
   * 行カーソルを移動させる事が可能です。<br/><br/>
   * 
   * 当メソッドを使用する前にインデックスを取得し、使用後そのインデックスにもどすこと処理が必要です。<br/>
   * これを行わないと、TableListenerオブジェクトがとらえている現在行と不整合が発生し、
   * 予期せぬ結果となります。<br/><br/>
   *
   * <em>※注意：</em>
   * 現在行と同じ行インデックスが指定された場合は、行移動は行わずに true を返します。<br/>
   * 
   * @param  index  行インデックス
   * @return  行の移動に成功したかどうか
   * @throws  DataException  例外発生時
   */
  public synchronized boolean internalMoveTo(int index) throws DataException {
    // 現在行と同じなので、何もしない。
    if (this.getRowIndex() == index)
      return true;
    
    // 指定インデックスが取扱行ベクタの上回るときは、追加行のチェック。
    if (visibleRows.size() < (index + 1)) {
      // 追加行がないので、あきらめてください！
      if (!checkMoreRow(false))
        return false;

      // 指定インデックスが最終行のときも心配なので、追加行のチェック。
    } else if (visibleRows.size() == (index + 1)) {
      checkMoreRow(false);
    }
    cursorRowIndex = index;
    return true;
  }
  
  
  /**
   * 内部データのセットアップします。<br/><br/>
   *
   * DataField設定を確定し、データロードの実行を行います。<br/>
   *
   * @throws  DataException  例外発生時
   */
  protected void internalSetup() throws DataException {
    if (dataProvider!=null) {
      dataProvider.setupColumn(this);
    } else {
      doSetupColumn(columns);
    }
    if (captionResource != null) {
      Field caption; Column col;
      for (int i = 0; i < getColumnCount(); i++) {
        col = this.getColumn(i);
        if (!col.getCaption().equals(col.getName())) continue;
        for (int x = 0; x < captionResource.length; x++) {
          try {
            caption = captionResource[x].getField(col.getName());
            col.setCaption(caption.get(captionResource).toString());
          } catch (Exception e) {
            try {
              caption = captionResource[x].getField(col.getDBColumnName());
              col.setCaption(caption.get(captionResource).toString());
            } catch (Exception e1) {
              logger.debug("キャプションが定義されていません。: [ Columns ] "+ col.getName());
            }
          }
        }
      }
      Params<?> p = this.getParams();
      for (int i = 0; i < p.getColumnCount(); i++) {
        col = p.getColumn(i);
        if (!col.getCaption().equals(col.getName())) continue;
        for (int x = 0; x < captionResource.length; x++) {
          try {
            caption = captionResource[x].getField(col.getName());
            col.setCaption(caption.get(captionResource).toString());
          } catch (Exception e) {
            try {
              caption = captionResource[x].getField(col.getDBColumnName());
              col.setCaption(caption.get(captionResource).toString());
            } catch (Exception e1) {
              logger.debug("キャプションが定義されていません。: [ Params ] "+ col.getName());
            }
          }
        }
      }
    }
    rows.clear();
    for (int i = 0; i < getColumnCount(); i++) {
      // カラム毎の更新フラグをクリア
      getColumn(i).setUpdated(false);
    }
    if (dataProvider!=null) {
      dataProvider.loadData(this);
    } else {
      doLoadData();
    }
  }
  
  /**
   * Table オブジェクトがアクティブ化された状態かどうか調べます。<br/>
   * 
   * @return True：アクティブ状態、False：非アクティブ状態
   */
  public boolean isActive() {
    return active;
  }
  
  /**
   * フィルターが適用中であるかどうかを取得します。<br/>
   * 
   * @return True：適用中、False：未適用
   */
  public boolean isApplyFilter() {
    return applyFilter;
  }

  
  /**
   * 行カーソルを最終行へ移動します。<br/>
   *
   * @return  True：行の移動に成功した、False：失敗した
   * @throws  DataException  例外発生時
   */
  public synchronized boolean last() throws DataException {
    checkMoreRow(true);
    return moveTo(visibleRows.size() - 1);
  }
  
  /**
   * 行カーソルを1行次へ移動します。<br/>
   *
   * @return  True：行の移動に成功した、False：失敗した
   * @throws  DataException  例外発生時
   */
  public synchronized boolean next() throws DataException {
    return moveTo(cursorRowIndex + 1);
  }
  
  /**
   * Table オブジェクトを非アクティブ化します。<br/>
   * 
   * @throws DataException 例外発生時
   */
  public void passivate() throws DataException {
    setActive(false);
  }
  
  /**
   * 行カーソルを1行前へ移動します。<br/>
   *
   * @return  True：行の移動に成功した、False：失敗した
   * @throws  DataException  例外発生時
   */
  public synchronized boolean previous() throws DataException {
    if (cursorRowIndex <= 0)
      return false;
    return moveTo(cursorRowIndex - 1);
  }
  
  /**
   * Tableにフィルターを解除する指示を行います。<br/><br/>
   * 
   * フィルターを解除した後に、TableEvent.DONE_FILTERイベントを発生させます。<br/>
   * 
   * @throws DataException フィルター解除時に発生した例外
   */
  public synchronized void releaseFilter() throws DataException {
    releaseFilter(true);
  }
  
  /**
   * Tableにフィルターを解除する指示を行います。<br/>
   * 
   * @param doneFilter True：フィルターを解除した後にTableEvent.DONE_FILTERイベントを発生させる、False：させない
   * @throws DataException フィルター解除時に発生した例外
   */
  public synchronized void releaseFilter(boolean doneFilter) throws DataException {
    applyFilter = false;
    internalApplyFilter(doneFilter, true);
  }
  
  /**
   * 全カラムを削除します。<br/>
   *
   * @throws  DataException  例外発生時
   */
  public synchronized void clearColumns() throws DataException {
    checkDeactive();
    columns.clear();
  }
  
  /**
   * Table オブジェクトへの変更を監視する ChangeListener オブジェクトを削除します。<br/>
   * 
   * @param l ChangeListener オブジェクト
   */
  @SuppressWarnings("unchecked")
  public synchronized void removeChangeListener(ChangeListener l) {
    if (changeListeners != null && changeListeners.contains(l)) {
      Vector<ChangeListener> v = (Vector<ChangeListener>) changeListeners.clone();
      v.removeElement(l);
      changeListeners = v;
    }
  }
  

  
  /**
   * 引数 column で指定されたフィールドを削除します。<br/><br/>
   * 
   * 指定した Column オブジェクトが存在しなかった場合は何もしません。<br/>
   * 
   * @param column Column オブジェクト
   */
  public void removeColumn(Column column) {
    int i;
    if ((i=columns.indexOfValue(column)) >= 0) {
      columns.removeAt(i);
    }
  }
  
  /**
   * TableListenerの削除します。<br/>
   *
   * @param  l  TableListenerインターフェースを実装したオブジェクト
   */
  @SuppressWarnings("unchecked")
  public synchronized void removeTableListener(TableListener l) {
    if (tableListeners != null && tableListeners.contains(l)) {
      Vector<TableListener> v = (Vector<TableListener>) tableListeners.clone();
      v.removeElement(l);
      tableListeners = v;
    }
  }
  
  /**
   * Table オブジェクトの変更内容をDBや外部ファイルに書き出します。
   * 
   * @throws DataException 例外発生時
   */
  public synchronized void resolveChanges() throws DataException {
    validateAllRows();
    if (dataResolver!=null) {
      dataResolver.execute(this);
    }
    fireDataChanged(new TableEvent<E>(this, null, -1), TableEvent.Operation.DONE_RESOLVE);
  }
  
  /**
   * Table オブジェクトの状態を設定します。<br/>
   * 
   * @param activation True：アクティブ状態、False：非アクティブ状態
   * @throws DataException 例外発生時
   */
  protected synchronized void setActive(boolean activation) throws DataException {
    setEntityShortcut(null);
    if (isActive()) {
      internalFree();
      fireOnPassivate(new ChangeEvent(this));
    }
    active = false;
    if (!activation)
      return;
    internalSetup();
    //cursorRow = new Row(this);
    active = true;
    if (getRowCount()>0) {
      setEntityShortcut(null);
    }
    fireOnColumnSetup(new ChangeEvent(this));
    internalOnActivate();
    try {
      fireOnActivate(new ChangeEvent(this));
    } catch (DataException e) {
      throw e;
    } catch (Exception e) {
      throw new DataException(e);
    }
  }
  
  /**
   * 列の見出し(Columnクラスのcaptionプロパティ)に設定するリソースクラスを設定します。<br/><br/>
   * 
   * リソースクラスには列名と完全に同名で、文字列フィールド(public static)を定義し、見出しにする値を設定してください。<br/>
   * 
   * @param captionResource リソースクラス
   */
  @SuppressWarnings("unchecked")
  public void setCaptionResource(Class captionResource) {
    this.captionResource = new Class[]{captionResource};
  }
  
  /**
   * 列の見出し(Columnクラスのcaptionプロパティ)に設定するリソースクラスの配列を設定します。 <br/><br/>
   * 
   * リソースクラスには列名と完全に同名で、文字列フィールド(public static)を定義し、見出しにする値を設定してください。<br/>
   * 
   * @param captionResource リソースクラスの配列
   */
  @SuppressWarnings("unchecked")
  public void setCaptionResource(Class[] captionResource) {
    this.captionResource = captionResource;
  }
  
  
  /**
   * 引数 index で指定されたインデックスへ行カーソルを移動します。<br/>
   *
   * @param  index  行インデックス
   * @return  True：行の移動に成功した、False：失敗した
   * @throws  DataException  例外発生時
   */
  public synchronized boolean moveTo(int index) throws DataException {
    if (visibleRows.size() == 0 || index < 0) {
      _inBounds = false;
      setEntityShortcut(null);
      return _inBounds;
    }
    _inBounds = true;
    //インデックスが変らない時は移動しない。
    if (cursorRowIndex == index) {
      setEntityShortcut(getRow());
      return _inBounds;
    }

    // 現在行が追加または編集中の場合、確定！
    //if (this.cursorRow.isEditing()) validateRows();
    
    if (_inBounds = internalMoveTo(index)) {
      fireDataChanged(new TableEvent<E>(this, getRow(), getRowIndex()), TableEvent.Operation.DONE_MOVE);
    }
    return _inBounds;
  }
  
  private void setEntityShortcut(Row<E> row) {
    if (row!=null && isTypedRow()) {
      entity = row.getEntity();
    } else {
      entity = null;
      return;
    }
  }
  /**
   * 表にデータを供給する DataProvider オブジェクトを設定します。<br/>
   * 
   * @param dataProvider DataProvider オブジェクト
   */
  public void setDataProvider(DataProvider dataProvider) {
    this.dataProvider = dataProvider;
    this.dataProvider.setParams(this._params);
  }
  
  /**
   * 表のデータをDBや外部ファイルに書き出す DataResolver オブジェクトを設定します。<br/>
   * 
   * @param dataResolver DataResolver オブジェクト
   */
  public void setDataResolver(DataResolver dataResolver) {
    this.dataResolver = dataResolver;
  }
  
  /**
   * Table オブジェクトの名前を設定します。<br/>
   * 
   * @param name Table オブジェクトの名前
   * @throws DataException 例外発生時
   */
  public void setName(String name) {
    this.name = name;
  }
  
  /**
   * 引数 nullable で指定されたNullable属性を全てのフィールドに設定します。<br/>
   * 
   * @param nullable Nullable属性
   * @throws DataException 例外発生時
   */
  public void setNullableAll(boolean nullable) throws DataException {
    for (int i=0; i<this.getColumnCount(); i++) {
      this.getColumn(i).setNullable(nullable);
    }
  }
  
  /**
   * 引数 nullables で指定されたNullable属性の配列を設定します。<br/>
   * 
   * @param nullables 各フィールドのNullable属性を格納した配列
   * @throws DataException 例外発生時
   */
  public void setNullables(boolean[] nullables) throws DataException {
    if (nullables == null || nullables.length != this.getColumnCount())
      throw new IllegalArgumentException("length of nullables parameter does not match to count of ColumnMap.");

    for (int i=0; i<nullables.length; i++) {
      this.getColumn(i).setNullable(nullables[i]);
    }
  }

  /**
   * ソートキー配列を設定します。<br/><br/>
   * @param sortKeys ソート用のフィールド名の配列
   */
  public void setSortKeys(SortKey... sortKeys) {
    this.sortKeys = sortKeys;
  }
  
  
  /**
   * 前回と同じソート順序で、ソートをやり直します。<br/>
   * 
   * @throws DataException フィルター解除時に発生した例外
   */
  public synchronized void sortRows() throws DataException {
    internalApplyFilter(true, true);    
  }
  
  /**
   * 引数 columnName で指定された列名の値により、行の並び替えを行います。<br/><br/>
   * 
   * ある列を指定すると最初は昇順で並び替えされ、再度同じ列を指定すると降順に並び替えられます。
   * 
   * @param columnName Column名
   */
  public void sortRows(String columnName) {
    if ( !this.isActive() || (this.getColumns().indexOfKey(columnName)==-1) || (columnName==null))  return;
    SortKey sortKey;
    if (sortKeyHistory==null) {
      sortKeyHistory = new ArrayList<SortKey>();
    }
    
    if (sortKeyHistory.size()==0) {
      sortKeyHistory.add(0, new SortKey(columnName));
    } else {
      sortKey = sortKeyHistory.get(0);
      if (sortKey.columnName.equals(columnName)) {
        sortKey.ascending = !sortKey.ascending;
      } else {
        sortKeyHistory.add(0, new SortKey(columnName));
        for (int i = 1; i<sortKeyHistory.size();) {
          sortKey = sortKeyHistory.get(i);
          if (sortKey.columnName.equals(columnName)) {
            sortKeyHistory.remove(i);
          } else {
            i++;
          }
        }
      }
    }
    
    while (sortKeyHistory.size()>this.getColumnCount()) {
      sortKeyHistory.remove(this.getColumnCount()-1);
    }
    SortKey[] sk = new SortKey[sortKeyHistory.size()];
    
    for (int i = 0; i<sortKeyHistory.size(); i++) {
      sk[i] = sortKeyHistory.get(i);
    }
    
    try { this.sortRows(sk); } catch (DataException e) { e.printStackTrace(); }
  }
  
  /**
   * 引数 sortKeyで指定された順序で行データを並び替えます。<br/><br/>
   * @param sortKey
   */
  public synchronized void sortRows(SortKey...sortKey) throws DataException{
    this.sortKeys = sortKey;
    internalApplyFilter(true, true);
  }
  
  /**
   * テーブルオブジェクトに含まれる全ての行を検証し、編集内容を確定します。<br/><br/>
   *
   * @throws  DataException  例外発生時
   */
  public synchronized void validateAllRows() throws DataException {
    String err = Const.NULLSTRING;
    for (Row<E> row : rows.values()) {
      try {
        row.validate();
      } catch (DataException e) {
        if (!err.equals(Const.NULLSTRING)) err += "\n";
        err += e.getLocalizedMessage();
      }
    }
    if (!Const.NULLSTRING.equals(err)) {
      throw new DataException(err);
    }
  }

  /**
   * 文字列項目の長さチェックを行う際の文字コードセット名を取得します。<br>
   * @return 文字列項目の長さチェックを行う際の文字コードセット名
   */
  public String getCharsetForCheckingLength() {
    return charsetForCheckingLength;
  }

  /**
   * 文字列項目の長さチェックを行う際の文字コードセット名を設定します。<br>
   * @param charsetForCheckingLength 文字列項目の長さチェックを行う際の文字コードセット名
   */
  public void setCharsetForCheckingLength(String charsetForCheckingLength) {
    this.charsetForCheckingLength = charsetForCheckingLength;
  }
  
  /*
  public Object checkValidation(Column column, Object value) throws DataException {
    try {
      ChangeEvent fieldHandleEvent = new ChangeEvent(this, column, value, cursorRow.getValueAsObject(columns.indexOfValue(column)));
      fireBeforeSetValue(fieldHandleEvent);
  
      if (fieldHandleEvent.getNewValue() == null) {
        if (this.isCheckNullable() && !column.isNullable()) {
          InvalidColumnException ex = new InvalidColumnException(null, null, "INVALID_NULL_VALUE",
              new String[]{column.getCaption(true)}, column, column.getCaption(true));
          throw ex;
        }
      } else {
        if (column.getType().equals(Column.Type.STRING) && column.isFillChar()) {
          String temp = (String) (fieldHandleEvent.getNewValue());
          temp = temp.trim();
          final int tempLength = temp.getBytes(charsetForCheckingLength).length;
          if (tempLength != column.getLength()) {
            String _caption = column.getCaption(true);
            throw new InvalidColumnException(null, null, "FILL_CHAR",
                new String[]{_caption, Integer.toString(column.getLength())},
                column, _caption);
          }
  
          if (tempLength == 0) {
            if (this.isCheckNullable() && !column.isNullable()) {
              throw new InvalidColumnException(null, null, "INVALID_NULL_VALUE",
                  new String[]{column.getCaption(true)},
                  column, column.getCaption(true));
            } else {
              return null;
            }
          }
          return temp;
        }
      }
      return fieldHandleEvent.getNewValue();
    } catch (DataException ex) {
      throw ex;
    } catch (Exception ex) {
      throw new DataException(ex);
    }
  }
  */
  
  /**
   * setMaxRowsで設定した件数よりも実際に発行したSQLで取得できるレコードが存在するか否かを取得します。<br/>
   * 
   * @return レコードが存在するか否か
   */
  public boolean isHasMoreRow() {
    return hasMoreRow;
  }

  /**
   * setMaxRowsで設定した件数よりも実際に発行したSQLで取得できるレコードが存在するか否かを設定します。<br/>
   * 
   * @param hasMoreRow レコードが存在するか否か
   */
  public void setHasMoreRow(boolean hasMoreRow) {
    this.hasMoreRow = hasMoreRow;
  }
  
  /**
   * データ値を格納するクラスを指定します。Rowオブジェクトで値格納には、このクラスのインスタンスが使われます。
   * @return データ値を格納するクラス
   */
  public Class<E> getEntityType() {
    return entityType;
  }

  public void setEntityType(Class<E> entityType) throws InstantiationException, IllegalAccessException, DataException {
    this.entityType = entityType;
    if (!isTypedRow()) return;
    List<Column> entityCols = new ITableDef.Columns(entityType, this.upperCaseDBColumnName);
    params = _params.entity;
    $$ = this.entityType.newInstance();
    Serializable dummy = null;
    for (Column col : entityCols) {
      addColumn(col);
      dummy = Column.createDefaultValue(col, 0);
      try {
        col.setEntityValue($$, dummy);
      } catch (Column.NoAccessibleObjectException x) {
        //ReadOnly項目
      }
    }

  }
}  
  

