/** Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.data.text;

import org.apache.log4j.*;
import java.io.*;
import java.util.*;

import jp.littlesoft.data.Column;
import jp.littlesoft.data.Table;
import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.util.Const;
import jp.littlesoft.util.StringUtil;
/**
 * テキストデータ(CSV、固定長)を扱うためのテーブルクラス。<br/><br/>
 *
 * テキストデータに対して、以下の操作を行うことができます。<br/><br/>
 *
 * ・テキストファイルを読み込み、Tableにデータを設定する。<br/>
 * ・テキストデータ(String)を読み込み、Tableにデータを設定する。<br/>
 * ・Tableのデータをテキストファイルとして保存する。<br/>
 * ・Tableのデータをテキストデータ(String)として取得する。<br/><br/>
 *
 * ■■■ CSVデータの入出力について ■■■<br/><br/>
 *
 * CSVデータを扱う場合、Schemaプロパティに以下の入出力形式を指定することができます。<br/>
 * 指定しない場合は、全てデータとして入出力を行います。<br/><br/>
 *
 * 1. jp.littlesoft.data.TextTable.Schema.SPECIFIED_NAME_AND_TYPE<br/>
 * ・１行目[カラム名] + ２行目[カラムタイプ] + ３行目以降[データ]<br/>
 * 2. jp.littlesoft.data.TextTable.Schema.SPECIFIED_NAME<br/>
 * ・１行目[カラム名] + ２行目以降[データ]<br/>
 * 3. jp.littlesoft.data.TextTable.Schema.ONLY_DATA<br/>
 * ・１行目以降[データ]（デフォルト）<br/><br/>
 *
 * なお、上記形式2、3を利用する場合は、以下の例2、例3のように、CSVデータの入出力前にカラムの追加処理を行う必要があります。<br/><br/>
 *
 * 例1: １行目がカラム名、２行目がカラムタイプ、３行目以降がデータのCSVファイルを読み込む場合。<br/>
 * TextTable text = new TextTable();<br/>
 * // テキストスキーマの設定<br/>
 * text.setSchema(TextTable.Schema.SPECIFIED_NAME_AND_TYPE);<br/>
 * // CSVファイルのロード<br/>
 * text.importTextFile("file:///c:/input_data.csv");<br/>
 * // CSVファイルのセーブ<br/>
 * text.exportTextFile("c:/output_data.csv");<br/><br/>
 *
 * 例2: １行目がカラム名、２行目以降がデータのCSVファイルを読み込む場合。<br/>
 * TextTable text = new TextTable();<br/>
 * // カラムの追加<br/>
 * text.addColumn("Column1", Column.Type.STRING);<br/>
 * text.addColumn("Column2", Column.Type.INT);<br/>
 * text.addColumn("Column3", Column.Type.DATE);<br/>
 * // テキストスキーマの設定<br/>
 * text.setSchema(TextTable.Schema.SPECIFIED_NAME);<br/>
 * // CSVファイルのロード<br/>
 * text.importTextFile("file:///c:/input_data.csv");<br/>
 * // CSVファイルのセーブ<br/>
 * text.exportTextFile("c:/output_data.csv");<br/><br/>
 *
 * 例3: 1行目以降がデータのCSVファイルを読み込む場合。<br/>
 * TextTable text = new TextTable();<br/>
 * // カラムの追加<br/>
 * text.addColumn("Column1", Column.Type.STRING);<br/>
 * text.addColumn("Column2", Column.Type.INT);<br/>
 * text.addColumn("Column3", Column.Type.DATE);<br/>
 * // CSVファイルのロード<br/>
 * text.importTextFile("file:///c:/input_data.csv");<br/>
 * // CSVファイルのセーブ<br/>
 * text.exportTextFile("c:/output_data.csv");<br/><br/>
 *
 * なお、データ読み込みの際、CSVデータの項目数 ＞ カラム数 である場合、<br/>
 * データの先頭からカラム数分のデータを読み込みます。（残りのデータは無視されます。）<br/>
 * また、読み込み時にCSVデータのある項目を読み飛ばすことはできません。<br/><br/>
 *
 * ■■■ 固定長データの入出力について ■■■<br/><br/>
 *
 * レコード長が固定のデータを読み込むことができます。<br/>
 * 各項目の長さは、Tableに追加したカラムに対して設定します。<br/>
 * データ読み込みの際、レコードの長さ ＞ カラムの長さの合計 である場合、<br/>
 * データの先頭からカラムの長さの合計分のデータを読み込みます。（残りのデータは無視されます。）<br/>
 * また、読み込み時にレコードのある項目を読み飛ばすことはできません。<br/>
 * データ読み込みの際、カラムのデータ型が以下のものに関しては、<br/>
 * 項目の前後半角スペースをトリミングしてTableにデータを設定します。<br/><br/>
 *
 * データ型: BYTE、SHORT、LONG、INT、FLOAT、DOUBLE、BOOLEAN、BIGDECIMAL、DATE、TIME、TIMESTAMP<br/><br/>
 *
 * 日付型のデータを読み込む際は、適切なフォーマットパターンを設定してください。<br/>
 * 数値型の場合でデータ出力時に左ゼロ詰めを行いたい場合は、フォーマットパターンを指定してください。<br/><br/>
 *
 * STRING型はカラムのTrimTypeプロパティの設定によりトリミングについての振る舞いが変化します。<br/>
 * ディフォルトではスペースのトリミングは行われません。<br/><br/>
 *
 * 以下の形式を設定することで、データ出力時の項目に対するスペースの補完処理のタイプを指定することができます。<br/><br/>
 *
 * 1.jp.littlesoft.ds.data.TextTable.PadType.LEFT<br/>
 * ・スペースを左から詰める<br/>
 * 2.jp.littlesoft.ds.data.TextTable.PadType.RIGHT<br/>
 * ・スペースを右から詰める（デフォルト）<br/>
 * 3.jp.littlesoft.ds.data.TextTable.PadType.NOT_ASSIGNED;<br/>
 * ・なにもしない<br/><br/>
 *
 * 例1:固定長データを読み込み、以下の設定で出力を行う場合。<br/>
 * ・Column1のデータの右側をトリミングする。<br/>
 * ・Column2に数値フォーマットパターン"0000000000"を設定し、左ゼロ詰めを行う<br/>
 * ・Column3に日時フォーマットパターン"yyyy/MM"を設定する。<br/>
 * ・補完処理（左詰）を設定する。<br/><br/>
 *
 * 【入力データ】<br/>
 * <table border=1 cellspacing=0 cellpadding=2>
 * <tr><td>Column1<td>Column2<td>Column3
 * <tr><td>あいうえお&nbsp;&nbsp;<td>1234567&nbsp;&nbsp;&nbsp;<td>2002/09/17
 * <tr><td>あいうえお&nbsp;&nbsp;<td>1234567&nbsp;&nbsp;&nbsp;<td>2002/09/17
 * <tr><td>あいうえお&nbsp;&nbsp;<td>1234567&nbsp;&nbsp;&nbsp;<td>2002/09/17
 * </table>
 * Column1・・・長さ12(データ10 + 半角スペース2)<br/>
 * Column2・・・長さ10(データ7 + 半角スペース3)<br/>
 * Column3・・・長さ10(データ10)<br/><br/>
 *
 * 【コーディング】<br/>
 * TextTable text = new TextTable(TextTable.Type.FIXED_LENGTH_TEXT);<br/>
 * // カラムの追加<br/>
 * text.addColumn("Column1", Column.Type.STRING);<br/>
 * text.addColumn("Column2", Column.Type.INT);<br/>
 * text.addColumn("Column3", Column.Type.DATE);<br/>
 * // カラムの長さを設定<br/>
 * text.getColumn("Column1").setLength(12);<br/>
 * text.getColumn("Column2").setLength(10);<br/>
 * text.getColumn("Column3").setLength(10);<br/>
 * // Column1のデータの右側をトリミング<br/>
 * text.getColumn("Column1").setTrimType(Column.TrimType.RIGHT);<br/>
 * // Column2に数値フォーマットパターンを設定（左ゼロ詰め）<br/>
 * text.getColumn("Column2").setFormatPattern("0000000000");<br/>
 * // Column3に日時フォーマットパターンを設定<br/>
 * text.getColumn("Column3").setFormatPattern("yyyy/MM");<br/>
 * // 出力時の補完処理の設定（左詰）<br/>
 * text.setTextPadType(TextTable.PadType.LEFT);<br/>
 * // 固定長テキストファイルのロード<br/>
 * text.importTextFile("file:///c:/input_data.txt");<br/>
 * // 固定長テキストファイルのセーブ<br/>
 * text.exportTextFile("c:/output_data.txt");<br/><br/>
 *
 * 【出力データ】<br/>
 * <table border=1 cellspacing=0 cellpadding=2>
 * <tr><td>Column1<td>Column2<td>Column3
 * <tr><td> あいうえお<td>0001234567<td>  2002/09
 * <tr><td> あいうえお<td>0001234567<td>  2002/09
 * <tr><td> あいうえお<td>0001234567<td>  2002/09
 * </table><br/>
 * 
 *
 * ※デフォルトではヘッダにカラム名が出力されますが、setHeaderType(TextTable.HeaderType.DATA_FIELD_CAPTION);
 * とすることでカラム名の代わりにキャプションを出力するようになります。<br/>
 * <br/>
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public class TextTable<E extends Serializable> extends Table<E> {

  private Type mTextType = Type.CSV;
  private Schema mTextSchema = Schema.ONLY_DATA;
  private PadType mTextPadType = PadType.RIGHT;
  private char mQuote = '\"';
  private char mDelimiter = Const.COMMA_CHAR;
  private boolean mQuoteInExport = true;
  private String mLineSeparator = null; // Const.LINE_SEPARATOR;
  private String mEncoding = null;
  private HeaderType headerType = HeaderType.COLUMN_NAME;  // export時に出力するヘッダ（カラム名またはキャプション）
  /**
   * export/importする際のバッファサイズです。<br/>
   */
  public static int BUFFER_SIZE = 4096;
  
  //エクスポート時のカラム名
  private String[] exportColumns = null;
  
  private static final transient Logger logger = LogManager.getLogger(TextTable.class.getName());

  /**
   * TextRowSetクラスのコンストラクタ。<br/>
   */
  public TextTable() {
  }

  /**
   * TextRowSetクラスのコンストラクタ。<br/>
   *
   * @param textType テキストタイプ（CSVまたは固定長テキスト）
   */
  public TextTable(Type textType) {
    this();
    mTextType = textType;
  }

  /**
   * テキストタイプ（CSVまたは固定長テキスト）を設定します。<br/>
   *
   * @param textType テキストタイプ
   */
  public void setType(Type textType) {
    mTextType = textType;
  }

  /**
   * テキストタイプ（CSVまたは固定長テキスト）を取得します。<br/>
   *
   * @return テキストタイプ
   */
  public Type getType() {
    return mTextType;
  }

  /**
   * 文字列の補完処理のタイプを設定するします。<br/><br/>
   *
   * 固定長データを扱う場合に利用します。<br/>
   *
   * @param textPadType 補完処理のタイプ
   * @see PadType
   */
  public void setPadType(PadType textPadType) {
    mTextPadType = textPadType;
  }

  /**
   * 文字列の補完処理のタイプを取得します。<br/><br/>
   *
   * 固定長データを扱う場合に利用します。<br/>
   *
   * @return 補完処理のタイプ
   * @see PadType
   */
  public PadType getPadType() {
    return mTextPadType;
  }

  /**
   * CSVデータ入出力時のテキストスキーマを設定します。<br/><br/>
   *
   * CSVデータを扱う場合に利用します。<br/><br/>
   *
   * @param textSchema テキストスキーマ
   * @see Schema
   */
  public void setSchema(Schema textSchema) {
    mTextSchema = textSchema;
  }

  /**
   * CSVデータ入出力時のテキストスキーマを取得します。<br/><br/>
   *
   * CSVデータを扱う場合に利用します。<br/>
   *
   * @return テキストスキーマ
   * @see Schema
   */
  public Schema getSchema() {
    return mTextSchema;
  }

  /**
   * 区切り文字を設定します。<br/><br/>
   *
   * CSVデータを扱う場合に利用します。<br/>
   *
   * @param delimiter 区切り文字 デフォルト値:カンマ[,]
   */
  public void setDelimiter(char delimiter) {
    mDelimiter = delimiter;
  }

  /**
   * 区切り文字を取得します。<br/><br/>
   *
   * CSVデータを扱う場合に利用します。<br/>
   *
   * @return 区切り文字
   */
  public char getDelimiter() {
    return mDelimiter;
  }

  /**
   * クォート文字を設定します。<br/><br/>
   *
   * CSVデータを扱う場合に利用します。<br/>
   *
   * @param quote クォート文字 デフォルト値:ダブルクォート["]
   */
  public void setQuote(char quote) {
    mQuote = quote;
  }

  /**
   * クォート文字を取得します。<br/><br/>
   *
   * CSVデータを扱う場合に利用します。<br/>
   *
   * @return クォート文字
   */
  public char getQuote() {
    return mQuote;
  }

  /**
   * CSVデータ出力時に、クォート文字を出力するかどうかを設定します。<br/><br/>
   *
   * CSVデータを扱う場合に利用します。<br/><br/>
   *
   * デフォルト値：true<br/>
   *
   * @param quoteInExport True：クォート文字を出力する、False：出力しない
   */
  public void setQuoteInExport(boolean quoteInExport) {
    mQuoteInExport = quoteInExport;
  }


  /**
   * CSVデータ出力時に、クォート文字を出力するかどうかのフラグを取得します。<br/><br/>
   *
   * CSVデータを扱う場合に利用します。<br/>
   *
   * @return True：クォート文字を出力する、False：出力しない
   */
  public boolean isQuoteInExport() {
    return mQuoteInExport;
  }


  /**
   * CSVデータ出力時に出力するヘッダタイプを指定します。<br/><br/>
   *
   * CSVデータを Schema.SPECIFIED_NAME_AND_TYPE または Schema.SPECIFIED_NAME のモードで
   * 出力する場合に利用します。<br/>
   *
   * デフォルト値:HeaderType.COLUMN_NAME<br/>
   *
   * @param headerType HeaderType.COLUMN_NAME:カラム名を出力する。 HeaderType.DATA_FIELD_CAPTION:キャプション出力する。
   * @see HeaderType
   */
  public void setHeaderType(HeaderType headerType) {
    this.headerType = headerType;
  }


  /**
   * CSVデータ出力時に出力するヘッダタイプを取得します。<br/><br/>
   *
   * CSVデータを Schema.SPECIFIED_NAME_AND_TYPE または Schema.SPECIFIED_NAME のモードで
   * 出力する場合に利用します。<br/>
   *
   * @return CSVデータ出力時に出力するヘッダタイプ
   * @see HeaderType
   */
  public HeaderType getHeaderType() {
    return this.headerType;
  }


  /**
   * 改行コードを設定します。<br/>
   *
   * @param lineSeparator 改行コード デフォルト値:システムの改行コード
   */
  public void setLineSeparator(String lineSeparator) {
    mLineSeparator = lineSeparator;
  }


  /**
   * 改行コードを取得します。<br/>
   *
   * @return 改行コード
   */
  public String getLineSeparator() {
  	if (this.mLineSeparator == null) return Const.NULLSTRING;
    return mLineSeparator;
  }

  /**
   * 改行コードを取得する。<br/><br/>
   * 
   * LineSeparatorプロパティが空文字の場合は、Const.LINE_SEPARATORを返します。<br/>
   * 
   * @return エンコーディング
   */
  private String getLineSeparatorInternal() {
    final String _lineSeparator = this.getLineSeparator();
    if (_lineSeparator.length() == 0) {
      return Const.LINE_SEPARATOR;
    } else {
      return _lineSeparator;
    }
  }

  /**
   * エンコーディングを設定します。<br/><br/>
   *
   * nullまたは空文字を設定すると、System.getProperty("file.encoding")の値が適用されます。<br/>
   *
   * @param encoding エンコーディング
   */
  public void setEncoding(String encoding) {
    mEncoding = encoding;
  }


  /**
   * エンコーディングを取得します。<br/>
   *
   * @return エンコーディング
   */
  public String getEncoding() {
  	if (this.mEncoding == null) return Const.NULLSTRING;
    return mEncoding;
  }

  /**
   * エンコーディングを取得する。<br/><br/>
   * 
   * Encodingプロパティが空文字の場合は、Const.FILE_ENCODINGを返します。<br/>
   * 
   * @return エンコーディング
   */
  private String getEncodingInternal() {
    final String _mEncoding = this.getEncoding();
    if (_mEncoding.length() == 0) {
      return Const.FILE_ENCODING;
    } else {
      return _mEncoding;
    }
  }

  /**
   * テキストファイルを読み込み、TextTableにデータをセットします。<br/><br/>
   *
   * このメソッドが呼ばれると、内部でsetActive(true)が呼ばれます。<br/>
   * ファイルタイプに従って、データをTextTableにセットします。<br/>
   *
   * @param fileURL ファイルのURL
   * @throws TextException 例外発生時
   */
  public void importTextFile(String fileURL) throws TextException {
    if (StringUtil.isNullString(fileURL)) {
      throw new IllegalArgumentException("fileURL is null or empty String.");
    }
    try {
      // write log
      StringBuffer infoBuf = new StringBuffer();
      infoBuf.append("*** reading ");
      infoBuf.append(mTextType.name().toLowerCase());
      infoBuf.append(" file : ");
      infoBuf.append(fileURL);
      logger.info(infoBuf.toString());

      java.net.URL url = new java.net.URL(fileURL);
      importTextFile(url.openStream());
    } catch (IOException ex) {
      throw new TextException(ex);
    }
  }

  /**
   * テキストデータを読み込み、TextTableにデータをセットします。<br/><br/>
   *
   * このメソッドが呼ばれると、内部でsetActive(true)が呼ばれます。<br/>
   * ファイルタイプに従って、データをTextTableにセットします。<br/>
   *
   * @param data テキストデータ
   * @throws TextException 例外発生時
   */
  public void importTextString(String data) throws TextException {
    //データが存在しない場合
    if (StringUtil.isNullString(data)) {
      throw new IllegalArgumentException("data is null or empty String.");
    }
    importTextFile(new StringReader(data));
  }

  /**
   * テキストファイルを読み込み、TextTableにデータをセットします。<br/><br/>
   *
   * このメソッドが呼ばれると、内部でsetActive(true)が呼ばれます。<br/>
   * ファイルタイプに従って、データをTextTableにセットします。<br/>
   *
   * @param inputStream テキストファイルから取得したInputStream
   * @throws TextException 例外発生時
   */
  public void importTextFile(InputStream inputStream) throws TextException {
    if (inputStream == null) {
      throw new IllegalArgumentException("inputStream is null.");
    }
    try {
      importTextFile(new InputStreamReader(inputStream, this.getEncodingInternal()));
    } catch (TextException ex) {
      throw ex;
    } catch (Exception ex) {
      throw new TextException(ex);
    } finally {
      try {
        inputStream.close();
      } catch (IOException ex) {
      }
    }
  }
  
  /**
   * テキストファイルを読み込み、TextTableにデータをセットします。<br/><br/>
   *
   * このメソッドが呼ばれると、内部でsetActive(true)が呼ばれます。<br/>
   * ファイルタイプに従って、データをTextTableにセットします。<br/>
   *
   * @param inputStream テキストファイルから取得したInputStream
   * @throws TextException 例外発生時
   */
  private void importTextFile(Reader reader) throws TextException {
    if (reader == null) {
      throw new IllegalArgumentException("reader is null.");
    }
    try {
      BufferedReader br = new BufferedReader(reader, BUFFER_SIZE);
      String data;
      this.setActive(false);
      
      // CSVから読み込むカラムのインデックス
      // 既に"COL_A","COL_B"があり、CSVのカラムの並びが"COL_B","COL_A"の場合、int[]{1,0} となる。
      int[] columnIndexes;
      
      // ※既存のカラムは削除しません。
      if (this.getType() == Type.CSV && Schema.SPECIFIED_NAME == mTextSchema) {
        /** １行目[カラム名] + ２行目以降[データ] */
        /**  1st line[column name] + lines after 2nd line[data] */
        // 1行目のカラム名を読み込む。
        data = br.readLine();
        Vector<String> columnNames = getTokens(data);
        columnIndexes = new int[columnNames.size()];
        for (int i = 0; i < columnNames.size(); i++) {
          String columnName = (String)columnNames.get(i);
          if (this.getColumn(columnName,true) == null) {
            this.addColumn(columnName, Column.Type.STRING);
          }
          columnIndexes[i] = this.getColumns().indexOfKey(columnName);
        }
      } else if (this.getType() == Type.CSV && Schema.SPECIFIED_NAME_AND_TYPE == mTextSchema) {
        /** １行目[カラム名] + ２行目[カラムタイプ] + ３行目以降[データ] */
        /** 1st line[column name] + 2nd line[column type] + lines after 3rd line[data]  */
        // 1行目のカラム名を読み込む。
        data = br.readLine();
        Vector<String> columnNames = getTokens(data);
        // 2行目のカラムタイプを読み込む。
        data = br.readLine();
        Vector<String> columnTypes = getTokens(data);
        // 数が一致しているか?
        if (columnNames.size() != columnTypes.size()) {
          throw new TextException(null, null, "COLUMN_NAME_AND_TYPE_NUMBERS_DIFFER",
            new String[]{this.getName(), Integer.toString(columnNames.size()), Integer.toString(columnTypes.size())});
        }
        
        columnIndexes = new int[columnNames.size()];
        for (int i = 0; i < columnNames.size(); i++) {
          String columnName = (String)columnNames.get(i);
          if (this.getColumn(columnName,true) == null) {
            this.addColumn(columnName, Column.Type.valueOf((String)columnTypes.get(i)));
          } else {
            //既にカラムが設定されている場合は型を変更する。
            this.getColumn(columnName).setType(Column.Type.valueOf((String)columnTypes.get(i)));
          }
          columnIndexes[i] = this.getColumns().indexOfKey(columnName);
        }
      } else {
        // 1行目からデータが入っている。
        // カラムは既存のものを利用する。
        columnIndexes = new int[this.getColumnCount()];
        for (int i=0; i<columnIndexes.length; i++) {
          columnIndexes[i] = i;
        }
      }

      if (columnIndexes.length == 0) {
        throw new TextException(null, null, "NO_FIELDS_ERROR",
            new String[]{this.getName()});
      }
      int columnCount = this.getColumnCount();
      Vector<String> tokens;
      int readColumnCount;
      //int maxRows = (this.getMaxRows() <= 0) ? Integer.MAX_VALUE : this.getMaxRows();
      int maxRows = Integer.MAX_VALUE;
      int dataRowCount = 0;
      this.setActive(true);
      while ((data = br.readLine()) != null && dataRowCount < maxRows) {
        dataRowCount++;
        tokens = getTokens(data);
        appendRow();
        readColumnCount = (tokens.size() < columnCount) ? tokens.size() : columnCount;
        for (int i=0; i<readColumnCount; i++) {
          this.getRow().setValueAsString(columnIndexes[i], parseValue(i, (String)tokens.get(i)));
        }
      }
      validateAllRows();
    } catch (IOException e) {
      if (e instanceof UnsupportedEncodingException) {
        throw new TextException(e, null, "INVALID_ENCODING", new String[]{this.getEncodingInternal()});
      }
      throw new TextException(e, null, "IMPORT_FILE_FAIL", new String[]{this.getName()});
    } catch (DataException e) {
      throw new TextException(e);
    } finally {
      try {
        reader.close();
      } catch (IOException ex) {
      }
    }
  }

  /**
   * TextTableのデータをテキストファイルとして保存します。<br/><br/>
   *
   * エンコードは、setEncoding()で設定したエンコードが適用されます。<br/>
   *
   * @param fileName ファイル名
   * @throws TextException 例外発生時
   */
  public void exportTextFile(String fileName) throws TextException {
  	  OutputStream out;
  	  try {
    	  out = new FileOutputStream(fileName);
    } catch (FileNotFoundException e) {
    	  throw new TextException(e, null, "EXPORT_FILE_FAIL",
              new String[]{this.getName(), fileName});
    }
    export(out);

    // write log
    if (logger.isDebugEnabled()) {
      	final StringBuffer infoBuf = new StringBuffer();
      	infoBuf.append("*** saving ")
    	         .append(mTextType.name().toLowerCase())
    	         .append(" file : ")
    	         .append(fileName);
    	  logger.debug(infoBuf.toString());
    }
  }

  /**
   * TextTableのデータをテキストデータとして取得します。<br/><br/>
   *
   * エンコードは、setEncoding()で設定したエンコードが適用されます。<br/>
   *
   * @return テキストデータ
   * @throws TextException 例外発生時
   */
  public String exportTextString() throws TextException {
    ByteArrayOutputStream bout = new ByteArrayOutputStream();
    export(bout);
    try {
      return bout.toString(this.getEncodingInternal());
    } catch (UnsupportedEncodingException e) {
      throw new TextException(e, null, "INVALID_ENCODING", new String[]{this.getEncodingInternal()});
    }
  }

  /**
   * TextTableのデータをOutputStreamに出力する。<br/><br/>
   *
   * エンコードは、setEncoding()で設定したエンコードが適用されます。<br/>
   * 
   * @param out 出力先のOutputStreamオブジェクト
   * @throws TextException
   */
  public void export(OutputStream out) throws TextException {
    if (out == null) {
      throw new IllegalArgumentException("OutputStream is null.");
    }
    Writer writer;
    try {
      writer = new BufferedWriter(
          new OutputStreamWriter(out, this.getEncodingInternal()), BUFFER_SIZE);
    } catch (IOException e) {
      if (e instanceof UnsupportedEncodingException) {
        throw new TextException(e, null, "INVALID_ENCODING", new String[]{this.getEncodingInternal()});
      } else {
        	throw new TextException(e, null, "EXPORT_DATA_FAIL", new String[]{this.getName()});
      }
    }
    export(writer);
  }

  /**
   * TextTableのデータをWriterに出力する。<br/><br/>
   * 
   * 出力後、writerオブジェクトはクローズされます。<br/>
   * 
   * @param writer 出力先のWriterオブジェクト
   * @throws TextException
   */
  public void export(Writer writer) throws TextException {
    if (writer == null) {
      throw new IllegalArgumentException("Writer is null.");
    }
    try {
      checkExportData();
      printTextData(writer);
    } finally {
      try {
        writer.close();
      } catch (IOException ex) {
      }
    }
  }
  
  /**
   * アクティブ時の内部設定. <br/><br/>
   *
   * このメソッドの処理は空です。<br/>
   *
   * @exception  jp.littlesoft.data.DataException  エラー発生時
   */
  protected void internalOnActivate() throws DataException {};

  /**
   * 追加行が存在するかどうかを調べます. <br/><br/>
   *
   * このメソッドは常にfalseを返します。<br/>
   *
   * @param   toLast True：最終行まで調べる、False：調べない
   * @return  False：追加行が存在しない（常にfalseが返されます）
   * @exception  jp.littlesoft.data.DataException  例外発生時
   */
  public boolean checkMoreRow(boolean toLast) throws DataException {
    return false;
  }



  /**
   * 
   * @see jp.littlesoft.data.Table#doLoadData()
   */
  protected void doLoadData() throws DataException {
  }

  /**
   * 出力するカラム名配列を設定します。<br/><br/>
   *
   * 設定されたカラム名のカラムデータを出力します。<br/>
   * nullを指定すると全カラムを出力します。<br/>
   *
   * @param exportFieldNames 出力するカラムデータのカラム名配列
   */
  public void setExportColumns(String[] exportFieldNames) {
    this.exportColumns = exportFieldNames;
  }

  /**
   * 出力するカラム名配列を取得します。<br/><br/>
   *
   * 設定されたカラム名のカラムデータを取得します。<br/>
   *
   * @return 出力するカラムデータのカラム名配列
   */
  public String[] getExportColumns() {
    return this.exportColumns;
  }

/*********************************************************************************
 以下、プライベートメソッド
**********************************************************************************/

  /**
   * 以下のデータ型の場合、両端のスペースをトリミングする。<br/><br/>
   *
   * BYTE、SHORT、LONG、INT、FLOAT、DOUBLE、BOOLEAN、BIGDECIMAL、DATE、TIME、TIMESTAMP<br/><br/>
   *
   * @param columnIndex カラムインデックス
   * @param data データ
   * @return トリミングされたデータ
   * @throws DataException 例外発生時
   */
  private String parseValue(int columnIndex, String data) throws DataException {
    Column.Type type = getColumn(columnIndex).getType();
    switch (type) {
    case BYTE:
    case SHORT:
    case LONG:
    case INT:
    case FLOAT:
    case DOUBLE:
    case BOOLEAN:
    case BIGDECIMAL:
    case DATE:
    case TIME:
    case TIMESTAMP:
      return data.trim();
    default:
      return data;
    }
  }

  /**
   * トークンを取得します。<br/>
   *
   * @param data データ
   * @return トークンの配列
   * @throws DataException 例外発生時
   */
  private Vector<String> getTokens(String data) throws DataException {
    /** CSVデータ */
    /** CSV data  */
    if (mTextType == Type.CSV) {
      return getCSVTokens(data);
    /** Text（固定長）データ */
    /** Fixed length data */
    } else if (mTextType == Type.FIXED_LENGTH_TEXT) {
      return getFixedLengthTextTokens(data);
    }
    return null;
  }

  /**
   * CSVデータのトークンを取得します。<br/>
   *
   * @param data データ
   * @return トークンの配列
   * @throws NoSuchElementException 例外発生時
   */
  private Vector<String> getCSVTokens(String data) throws java.util.NoSuchElementException {
    //int length = data.length();
    int count = countTokens(data);
    int curIndex = 0;
    Vector<String> result = new Vector<String>();
    for (int i = 0; i < count; i++) {
      if (curIndex > data.length()) {
        // 文字列にトークンが残っていないとき
        throw new java.util.NoSuchElementException();
      }
      int index = nextDerimIndex(data, curIndex);
      String next = data.substring(curIndex, index);
      curIndex = index + 1;
      result.add(replaceQuoteForImport(next));
    }
    return result;
  }

  /**
   * 固定長データのトークンを取得します。<br/>
   *
   * @param data データ
   * @return トークンの配列
   * @throws DataException 例外発生時
   */
  private Vector<String> getFixedLengthTextTokens(String data) throws DataException {
    int[] columnLength = getColumnLength();
    int startIndex = 0;
    Vector<String> result = new Vector<String>();
   	final String _encode = this.getEncodingInternal();
    try {
      byte[] byteData = data.getBytes(_encode);
      for (int i = 0; i < columnLength.length; i++) {
        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        if (byteData.length > startIndex + columnLength[i]) {
          bout.write(byteData, startIndex, columnLength[i]);
          result.addElement(bout.toString(_encode));
          startIndex += columnLength[i];
        //レコード長 ＜= カラムの長さの場合、レコード長 - startIndex をセットしてbreakする。
        } else {
          bout.write(byteData, startIndex, byteData.length - startIndex);
          result.addElement(bout.toString(_encode));
          break;
        }
      }
    }
    catch (UnsupportedEncodingException e) {
      // TextExceptionじゃなくていいの?
      throw new DataException(e, null, "INVALID_ENCODING", new String[]{_encode});
    }
    return result;
  }

  /**
   * カラムの長さを配列で取得します。<br/>
   *
   * @return カラムの長さの配列
   * @throws DataException 例外発生時
   */
  private int[] getColumnLength() throws DataException {
    int[] columnLength = new int[getColumnCount()];
    for (int i = 0; i < getColumnCount(); i++) {
      columnLength[i] = getColumn(i).getLength();
    }
    return columnLength;
  }

  /**
   * テキストデータを出力する。<br/>
   *
   * @param out Writerクラス
   * @throws TextException 例外発生時
   */
  private void printTextData(Writer out) throws TextException {
    if (mTextType == Type.CSV) {
      printCSVData(out);
    } else if (mTextType == Type.FIXED_LENGTH_TEXT) {
      printFixedLengthTextData(out);
    }
  }

  /**
   * CSVデータを出力します。<br/>
   *
   * @param out Writerクラス
   * @throws TextException 例外発生時
   */
  private void printCSVData(Writer out) throws TextException {
    if (mTextSchema == Schema.SPECIFIED_NAME) {
      printColumnNames(out);
    } else if (mTextSchema == Schema.SPECIFIED_NAME_AND_TYPE) {
      printColumnNames(out);
      printColumnTypes(out);
    }
    printData(out);
  }

  /**
   * 出力データのチェックを行う。<br/><br/>
   *
   * Tableの行数が０かつ、TextSchemaがデータのみの場合、エラーを投げる。<br/>
   *
   * @throws TextException 例外発生時
   */
  private void checkExportData() throws TextException {
    if (getRowCount() == 0 && mTextSchema == Schema.ONLY_DATA) {
      throw new TextException(null, null, "EXPORT_DATA_NOT_EXISTED",
          new String[]{this.getName()});
    }
  }

  /**
   * カラム名を出力します。<br/>
   *
   * @param out Writerクラス
   * @throws TextException 例外発生時
   */
  private void printColumnNames(Writer out) throws TextException {
    boolean isFirstColumn = true;
    try {
      if (this.getExportColumns() != null) {
      	if (HeaderType.CAPTION == this.getHeaderType()) {
      		// キャプションを出力する。
          for (int i = 0; i < getExportColumns().length; i++) {
            print(isFirstColumn, out, this.getColumn(getExportColumns()[i]).getCaption());
            isFirstColumn = false;
          }
      	} else {
      		// カラム名を出力する。
          for (int i = 0; i < getExportColumns().length; i++) {
            print(isFirstColumn, out, getExportColumns()[i]);
            isFirstColumn = false;
          }
      	}
      } else {
      	if (HeaderType.CAPTION == this.getHeaderType()) {
      		// キャプションを出力する。
          for (int i = 0; i < this.getColumnCount(); i++) {
            print(isFirstColumn, out, this.getColumn(i).getCaption());
            isFirstColumn = false;
          }
      	} else {
      		// カラム名を出力する。
          for (int i = 0; i < this.getColumnCount(); i++) {
            print(isFirstColumn, out, this.getColumn(i).getName());
            isFirstColumn = false;
          }
      	}
      }
    }
    catch (DataException ex) {
      throw new TextException(ex);
    }
    println(out);
  }


  /**
   * カラムタイプを出力します。<br/>
   *
   * @param out Writerクラス
   * @throws TextException 例外発生時
   */
  private void printColumnTypes(Writer out) throws TextException {
    boolean isFirstColumn = true;
    try {
      if (getExportColumns() != null) {
        for (int i = 0; i < getExportColumns().length; i++) {
          print(isFirstColumn, out, this.getColumn(getExportColumns()[i]).getType().toString());
          isFirstColumn = false;
        }
      } else {
        for (int i = 0; i < this.getColumnCount(); i++) {
          print(isFirstColumn, out, this.getColumn(i).getType().toString());
          isFirstColumn = false;
        }
      }
    }
    catch (DataException ex) {
      throw new TextException(ex);
    }
    println(out);
  }

  /**
   * データを出力します。<br/>
   *
   * @param out Writerクラス
   * @throws TextException 例外発生時
   */
  private void printData(Writer out) throws TextException {
    try {
      if (this.getSortKeys() != null && this.getSortKeys().length > 0) {
        // ソートを行う。
        this.sortRows();
      }
    
      for (int row = 0; row < this.getRowCount(); row++) {
        boolean isFirstColumn = true;
        if (getExportColumns() != null) {
      	  this.moveTo(row);
          for (int col = 0; col < getExportColumns().length; col++) {
            print(isFirstColumn, out, this.getRow().getValueAsString(getExportColumns()[col]));
            isFirstColumn = false;
          }
        } else {
          for (int col = 0; col < this.getColumnCount(); col++) {
            print(isFirstColumn, out, this.getRow(row).getValueAsString(col));
            isFirstColumn = false;
          }
        }
        println(out);
      }
    } catch (DataException ex) {
      throw new TextException(ex);
    }
  }

  /**
   * 固定長データを出力します。<br/>
   *
   * @param out Writerクラス
   * @throws TextException 例外発生時
   */
  private void printFixedLengthTextData(Writer out) throws TextException {
    try {
      if (this.getSortKeys() != null && this.getSortKeys().length > 0) {
        // ソートを行う。
        this.sortRows();
      }

      for (int i = 0; i < this.getRowCount(); i++) {     
        if (getExportColumns() != null) {
          for (int col = 0; col < getExportColumns().length; col++) {
              int j = this.getColumns().indexOfKey(getExportColumns()[col]);
              out.write(padSpace(i, j));
          }
        } else {
          for (int j = 0; j < this.getColumnCount(); j++) {
            out.write(padSpace(i, j));
          }
        }
        println(out);
      }
    } catch (DataException ex) {
      throw new TextException(ex);
    } catch (IOException ex) {
      throw new TextException(ex);
    }
  }

  /**
   * TextPadTypeの値（左詰、右詰）を元に、半角スペースで補完処理を行います。<br/>
   *
   * @param rowCount 行番号
   * @param columnCount カラム番号
   * @return 補完処理後の文字列
   * @throws TextException 例外発生時
   */
  private String padSpace(int rowCount, int columnCount) throws DataException {
    int columnLength = getColumn(columnCount).getLength();
    String data = getRow(rowCount).getValueAsString(columnCount);
    int dataLength = data.getBytes().length;
    //カラムの長さとデータの長さが同じ場合は何もしない。
    if (dataLength == columnLength) {
      return data;
    }
    String result;
    if (mTextPadType == PadType.LEFT || mTextPadType == PadType.RIGHT) {
      // 余分に追加する空白を作成
      StringBuffer sb = new StringBuffer();
      for (int i = columnLength - dataLength; i > 0; i--) {
        sb.append(' ');
      }
      
      if (mTextPadType == PadType.LEFT) {
        result = sb.append(data).toString();
      } else {
        // RIGHT
        result = data + sb.toString();
      }
    } else {
      result = data;
    }
    return result;
  }

  /**
   * 文字列を出力します。<br/>
   *
   * @param columnNo カラムNo
   * @param out Writerクラス
   * @param value 文字列
   */
  private void print(boolean isFirstColumn, Writer out, String value) throws TextException {
    try {
      if (!isFirstColumn) {
        // 先頭の要素でない場合は、区切り文字を出力
        out.write(String.valueOf(mDelimiter));
      }
      if (mQuoteInExport) {
        // 前後をQuote文字で囲む。
        out.write(mQuote);
        out.write(replaceQuoteForExport(value));
        out.write(mQuote);
      } else {
        out.write(replaceQuoteForExport(value));
      }
    } catch (IOException ex) {
      throw new TextException(ex);
    }
  }

  /**
   * 改行を出力します。<br/>
   *
   * @param out Writerクラス
   * @throws TextException 例外発生時
   */
  private void println(Writer out) throws TextException {
    try {
      out.write(this.getLineSeparatorInternal());
    } catch (Exception ex) {
      throw new TextException(ex);
    }
  }

  /**
   * 残りの要素の個数を返します。<br/><br/>
   *
   * CSVデータの読み込み時に利用。<br/>
   *
   * @param data 文字列
   * @return int 残りの要素の個数
   */
  private int countTokens(String data) {
    int count = 0;
    int index = 0;
    while (index < data.length()) {
      count++;
      index = nextDerimIndex(data, index) + 1;
    }
    return count;
  }

  /**
   * 次の区切り文字の位置を返します。<br/><br/>
   *
   * 存在しない場合は、最大の位置 + 1 (文字列の長さ) を返す。<br/>
   * CSVデータの読み込み時に利用。<br/>
   *
   * @param data 文字列
   * @param index インデックス
   * @return int 次の区切り文字の位置
   */
  private int nextDerimIndex(String data, int index) {
    // クォート文字で囲まれた中かどうかのフラグ
    boolean in = false;
    while (index < data.length()) {
      if (data.charAt(index) == mQuote) {
        // クォート文字が現れるたびに、フラグ反転
        in = !in;
      } else if ((data.charAt(index) == mDelimiter) && !in) {
        // クォート文字で囲まれない状態で , が現れたらループ終了
        break;
      }
      index++;
    }
    return index;
  }

  /**
   * 先頭と最後の クォート文字 を取り除き、
   * 連続したクォート文字を 単一のクォート文字に変換します。<br/><br/>
   *
   * CSVデータの読み込み時に利用。<br/>
   *
   * @param s 置換前の文字列
   * @return String 置換後の文字列
   */
  private String replaceQuoteForImport(String s) {
    int index = 0;
    int maxIndex = s.length() - 1;
    StringBuffer sb = new StringBuffer(s.length());
    // 空文字列の場合
    if (s.length() == 0) {
      return sb.toString();
    }
    // 先頭のクォート文字を取り除く
    if (s.charAt(index) == mQuote) {
      index++;
    }
    // 最後のクォート文字を取り除く
    if (s.charAt(maxIndex) == mQuote) {
      maxIndex--;
    }
    // 連続したクォート文字を 単一のクォート文字に変換する
    while (index <= maxIndex) {
      sb.append(s.charAt(index));
      if (s.charAt(index) == mQuote) {
        if (((index + 1) <= maxIndex) && (s.charAt(index + 1) == mQuote)) {
          index++;
        }
      }
      index++;
    }
    return sb.toString();
  }

  /**
   * クォート文字を２つのクォート文字に変換します。<br/><br/>
   *
   * CSVデータの出力時に利用。<br/>
   *
   * @param s 置換前の文字列
   * @return String 置換後の文字列
   */
  private String replaceQuoteForExport(String s) {
    int index = 0;
    int maxIndex = s.length() - 1;
    StringBuffer sb = new StringBuffer(s.length());
    // 空文字列の場合
    if (s.length() == 0) {
      return sb.toString();
    }
    // クォート文字を２つのクォート文字に変換する
    while (index <= maxIndex) {
      sb.append(s.charAt(index));
      if (s.charAt(index) == mQuote) {
      	sb.append(mQuote);
      }
      index++;
    }
    return sb.toString();
  }

  /**
   * 出力の種類を管理するクラス。<br/>
   * 
   * @author LittleSoft Corporation
   */
  public enum Type {
    /**
     * CSVデータを表す定数です。
     */
    CSV,
    /**
     * Text（固定長）データを表す定数です。
     */
    FIXED_LENGTH_TEXT,
  }

  /**
   * ヘッダの種類を管理するクラス。<br/>
   * 
   * @author LittleSoft Corporation
   */
  public enum HeaderType {
    /**
     * カラム名を出力を表す定数です。
     */
    COLUMN_NAME,
    /**
     * キャプションを出力を表す定数です。
     */
    CAPTION
  }

  /**
   * 補完の種類を管理するクラス。<br/>
   * 
   * @author LittleSoft Corporation
   */
  public enum PadType {
    /**
     * 補完しないを表す定数です。
     */
    COLUMN_NOT_ASSIGNED,
    /**
     * 左側に補完処理を表す定数です。
     */
    LEFT,
    /**
     * 右側に補完処理を表す定数です。
     */
    RIGHT
  }	

  /**
   * 出力構成の種類を管理するクラス。<br/>
   * 
   * @author LittleSoft Corporation
   */
  public enum Schema {
    /**
     * 「１行目以降[データ]」を表す定数です。
     */
    ONLY_DATA,
    /**
     * 「１行目[カラム名] + ２行目以降[データ] 」を表す定数です。
     */
    SPECIFIED_NAME, 
    /**
     * 「１行目[カラム名] + ２行目[カラムタイプ] + ３行目以降[データ]」を表す定数です。
     */
    SPECIFIED_NAME_AND_TYPE
  }	
}
