/** Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.data.validator;

import java.io.Serializable;
import java.math.BigDecimal;
import java.math.BigInteger;

import jp.littlesoft.data.Column;
import jp.littlesoft.data.Row;
import jp.littlesoft.data.exception.InvalidValueException;

/**
 * 数値検証クラス。<br/><br/>
 * BIGDECIMAL型カラムのディフォルト検証クラスです。
 * 以下のチェックを行います。<br/>
 * ・整数部の桁数チェック: column.getPrecision()およびgetScale()を参照します。<br/>
 * 
 * @see jp.littlesoft.data.Column#setValidator
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public class NumberValidator implements Validator, Serializable {
  /**
   * コンストラクタ。<br/>
   */
  public NumberValidator() {
  }

	/**
	 * 数値の検証を行います。 <br/><br/>
   * 
	 * 引数で渡されるカラムオブジェクトの定義をもとに、渡された値を検証します。<br/><br/>
	 *
	 * @param  column 対象となるオブジェクト
	 * @param  value  検証対象値(java.math.BigDecimalオブジェクト)
	 * @throws InvalidValueException エラーが検出された場合に送出される。
   * @see Column#setRegexp(String)
   */
	@SuppressWarnings("unchecked")
  public void execute(Row row, Column column, Object value) throws InvalidValueException {
    if(value == null) {
    	if (column.isNullable()) return;
      throw new InvalidValueException("INVALID_NULL_VALUE",
          new String[]{column.getCaption()}, row, column, null);
    }

    if (!(value instanceof java.lang.Number))
      throw new IllegalArgumentException("Argument \"value\" must be java.lang.Number Object! " +value);

    if (column.getPrecision() >= 0) {
      // 整数部の桁数をチェックする。
      // 小数部は誤差が混入する可能性があるためチェックできない。

      // 整数部だけにする。
      BigInteger bigInt;
      if (value instanceof java.math.BigDecimal) {
        bigInt = ((BigDecimal)value).abs().toBigInteger();
      } else if (value instanceof java.math.BigInteger){
        bigInt = ((BigInteger)value).abs();
      } else {
        bigInt = new BigDecimal(((Number)value).doubleValue()).abs().toBigInteger();
      }
      
      int _pre = (column.getScale() >= 0) ?
          column.getPrecision() - column.getScale() : column.getPrecision();
      if (_pre == 0) {
        // 0.12345みたいなケース
        if (bigInt.intValue() != 0) {
          throw new InvalidValueException("PARSE_ERROR_INVALID_PRECISION",
              new Object[]{column.getCaption(), value},
              row, column, value);
        }
      } else if (_pre > 0) {
        if (bigInt.toString().length() > _pre) {
          throw new InvalidValueException("PARSE_ERROR_INVALID_PRECISION",
              new Object[]{column.getCaption(), value},
              row, column, value);
        }
      }
    }
	}
}
