/** Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.data.validator;

import java.io.Serializable;
import java.util.regex.*;

import jp.littlesoft.data.Column;
import jp.littlesoft.data.Row;
import jp.littlesoft.data.exception.InvalidValueException;
import jp.littlesoft.util.Const;
import jp.littlesoft.util.DataUtil;

/**
 * 文字列値検証クラス。<br/><br/>
 * 以下のチェックを行います。<br/>
 * ・ヌルチェック: column.isNullable()を参照します.<br/>
 * ・長さチェック: column.getLength()を参照します.<br/>
 * ・制約チェック: column.getConstraint()を参照します.<br/>
 *
 * 文字型のカラムのディフォルト検証クラスです。<br/>
 * ただし、PostgreSQL 7.2以上のデータベースから取得した文字型カラムの場合は、StringLengthValidatorクラスが検証クラスとして割り当てられます。<br/><br/>
 *
 * @see jp.littlesoft.data.Column#setConstraint
 * @see jp.littlesoft.data.Column#setValidator
 *
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public class StringValidator implements Validator, Serializable {

  /**
   * コンストラクタ。<br/>
   */
  public StringValidator() {
  }

	/**
	 * 文字列値の検証を行います。 <br/><br/>
   * 
	 * 引数で渡されるカラムオブジェクトの定義をもとに、渡された値を検証します。<br/>
	 * 長さチェックはバイト数で行います。<br/>
   * <br/>
   * 下記の条件を満たす場合に正規表現によるチェックを行います。<br/>
   * ・カラムにチェック用の正規表現が設定されていること。<br/>
   * ・カラムに格納しようとしている値の長さが1以上であること。<br/>
   * 　（長さ０のチェックはColumn#setNullable(false)で行ってください。）<br/><br/>
	 *
	 * @param  column カラム
	 * @param  value  検証対象値(java.lang.Stringオブジェクト)
	 * @throws InvalidValueException エラーが検出された場合に送出される。
   * @see Column#setRegexp(String)
   */
	@SuppressWarnings("unchecked")
  public void execute(Row row, Column column, Object value) throws InvalidValueException {

    if(value == null) {
    	if(column.isNullable()) return;
      throw new InvalidValueException(
          "INVALID_NULL_VALUE",
          new String[]{column.getCaption()}, row, column, null);
    }
    if(!(value instanceof String)) throw new IllegalArgumentException("Argument \"value\" must be java.lang.String Object! " +value);
    String str = (String)value;

    int valLen = 0;
    if (row.getTable()!=null) {
      // バイト数でチェックする。
      try {
        valLen = str.getBytes(row.getTable().getCharsetForCheckingLength()).length;
      } catch(java.io.UnsupportedEncodingException e) {
        throw new IllegalArgumentException(e.getMessage()+": "+value);
      }
      
      // 長さエラー条件 column.getLength()>Const.NOT_ASSIGNED について:
      //
      // DBMSによっては、MetaDataから正しく長さが取得できない。
      // PostgreSQLでは、-5 などの値が取得されてしまうので、負値の場合は検査対象から
      // はずすことにする。
      if(column.getLength()>0 && valLen > column.getLength()){
        String _caption = column.getCaption();
        String _dispValue = (String)value;
        if (((String)value).length() > Const.ERR_MESSAGE_SPLIT_CHAR_COUNT) {
          _dispValue = ((String)value).substring(0, Const.ERR_MESSAGE_SPLIT_CHAR_COUNT) + " ...";
        }
        throw new InvalidValueException("PARSE_ERROR_TOO_LONG_STRING",
            new Object[]{_caption, _dispValue}, row, column, value);
      }
    }
		// checking constraints
    switch (column.getConstraint()) {
      case REQUIRED_ALPHA_NUMERICS:
        if(!(DataUtil.isHalfSize(str) && DataUtil.isAlphOrNumber(str)))
          throw new InvalidValueException("PARSE_ERROR_NON_ALPHA_NUMERICS",
            new String[]{column.getCaption()}, row, column, value);
        break;
      case REQUIRED_SINGLE_BYTES:
        if(!DataUtil.isHalfSize(str))
          throw new InvalidValueException("PARSE_ERROR_NON_SINGLE_BYTES",
            new String[]{column.getCaption()}, row, column, value);
        break;
      case REQUIRED_DOUBLE_BYTES:
        if(!DataUtil.isFullSize(str))
          throw new InvalidValueException("PARSE_ERROR_NON_DOUBLE_BYTES",
            new String[]{column.getCaption()}, row, column, value);
        break;
      case REQUIRED_SINGLE_BYTES_WITHOUT_KANA:
        if(!DataUtil.isHalfSizeWithoutKana(str))
          throw new InvalidValueException("PARSE_ERROR_NON_SINGLE_BYTES_WITHOUT_KANA",
              new String[]{column.getCaption()}, row, column, value);
        break;
    }

    // Regular Expression Check
    Pattern _regexpPattern = column.getRegexpPattern();
    if (valLen > 0 && _regexpPattern != null) {
      validateRegexp(row, column, _regexpPattern, str);
    }
	}
  
  @SuppressWarnings("unchecked")
  void validateRegexp(Row row, Column column, Pattern regexpPattern, String inStr) throws InvalidValueException{
  	Matcher m = regexpPattern.matcher(inStr);
		if (!m.find()) {
	      String _caption = column.getCaption();
	      throw new InvalidValueException("PARSE_ERROR_REGEXP",
	          new String[]{_caption, inStr}, row, column, inStr);
		}
  }
}
