/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.file.client;

import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStreamWriter;
import java.io.Serializable;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLEncoder;
import java.text.NumberFormat;
import java.util.ArrayList;

import jp.littlesoft.file.common.Attributes;
import jp.littlesoft.file.common.IProgressListener;
import jp.littlesoft.file.common.SplitDownloadParams;
import jp.littlesoft.soja.LSJson;
import jp.littlesoft.util.StringUtil;
import jp.littlesoft.web.ActionPage;

/**
 * ダウンロード クライアントアプリケーションクラス<br /><br />
 * 
 * ファイルを分割してダウンロードする、Javaクライアントアプリケーションクラスです。
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public class SplitDownloadClient implements Serializable {

  // 呼出し元JNLPファイルに定義されたJSONオブジェクトのパラメータ群
  private Attributes attributes;

  // ダウンロードファイルの出力ストリーム
  private FileOutputStream fileOutputStream;

  // ダウンロード状況の書込みクラス
  private IProgressListener iProgress;


  /**
   * コンストラクタ。
   * 
   * @param attributes JSONオブジェクトのパラメータ群
   * @param fileOutputStream ファイル出力ストリーム
   * @param iProgress ダウンロード状況の書込みクラス
   * @throws Exception 例外
   */
  public SplitDownloadClient(Attributes attributes, FileOutputStream fileOutputStream, IProgressListener iProgress) throws Exception {
    this.attributes = attributes;
    this.fileOutputStream = fileOutputStream;
    this.iProgress = iProgress;
  }


  /**
   * ダウンロードを実行する。
   * 
   * @throws Exception 例外
   */
  public void execute() throws Exception {

    SplitDownloadParams params = new SplitDownloadParams();
    params.setAttributes(attributes);

    ArrayList<String> cookies = new ArrayList<String>();
    cookies.add("JSESSIONID=" + attributes.jsessionId);

    // --- URLConnection、XxxxStreamを定義 ---
    URL url = null;
    HttpURLConnection connection = null;
    OutputStreamWriter out = null;
    InputStream inputStream = null;
    NumberFormat sizeFormat = NumberFormat.getNumberInstance();

    // ファイル保存先(BTableのkey)を取得
    String fileStorageId = (String)attributes.get("fileStorageId");
    int totalSize = 0;
    try {

      System.out.println("###### Download start");
      while (true) {
        // --- サーバー接続 ---
        url = new URL(attributes.webappUrl);
        connection = (HttpURLConnection) url.openConnection();
        connection.setDoOutput(true);
        connection.setRequestMethod("POST");

        if (cookies!=null) {
          String cookie = "";
          for (String s : cookies) {
            cookie += s;
          }
          connection.setRequestProperty("Cookie", cookie);
        }

        out = new OutputStreamWriter( connection.getOutputStream() );
        if (attributes.sequence != null) {
          out.write(ActionPage.PN_SEQ + "=" + URLEncoder.encode(attributes.sequence, "UTF-8") + "&");
        }
        out.write(ActionPage.PN_JSESSION_ID + "=" + URLEncoder.encode(attributes.jsessionId, "UTF-8") + "&");
        out.write(ActionPage.PN_SUBSESSION_ID + "=" + URLEncoder.encode(attributes.subsessionId, "UTF-8") + "&");
        out.write(ActionPage.PN_ACTIONPAGE_ID + "=" + URLEncoder.encode(attributes.actionPageId, "UTF-8") + "&");
        out.write(ActionPage.PN_METHOD + "=" + URLEncoder.encode("onJavaAppCallBack", "UTF-8") + "&");
        out.write(ActionPage.PN_APPLICATION_PARAMS + "=" + URLEncoder.encode(LSJson.serialize(params), "UTF-8") + "&");
        if ( StringUtil.isNotNullString(fileStorageId) ) {
          out.write("fileStorageId=" + URLEncoder.encode(fileStorageId, "UTF-8") + "&");
        }
        // <--- ダウンロードサイト用 画面間パラメータ ----
        // 暗号化文字列
        if ( StringUtil.isNotNullString((String)attributes.get("_i")) ) {
          out.write("_i=" + URLEncoder.encode((String)attributes.get("_i"), "UTF-8") + "&");
        }
        // アカウント・キー
        if ( StringUtil.isNotNullString((String)attributes.get("_a")) ) {
          out.write("_a=" + URLEncoder.encode((String)attributes.get("_a"), "UTF-8") + "&");
        }
        // ---------------------------------------------->
        out.flush();
        out.close();

        // --- HTTP接続 ---
        int responseCode = connection.getResponseCode();

        // --- HTTP送信 - 有効 ---
        if (responseCode == HttpURLConnection.HTTP_OK) {
          System.out.println("### Complete: POST OK");

          // InputStream取得
          inputStream = connection.getInputStream();

          // レスポンス・ヘッダーに"error"があれば、Serverから返却されたエラーメッセージ
          String error = connection.getHeaderField("error");
          if ( "error".equals(error) ) {
            System.out.println("###### Download error end");

            byte[] currentResponseByte = null;
            while (true) {
              byte[] b = new byte[1];
              int ret = inputStream.read(b, 0, 1);
              if (ret < 0) {
                break;
              }
              // byte配列の連結
              if (currentResponseByte == null) {
                currentResponseByte = b;
              } else {
                byte[] wk = currentResponseByte;
                currentResponseByte = new byte[wk.length + b.length];
                System.arraycopy(wk, 0, currentResponseByte, 0, wk.length);
                System.arraycopy(b, 0, currentResponseByte, wk.length, b.length);
              }
            }
            // エラー返却
            if (currentResponseByte != null) {
              String errorMessage = new String(currentResponseByte, "UTF-8");
              throw new Exception(errorMessage);
            } else {
              throw new Exception("fatal error occured.");
            }
          }

          // ファイルサイズ取得
          String fileSize = connection.getHeaderField("Content-Length");
          if (fileSize != null) {
            fileSize = fileSize.replaceAll("size=", "").replaceAll("\"", "");
          } else {
            //TODO 画面更新後のダウンロードなどで、ダウンロードできなかった場合はエラー
            throw new Exception("ダウンロードに失敗しました。");
          }
          totalSize += Integer.parseInt(fileSize);
          System.out.println("### Downloading : " + fileSize + " Byte");
          System.out.println("### Total size : " + totalSize + " Byte");

          // 読込んだ総ファイルサイズをダイアログに反映
          iProgress.notifyProgress(sizeFormat.format(totalSize));
          Thread.sleep(100);

          // --- ファイル読込み ---
          byte[] buffer = new byte[1024];
          int length = 0;

          // --- ストリームからコンテンツデータを読込み ---
          while ( (length = inputStream.read(buffer)) >= 0 ) {
            fileOutputStream.write(buffer, 0, length);
          }

          // 次のFileStorageIdを取得
          fileStorageId = connection.getHeaderField("fileStorageId");

          // --- 次のFileStorageIdがなければ終了 ---
          if ( StringUtil.isNullString(fileStorageId) ) {
            System.out.println("###### Download end");
            break;
          }
          System.out.println("### Next FileStorage : true");
          System.out.println("###-----------------------------");
        }
        // --- HTTP送信 - 無効 ---
        else {
          // HTTPレスポンスメッセージを返却
          System.out.println("### Complete: POST NG");
          System.out.println("### ResponseCode:"+ responseCode);
          System.out.println("### ResponseMessage:"+ connection.getResponseMessage());
          System.out.println("###### Download error end");
          throw new Exception( "POST ERR (ResponseCode->" + responseCode + ") " + connection.getResponseMessage() );
        }

        // コネクションとストリームを閉じる
        try { 
          if (inputStream != null) inputStream.close();
          if (out != null) out.close();
          if (connection != null) connection.disconnect();
        } catch (Exception x) {
          x.printStackTrace();
          throw x;
        }
      }

    } finally {
      try {
        if (inputStream != null) inputStream.close();
        if (connection != null) connection.disconnect();
      } catch (Exception x) {}
    }
  }

}
