/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.file.client;

import java.io.DataOutputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import jp.littlesoft.file.common.Attributes;
import jp.littlesoft.file.common.SplitUploadParams;
import jp.littlesoft.file.common.IProgressListener;
import jp.littlesoft.soja.LSJson;
import jp.littlesoft.web.ActionPage;

/**
 * アップロード クライアントアプリケーションクラス<br /><br />
 * 
 * ファイルを分割してアップロードする、Javaクライアントアプリケーションクラスです。
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
public class SplitUploadClient {

  // 呼出し元JNLPファイルに定義されたJSONオブジェクトのパラメータ群
  private Attributes attributes;

  // アップロードファイルの入力ストリーム
  private FileInputStream fileInputStream;

  private String boundary = Long.toString(System.currentTimeMillis()) + Integer.toString(this.hashCode());

  // アップロード状況の書込みクラス
  private IProgressListener iProgress;


  /**
   * コンストラクタ。
   * 
   * @param attributes JSONオブジェクトのパラメータ群
   * @param fileInputStream ファイル入力ストリーム
   * @param iProgress ダウンロード状況の書込みクラス
   * @throws Exception 例外
   */
  public SplitUploadClient(Attributes attributes, FileInputStream fileInputStream, IProgressListener iProgress) throws Exception {
    this.attributes = attributes;
    this.fileInputStream = fileInputStream;
    this.iProgress = iProgress;
  }


  /**
   * アップロードを実行する。
   * 
   * @throws Exception 例外
   */
  public void execute() throws Exception {

    SplitUploadParams params = new SplitUploadParams();
    params.setAttributes(attributes);
    params.isSplit = true;
    params.fileName = (String)attributes.get("parentPath");
    params.fileName = (String)attributes.get("fileName");

    URL url = new URL(attributes.webappUrl);
    NumberFormat sizeFormat = NumberFormat.getNumberInstance();

    HttpURLConnection connection = null;
    InputStream inputStream = null;

    try {
      byte[] buff = new byte[SplitUploadParams.BUFFER_SIZE];
      int size = buff.length;
      int total = 0;
      int seq = 0;

      ArrayList<String> cookies = new ArrayList<String>();
      cookies.add("JSESSIONID=" + attributes.jsessionId);

      while(size == SplitUploadParams.BUFFER_SIZE) {
        size = fileInputStream.read(buff, 0, SplitUploadParams.BUFFER_SIZE);
        total += size;
        if (size < SplitUploadParams.BUFFER_SIZE) {
          byte[] finalBuff = new byte[size];
          System.arraycopy(buff, 0, finalBuff, 0, size);
          buff = finalBuff;
          params.isFinal = true;
        } else {
          params.isFinal = false;
        }
        connection = (HttpURLConnection) url.openConnection();
        connection.setDoOutput(true);
        connection.setRequestMethod("POST");

        if (cookies != null) {
          String cookie = "";
          for (String s : cookies) {
            cookie += s;
          }
          connection.setRequestProperty("Cookie", cookie);
        }
        // multipart/form-data形式
        connection.setRequestProperty("Content-type", "multipart/form-data;  boundary=" + boundary);
        connection.connect();

        params.splitSequence = seq++;
        params.individualDataSize = (long)buff.length;
        params.cumulativeDataSize = (long)total;
        DataOutputStream out = new DataOutputStream(connection.getOutputStream());
        postData(params, out, buff);
        out.flush();
        out.close();

        //ファイルサイズのアップロード済み
        iProgress.notifyProgress(sizeFormat.format(total));
        Thread.sleep(100);     

        // --- HTTP接続 ---
        int responseCode = connection.getResponseCode();

        // --- HTTP送信 - 有効 ---
        if (responseCode == HttpURLConnection.HTTP_OK) {
          System.out.println("###### Complete: POST OK");

          // レスポンス・ヘッダーに"error"があれば、Serverから返却されたエラーメッセージ
          String error = connection.getHeaderField("error");
          if ( "error".equals(error) ) {
            System.out.println("###### Upload error end");

            inputStream = connection.getInputStream();
            byte[] currentResponseByte = null;
            while (true) {
              byte[] b = new byte[1];
              int ret = inputStream.read(b, 0, 1);
              if (ret < 0) {
                break;
              }
              // byte配列の連結
              if (currentResponseByte == null) {
                currentResponseByte = b;
              } else {
                byte[] wk = currentResponseByte;
                currentResponseByte = new byte[wk.length + b.length];
                System.arraycopy(wk, 0, currentResponseByte, 0, wk.length);
                System.arraycopy(b, 0, currentResponseByte, wk.length, b.length);
              }
            }

            // エラー返却
            if (currentResponseByte != null) {
              String errorMessage = new String(currentResponseByte, "UTF-8");
              throw new Exception(errorMessage);
            } else {
              throw new Exception("fatal error occured.");
            }
          }

          // 最後の書き込みの場合、レスポンス・ヘッダーに"final"がなければ、サーバ側で受入れ処理が出来ていないためエラー
          String fin = connection.getHeaderField("final");
          if ( size < SplitUploadParams.BUFFER_SIZE && !"final".equals(fin) ) {
            //TODO 画面更新後のアップロードなどで、アップロードできなかった場合
            throw new Exception("アップロードに失敗しました。");
          }

          // --- 書き込み結果の標準出力 ---
          Map<String, java.util.List<String>> hFields = connection.getHeaderFields();
          for (String key : hFields.keySet()) {
            System.out.println("### key:" + key);
            for (String value : hFields.get(key)) {
              System.out.println("### value:" + value);
            }
          }
          if (hFields.get("Set-Cookie") != null) {
            List<String> cList = hFields.get("Set-Cookie");
            for (String s : cList) {
              if (!cookies.contains(s)) cookies.add(s);
            }
          }

        }
        // --- HTTP送信 - 無効 ---
        else {
          // HTTPレスポンスメッセージを返却
          System.out.println("### Complete: POST NG");
          System.out.println("### ResponseCode:"+ responseCode);
          System.out.println("### ResponseMessage:"+ connection.getResponseMessage());
          System.out.println("###### Upload error end");
          throw new Exception ("POST ERR (ResponseCode->" + responseCode + ") " + connection.getResponseMessage());
        }

        try {
          if (connection != null) connection.disconnect();
        } catch (Exception x) {
          x.printStackTrace();
          throw x;
        }
      }

    } finally {
      try {
        if (inputStream != null) inputStream.close();
        if (connection != null) connection.disconnect();
        if (fileInputStream != null) fileInputStream.close();
      } catch (Exception x) {}
    }
  }


  /**
   * multi-part送信用データ作成.
   * 
   * @param params
   * @param out
   * @param buff
   * @throws Exception 例外
   */
  public void postData(SplitUploadParams params, DataOutputStream out, byte[] buff) throws Exception {
    String jsonString = LSJson.serialize(params);
    System.out.println("@@@ " +jsonString);
    if (attributes.sequence != null) {
      writeFormData(out, ActionPage.PN_SEQ, attributes.sequence, boundary);
    }
    writeFormData(out, ActionPage.PN_SUBSESSION_ID, attributes.subsessionId, boundary);
    writeFormData(out, ActionPage.PN_ACTIONPAGE_ID, attributes.actionPageId, boundary);
    writeFormData(out, ActionPage.PN_METHOD, "onJavaAppCallBack", boundary);
    writeFormData(out, ActionPage.PN_APPLICATION_PARAMS, jsonString, boundary);
    //MultiPartのデータ作成：ファイル部分
    out.writeBytes("--" + boundary + "\r\n");
    out.writeBytes("Content-Disposition: form-data; ");
    out.writeBytes("name=\"" + "fn" + "\"; ");
    out.writeBytes("filename=\"" + java.net.URLEncoder.encode(params.fileName, SplitUploadParams.DEFAULT_ENC) + "\"\r\n");
    out.writeBytes("name=\"" + "fn" + "\"; ");
    out.writeBytes("Content-Type: application/octet-stream\r\n");
    out.writeBytes("\r\n");
    out.write(buff);
    out.writeBytes("\r\n");
    out.writeBytes("--" + boundary + "--");
  }


  /**
   * 
   * @param out
   * @param name
   * @param value
   * @param boundary
   * @throws IOException 例外
   */
  private void writeFormData(DataOutputStream out, String name, String value, String boundary) throws IOException {
    if (name==null || value == null) return;

    out.writeBytes("--" + boundary + "\r\n");
    out.writeBytes("Content-Disposition: form-data; ");
    out.writeBytes("name=\"" + java.net.URLEncoder.encode(name, SplitUploadParams.DEFAULT_ENC) + "\"\r\n");
    out.writeBytes("\r\n");
    out.writeBytes(value);
    out.writeBytes("\r\n");
  }
 
}
