/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.file.server;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.io.UnsupportedEncodingException;

import jp.littlesoft.file.common.SplitUploadParams;
import jp.littlesoft.soja.LSJson;

import org.apache.commons.fileupload.FileItem;
import org.json.JSONException;

/**
 * ファイルアップロードFileItemクラス。
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
abstract public class AbstractFileItem implements FileItem,Serializable {

  protected boolean temporary = true;
  protected String fieldName;
  protected String contentType;
  protected boolean isFormField;
  protected String fileName;
  
  protected SplitUploadParams params = null;
  protected AbstractFileItem prevItem = null;
  protected AbstractFileItem nextItem = null;

  /**
   * 指定されたファイルへ読み書きを行うファイルアイテムを生成する.
   * @param File
   */
  public AbstractFileItem(File file) {
    this.fileName = file.getName();
  }
  
  /**
   * FileItemFactoryから呼ばれるコンストラクタ
   *
   * @param fieldName フォームフィールド名
   * @param contentType フォームフィールドのコンテントタイプ
   * @param isFormField 単純なフォームフィールドを示す場合はtrue、それ以外の場合はfalse
   * @param fileName ブラウザまたは他のクライアントから渡されるアップロードファイル名
   */
  public AbstractFileItem(String fieldName, String contentType, boolean isFormField, String fileName) {
    this.fieldName   = fieldName;
    this.contentType = contentType;
    this.isFormField = isFormField;
    this.fileName    = fileName;
  }


  /**
   * 関係する一時ディスク領域も含むストレージ上のファイルアイテムを削除します。<br/>
   * FileItem インスタンスがガベージコレクションにかかった時にこのストレージは削除されますが、<br/>
   * このメソッドは早く確実に削除を実施し、システムリソースを保護します。<br/>
   *
   * @param file アップロードアイテムの出力先となる File
   * @throws Exception エラーが発生した場合
   */
  abstract public void delete();

  /**
   * ファイルアイテムの内容をバイト配列で返します。
   *
   * @return ファイルアイテムの内容を示すバイト配列
   */
  abstract public byte[] get();


  /**
   * ブラウザによって付加されたコンテントタイプを返します。 定義されていない場合には null を返します。 
   *
   * @return ブラウザによって付加されたコンテントタイプ。定義されていない場合には null
   */
  public String getContentType() {
    return contentType;
  }


  /**
   * ファイルアイテムに対応したマルチパートフォームフィールド名を返します。
   *
   * @return フォームフィールド名
   */
  public String getFieldName() {
    return fieldName;
  }


  /**
   * ファイルの内容を取得するための InputStream を返します.<br/>
   *
   * @return ファイルの内容を取得するための InputStream
   * @throws IOException エラーが発生した場合
   */
  abstract public InputStream getInputStream() throws IOException;


  /**
   * ブラウザ(または他のクライアントソフトウェア)によって渡された クライアントの<br/>
   * ファイルシステム上でのオリジナルのファイル名を返します。<br/>
   * 大体の場合、これはパス情報を除いた元のファイル名になります。<br/>
   * しかし、Operaブラウザ等のいくつかのクライアントではパス情報が含まれます。<br/>
   *
   * @return クライアントのファイルシステム上でのオリジナルのファイル名
   */
  public String getName() {
    return fileName;
  }


  /**
   * ファイルの内容を保持するための OutputStream を返します。
   *
   * @return ファイルの内容を保持するための OutputStream 
   * @throws IOException エラーが発生した場合
   */
  abstract public OutputStream getOutputStream() throws IOException;

  /**
   * ファイルアイテムのサイズを返します。
   *
   * @return ファイルアイテムのバイト単位のサイズ
   */
  public long getSize() {
    return get().length;
  }


  /**
   * デフォルトのキャラクタエンコーディングを使用してファイルアイテムの内容の文字列表現を返します。<br/>
   * このメソッドはアイテムの内容を取得するために get() メソッドを使用します。
   *
   * @return アイテムの内容の文字列表現
   */
  public String getString() {
    return new String(get());
  }


  /**
   * 指定されたエンコーディングを使用してファイルアイテムの内容の文字列表現を返します。<br/>
   * このメソッドはアイテムの内容を取得するために get() メソッドを使用します。
   *
   * @param encoding 使用するキャラクタエンコーディング
   * @return アイテムの内容の文字列表現
   * @throws UnsupportedEncodingException 指定されたエンコーディングが利用できなかった場合
   */
  public String getString(String encode) throws UnsupportedEncodingException {
    return new String(get(), encode);
  }


  /**
   * FileItem インスタンスが単純なフォームフィールドを示すかどうかを判断します。 
   * 
   * @return インスタンスが単純なフォームフィールドを示す場合はtrue、インスタンスがアップロードされたファイルを示す場合はfalse
   */
  public boolean isFormField() {
    return isFormField;
  }


  /**
   * ファイルの内容をメモリ上から読み込むことができるかどうかを示します。 
   *
   * @return ファイルの内容をメモリ上から読み込むことができる場合はtrue、その他の場合はfalse
   */
  abstract public boolean isInMemory() ;


  /**
   * このファイルアイテムを参照するためのフィールド名を設定します。
   *
   * @param name フォームフィールド名
   */
  public void setFieldName(String fieldName) {
    this.fieldName = fieldName;
  }


  /**
   * FileItem インスタンスが単純なフォームフィールドを示すかどうかを設定します。
   *
   * @param state インスタンスが単純なフォームフィールドを示す場合はtrue、インスタンスがアップロードされたファイルを示す場合はfalse
   */
  public void setFormField(boolean isFormField) {
    this.isFormField = isFormField;
  }

  /**
   * 分割されたファイルを先頭からすべて読み取る入力ストリームを取得します.<br>
   * @return InputStream;
   */
  abstract public InputStream getFullReadInputStream(File file)  throws Exception ;
  
  
  /**
   * fileで指定された永続化ストレージにデータを書き出します.<br/>
   * 分割されたファイルの場合、先頭アイテムをwriteすれば残りも同時にwriteされます.<br/>
   * 当メソッドによりファイルの状態はisTemporary:falseとなります.
   * また、Temporaryファイル等の後始末も行います.
   *
   * @param file アップロードアイテムの出力先となる File
   * @throws Exception エラーが発生した場合
   */
  abstract public void write(File file) throws Exception;
  
  /**
   * アップロードアイテムが生成された時点ではisTemporary:trueです.
   * isTemporary:trueの場合、FileManagerによって当該セッション情報が破棄されると同時にファイルも削除されます
   * @return isTemporary
   */
  public boolean isTemporary() {
    return temporary;
  }

  /**
   * アップロードアイテムが生成された時点ではisTemporary:trueです.
   * isTemporary:trueの場合、FileManagerによって当該セッション情報が破棄されると同時にファイルも削除されます.
   * isTemporary:falseに設定すると上記削除対象からは除外されますが、永続的なストレージへの移動と一時ファイルの消去は個別に実装する必要があります.
   * @return isTemporary
   */
  public void setTemporary(boolean temporary) {
    this.temporary = temporary;
    AbstractFileItem item = this;
    while(item.nextItem!=null) {
      item.nextItem.temporary = temporary;
      item = item.nextItem;
    }
  }
  
  /**
   * 分割時の累積データサイズを取得する。
   * 
   * @return
   */
  public long getTotalSize() {
    AbstractFileItem item = this;
    while(item.nextItem!=null) {
      item = item.nextItem;
    }
    return item.params.cumulativeDataSize;
  }
  
  @Override
  public String toString() {
    try {
      return "{fileName:" + fileName + 
             ",fieldName:" + fieldName +
             ",temporary:" + temporary +
             ( (prevItem!=null) ? ",prevItem:" + prevItem.fieldName : "" ) +
             ( (nextItem!=null) ? ",nextItem:" + nextItem.fieldName : "" ) +
             ( (params!=null) ? ",params:" + LSJson.serialize(params) : "" ) +
             " }";
    } catch (JSONException e) {
      e.printStackTrace();
      return null;
    }
  }
  
  /**
   * ファイルを識別する一意のIDを取得する.<br.>
   * 通常のDiskFileであればファイルパス、GAEではentityのKeyを文字列化した値
   * @return id
   */
  abstract public String getId();


  /**
   * 前のFileItemオブジェクトを取得する。
   * 
   * @return 前のFileItemオブジェクト
   */
  public AbstractFileItem getPrevItem() {
    return prevItem;
  }

  /**
   * 次のFileItemオブジェクトを取得する。
   * 
   * @return 次のFileItemオブジェクト
   */
  public AbstractFileItem getNextItem() {
    return nextItem;
  }


  public SplitUploadParams getParams() {
    return params;
  }


  protected void setParams(SplitUploadParams params) {
    this.params = params;
  }


  protected void setPrevItem(AbstractFileItem prevItem) {
    this.prevItem = prevItem;
  }


  protected void setNextItem(AbstractFileItem nextItem) {
    this.nextItem = nextItem;
  }
}
