/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.sql;

import java.io.ByteArrayInputStream;
import java.io.CharArrayReader;
import java.io.Serializable;
import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.List;
import java.util.Properties;
import java.util.StringTokenizer;

import jp.littlesoft.data.Column;
import jp.littlesoft.data.Params;
import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.data.exception.InvalidColumnException;
import jp.littlesoft.data.validator.Validator;
import jp.littlesoft.data.validator.StringLengthValidator;
import jp.littlesoft.soja.LSJson;
import jp.littlesoft.sql.exception.LSSQLException;
import jp.littlesoft.util.Const;
import jp.littlesoft.util.DataUtil;
import jp.littlesoft.util.StringUtil;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.json.JSONException;

/**
 * SQLアダプタクラス。<br/>
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public class SQLAdaptor implements Serializable {
  
  /**
   * コネクションオブジェクトです。<br/>
   */
  protected transient java.sql.Connection jdbcConnection = null;
  
  private String dbUrl = null;
  private String dbUserName = null;
  private String dbPassword = null;
  private String driverName = null;
  private Properties connectProps = null;
  private boolean connectionOwner = false;
  
  /**
   * ログ出力オブジェクトです。<br/>
   */
  protected static final transient Logger logger = LogManager.getLogger(SQLDataResolverImpl.class.getName());
  
  /**
   * Validatorのインスタンスです。<br/><br/>
   * 
   * PostgreSQL 7.2以上の場合に設定する。<br/>
   */
  private Validator validator = null;
  
  /**
   * 元プロセスを識別するための識別子です。<br/>
   */
  private String clientId = null;
  
  /**
   * クエリーのタイムアウト時間を表します。<br/>
   */
  protected int queryTimeout = -1;
  
  /**
   * DBサーバのタイプです。<br/>
   */
  private SQLServerType sqlServerType;
  
  boolean autoCommitAvailable = true;
  
  /**
   * getPrimaryKeyを利用可能かどうかを表します。<br/><br/>
   * 
   * True：利用可、False：利用不可<br/>
   */
  protected boolean getPrimaryKeyAvailable = true;
  
  private boolean searchMetaDataByUpperCase = false;
  private boolean searchMetaDataByLowerCase = false;
  private boolean searchMetaDataWithinSchema = true;
  
  
  /**
   * コンストラクタ。
   */
  public SQLAdaptor() {}

  /**
   * JDBCのConnectionオブジェクトを取得します。<br/>
   *
   * @return  JDBCのConnectionオブジェクト
   */
  public java.sql.Connection getConnection() {
    return jdbcConnection;
  }

  /**
   * JDBCのConnectionオブジェクトを設定します。<br/><br/>
   *
   * APサーバーのコネクションを利用する場合は、このメソッドを使用して、
   * APサーバーから取得したjava.sql.Connectionを設定してください。<br/>
   *
   * @param  newJdbcConnection  JDBCのConnection
   * @exception  jp.littlesoft.sql.NJSQLException  例外発生時
   */
  public void setConnection(java.sql.Connection newJdbcConnection) throws LSSQLException {
    if (jdbcConnection==newJdbcConnection) return;
    if (isConnectionOwner() && isConnecting()) close();
    jdbcConnection = null;
    jdbcConnection = newJdbcConnection;
    connectionOwner = false;
    try {
      this.setSQLServerType(SQLServerType.checkSQLServerType(newJdbcConnection.getMetaData().getDatabaseProductName()));      
    } catch (Exception x) {
      throw new LSSQLException(x); 
    }
    setColumnValidatorByConnection(jdbcConnection);
  }

  /**
   * ターゲットDBMSのURLを設定します。<br/>
   *
   * @param  newDbUrl  DBのURL
   */
  public void setDBURL(String newDbUrl) {
    dbUrl = newDbUrl;
  }

  /**
   * ターゲットDBMSのURLを取得します。<br/>
   *
   * @return  DBのURL
   */
  public String getDBURL() {
    return dbUrl;
  }

  /**
   * ターゲットDBMSのユーザー名を設定します。<br/>
   *
   * @param  newDbUserName  DBの接続ユーザー名
   */
  public void setDBUserName(String newDbUserName) {
    dbUserName = newDbUserName;
  }

  /**
   * ターゲットDBMSのユーザー名を取得します。<br/>
   *
   * @return  DBの接続ユーザー名
   */
  public String getDBUserName() {
    // ユーザ名がなしの場合に、接続プロパティを指定してconnect
    // しようとするとNPEが発生する障害を修正
    if (this.dbUserName == null) {
      return Const.NULLSTRING;
    }
    return this.dbUserName;
  }

  /**
   * ターゲットDBMSのパスワードを設定します。<br/>
   *
   * @param  newDbPassword  DBのパスワード
   */
  public void setDBPassword(String newDbPassword) {
    dbPassword = newDbPassword;
  }

  /**
   * ターゲットDBMSのパスワードを取得します。<br/>
   *
   * @return  DBのパスワード
   */
  public String getDBPassword() {
    // パスワードがなしの場合に、接続プロパティを指定してconnect
  	  // しようとするとNPEが発生する障害を修正
  	  if (this.dbPassword == null) {
  		    return Const.NULLSTRING;
  	  }
    return this.dbPassword;
  }

  /**
   * ターゲットDBMSのJDBCドライバークラス名を設定します。<br/>
   *
   * @param  newDriverName  JDBCドライバーのクラス名
   */
  public void setJDBCDriverClassName(String newDriverName) {
    this.driverName = newDriverName;
    this.setSQLServerType(SQLServerType.checkSQLServerType(this.driverName));
  }

  /**
   * ターゲットDBMSのJDBCドライバークラス名を取得します。<br/>
   *
   * @return  JDBCドライバーのクラス名
   */
  public String getJDBCDriverClassName() {
    return driverName;
  }

  /**
   * 接続時に渡すプロパティ変数を設定します。<br/><br/>
   *
   * プロパティには "user", "password" は含めないでください。<br/>
   *
   * @param  p 接続時に渡すプロパティ
   */
  public void setConnectProperty(Properties p) {
    connectProps = p;
  }

  /**
   * 接続時に渡すプロパティ変数を取得します。<br/><br/>
   *
   * プロパティには user, password は含まれません。<br/>
   *
   * @return  接続時に渡すプロパティ
   */
  public Properties getConnectProperty() {
    return connectProps;
  }

  /**
   * データベースに接続します。<br/><br/>
   *
   * 当メソッドはsetJDBCConnectionによる外部コネクション利用と排他的です。<br/>
   *
   * @param  newDriverName  JDBCドライバーのクラス名
   * @param  newDbUrl       DBのURL
   * @param  newDbUserName  DBのユーザー名
   * @param  newDbPassword  DBのパスワード
   * @param  p              接続時に渡すプロパティ("user", "password"は除く）
   * @exception  jp.littlesoft.sql.NJSQLException  例外発生時
   */
  public void connect(String newDriverName, String newDbUrl, String newDbUserName, String newDbPassword, Properties p) throws LSSQLException {
    setJDBCDriverClassName(newDriverName);
    setDBURL(newDbUrl);
    setDBUserName(newDbUserName);
    setDBPassword(newDbPassword);
    setConnectProperty(p);
    connect();
  }

  /**
   * データベースに接続します。<br/><br/>
   *
   * 当メソッドはsetJDBCConnectionによる外部コネクション利用と排他的です。<br/>
   *
   * @exception  jp.littlesoft.sql.NJSQLException  例外発生時
   */
  public void connect() throws LSSQLException {
    if (isConnectionOwner() && isConnecting()) close();
    jdbcConnection = null;
    if (StringUtil.isNullString(getJDBCDriverClassName()))
      throw new LSSQLException(null, null, "JDBC_DRIVER_CLASSNAME_NOT_SPECIFIED", null);
    if (StringUtil.isNullString(getDBURL()))
      throw new LSSQLException(null, null, "DB_URL_NOT_SPECIFIED", null);
  
    try {
      Class.forName(getJDBCDriverClassName());
      if (this.getConnectProperty() == null || this.getConnectProperty().isEmpty()) {
        // 接続時に渡す情報としてユーザ名とパスワードしかない場合
        jdbcConnection = DriverManager.getConnection(getDBURL(), getDBUserName(), getDBPassword());
      } else {
          // 接続時に渡す情報としてユーザ名とパスワード以外にも何かある場合
          Properties p = new Properties();
          p.putAll(this.getConnectProperty());
          p.setProperty("user", this.getDBUserName());
          p.setProperty("password", this.getDBPassword());
          jdbcConnection = DriverManager.getConnection(getDBURL(), p);
      }
      connectionOwner = true;
      setAutoCommit(false);
    
      this.setColumnValidatorByConnection(jdbcConnection);
    } catch (ClassNotFoundException c) {
      LSSQLException ex = new LSSQLException(c, null, "ERROR_LOAD_DRIVER",
          new String[]{getJDBCDriverClassName()});
      throw ex;
    } catch (Throwable t) {
      // 本来はSQLExceptionをcatchすべきだが、MySQLのJDBC DriverがThrowableを投げることがあるので(バグ?)
      // 仕方なく、Throwableをcatchする。
      throw new LSSQLException(t, null, "ERROR_CONNECT_DB",
          new String[]{getDBURL(), getDBUserName()});
    }
  }

  /**
   * データベース接続を切断します。<br/>
   *
   * @exception  jp.littlesoft.sql.NJSQLException  例外発生時
   */
  public void close() throws LSSQLException {
    try {
      if (isConnecting()) jdbcConnection.close();
      connectionOwner = false;
    } catch (SQLException x) {
      throw new LSSQLException(x);
    }
  }

  /**
   * インスタンスが破棄されるときの処理です。<br/><br/>
   *
   * JDBCのConnectionオブジェクトが残っていれば切断します。<br/>
   *
   * @exception  java.lang.Throwable  例外発生時
   */
  protected void finalize() throws Throwable {
    if (isConnectionOwner() && isConnecting()) close();
    jdbcConnection = null;
    super.finalize();
  }

  /**
   * データベースに接続しているかどうか調べます。<br/>
   *
   * @return True：接続、False：未接続
   */
  public boolean isConnecting() {
    try {
      return (jdbcConnection!=null && !jdbcConnection.isClosed());
    } catch (SQLException x) {
      return false;
    }
  }

  /**
   * 現在のコネクションは当オブジェクト内部でされたものかどうか調べます。<br/>
   *
   * @return True：内部、False：外部
   * @see #setJDBCConnection
   */
  public boolean isConnectionOwner() {
    return connectionOwner;
  }

  /**
   * 自動的なコミットを行うかどうか設定します。<br/><br/>
   * 
   * ※JDBCコネクションが確立した状態でないと設定できません。<br/>
   *
   * @param  newAutoCommit True：自動的コミットを行う、False：自動コミットを行わない
   * @exception  jp.littlesoft.sql.NJSQLException  例外発生時
   */
  public void setAutoCommit(boolean newAutoCommit) throws LSSQLException {
    if (!this.autoCommitAvailable) {
      // setAutoCommitをサポートしていない場合は何もしない。
      return;
    }
    checkConnection();
    try {
        if(jdbcConnection.getAutoCommit()!= newAutoCommit)
          jdbcConnection.setAutoCommit(newAutoCommit);
    } catch (SQLException x) {
      throw new LSSQLException(x);
    }
  }

  /**
   * 自動コミットを行うかどうか調べます。<br/>
   *
   * @return True：自動的コミットを行う、False：自動コミットを行わない
   * @exception  jp.littlesoft.sql.exception.LSSQLException  例外発生時
   */
  public boolean isAutoCommit() throws LSSQLException {
    if (!this.autoCommitAvailable) {
      // setAutoCommitをサポートしていない場合はtrueを返す。
      return true;
    }
    checkConnection();
    try {
      return jdbcConnection.getAutoCommit();
    } catch (SQLException x) {
      throw new LSSQLException(x);
    }
  }

  /**
   * JDBCのConnectionを保持しているかどうか調べます。<br/><br/>
   *
   * 保持していない場合、例外が発生します。<br/>
   *
   * @exception  jp.littlesoft.sql.exception.LSSQLException  接続を保持していない場合
   */
  protected void checkConnection() throws LSSQLException {
    if (jdbcConnection==null)
      throw new LSSQLException(null, null, "ERROR_NULL_CONNECTION", null);
  }

  /**
   * SQL文をデータベースに送るためのjava.sql.Statementオブジェクトを生成します。<br/><br/>
   *
   * 内部で呼び出しています。<br/>
   * 開発者がこのメソッドを呼び出す必要はありません。<br/>
   *
   * @return  Statementオブジェクト
   * @exception  jp.littlesoft.sql.exception.LSSQLException  例外発生時
   *
   * @see java.sql.Statement
   */
  public synchronized Statement createStatement() throws LSSQLException {
    checkConnection();
    if (!isConnecting()) connect();
    try {
      Statement st = jdbcConnection.createStatement();
      // クエリーのタイムアウト時間を設定する。
      setQueryTimeoutToStatement(st, getQueryTimeout());
      //JDBCコネクションのURL情報を取得する。
      //setJDBCURLfromConn();
      return st;
    } catch (SQLException x) {
      throw new LSSQLException(x);
    }
  }

  /**
   * パラメータ付きSQL文をデータベースに送るためのjava.sql.PreparedStatementオブジェクトを生成します。<br/><br/>
   *
   * SQL文中のプレースホルダは"?"で指定します。<br/><br/>
   *
   * 内部で呼び出しています。<br/>
   * 開発者がこのメソッドを呼び出す必要はありません。<br/>
   *
   * @param   sqlText  SQL文
   * @return  PreparedStatementオブジェクト
   * @exception  jp.littlesoft.sql.exception.LSSQLException  例外発生時
   * @exception  jp.littlesoft.sql.DataException  例外発生時
   *
   * @see java.sql.PreparedStatement
   *
   */
  public synchronized PreparedStatement preparedStatement(String sqlText) throws LSSQLException, DataException {
    checkConnection();
    if (!isConnecting()) connect();
    try {
      PreparedStatement st = jdbcConnection.prepareStatement(sqlText);
      // クエリーのタイムアウト時間を設定する。
      setQueryTimeoutToStatement(st, getQueryTimeout());
      // JDBCコネクションのURL情報を取得する。
      //setJDBCURLfromConn();
      return st;
    } catch (SQLException x) {
      logger.error("preparedStatement :" + sqlText);
      throw new LSSQLException(x);
    }
  }

  /**
   * データベースのストアドプロシージャを呼び出すためのjava.sql.CallableStatementオブジェクトを生成します。<br/><br/>
   *
   * SQL文中のプレースホルダは"?"で指定します。<br/>
   * 実行するステートメントは次の様な形式で設定します。<br/>
   * "{call &lt;stored procedure name&gt; (?,?,... )}" <br/>
   * "? = {call &lt;stored procedure name&gt; (?,?,... )}"<br/><br/>
   *
   * 内部で呼び出しています。<br/>
   * 開発者がこのメソッドを呼び出す必要はありません。<br/><br/>
   *
   * @param   sqlText  SQL文
   * @return  CallableStatementオブジェクト
   * @exception  jp.littlesoft.sql.exception.LSSQLException  例外発生時
   * @exception  jp.littlesoft.sql.DataException  例外発生時
   *
   * @see java.sql.CallableStatement
   */
  public synchronized CallableStatement prepareCall(String sqlText) throws LSSQLException, DataException {
    checkConnection();
    if (!isConnecting()) connect();
    try {
      CallableStatement st = jdbcConnection.prepareCall(sqlText);
      // クエリーのタイムアウト時間を設定する。
      setQueryTimeoutToStatement(st, getQueryTimeout());
      //JDBCコネクションのURL情報を取得する。
      //setJDBCURLfromConn();
      return st;
    } catch (SQLException x) {
      throw new LSSQLException(x);
    }
  }

  /**
   * このJDBCAdaptorオブジェクトの識別子を設定します。<br/><br/>
   * 
   * この識別子はログ出力に利用されます。<br/>
   *
   * @param clientId このJDBCAdaptorの識別子
   */
  public void setClientId(String clientId) {
    this.clientId = clientId;
  }

  /**
   * このJDBCAdaptorオブジェクトの識別子を取得します。<br/>
   *
   * @return このJDBCAdaptorの識別子
   */
  public String getClientId() {
    return this.getClientId(false);
  }

  /**
   * このJDBCAdaptorオブジェクトの識別子を取得します。<br/><br/>
   * 
   * flagがtrueの場合、ClientIdが設定されていない場合は、Nameプロパティの値を返します。<br/>
   * 
   * @param flag ClientIdが設定されていない場合に、Nameプロパティの値を返すか?
   * @return このJDBCAdaptorの識別子
   */
  public String getClientId(boolean flag) {
    if (this.clientId == null) {
      if (flag)
        return this.getClass().getName();
      else
        return Const.NULLSTRING;
    } else {
      return clientId;
    }
  }

  /**
   * SQLServerTypeの設定します。<br/><br/>
   * 
   * SQLServerTypeクラスの定数を使用して、DBMSの種類を設定します。<br/>
   * この値により、JDBCAdaptorはDBMS毎の差異を吸収します。<br/>
   * この差異とはJDBCAdaptorHelperが吸収している粒度の大きい差異ではなく
   * SQLの構文等非常に繊細な部分で発生する差異です。<br/><br/>
   *
   * なお、SQLServerTypeはJDBCDriverClassNameを設定することで自動的に設定されますので
   * 通常はこのメソッドを使用する必要はありません。<br/>
   *
   * @param sqlServerType DBMSの種類を示す整数
   * @see SQLServerType
   */
  public void setSQLServerType(SQLServerType sqlServerType) {
    this.sqlServerType = sqlServerType;
    SQLServerType.setupDefault(this);
    this.resetBySQLServerType(sqlServerType);
  }

  /**
   * SQLServerTypeを取得します。<br/>
   * 
   * @return DBMSの種類を示す整数
   * @see SQLServerType
   */
  public SQLServerType getSQLServerType() {
    return sqlServerType;
  }

  // 文字型検証クラスを設定します。
  // 接続先データベースがPostgres 7.2以上の場合、StringLengthValidatorクラスを設定する。
  // それ以外の場合は null を設定する。
  // connect()の最後、またはsetJdbcConnection()の最後に行う。

  /**
   * Sets String Validator Class. <br/>
   */
  void setColumnValidatorByConnection(Connection jdbcConnection) {
    if (jdbcConnection == null) {
      this.setColumnValidator(null);
    } else {
      try {
        this.setColumnValidator(createStringValidator(
              jdbcConnection.getMetaData().getDatabaseProductName(),
              jdbcConnection.getMetaData().getDatabaseProductVersion()));
      } catch (Throwable t) {
        t.printStackTrace();
      }
    }
  }

  /**
   * Validatorを設定します。
   * 
   * @param validator Validatorのインスタンス
   */
  public void setColumnValidator(Validator validator) {
    this.validator = validator;
  }
  
  /**
   * Creates String Validator Instance. <br/>
   */
  private static Validator createStringValidator(String databaseProductName, String databaseProductVersion) {
    // データベースのメタデータから接続先データベース製品名とバージョンを取得する。
    if (! "PostgreSQL".equals(databaseProductName)) return null;
    
    if (StringUtil.isNullString(databaseProductVersion)) return null;
    
    // Version x.y.z の x.y の部分を取り出す。
    StringTokenizer st = new StringTokenizer(databaseProductVersion, ".");
    int majorVersion = 0;
    int minorVersion = 0;
    if (st.hasMoreTokens()) majorVersion = Integer.parseInt(st.nextToken());
    if (st.hasMoreTokens()) minorVersion = Integer.parseInt(st.nextToken());
    
    if (majorVersion > 7 || (majorVersion == 7 && minorVersion > 1)) {
      // 7.2以上の場合のみ、VARCHAR(n)をバイトではなく文字数で数える。
      return new StringLengthValidator();
    }
    return null;
  }

  /**
   * 文字型カラム用の検証クラスのインスタンスを取得します。<br/>
   * 
   * @return 文字型カラム用の検証クラスのインスタンス
   */
  public Validator getColumnValidator() {
    return this.validator;
  }

  /**
   * クエリーのタイムアウト時間を設定します。<br/><br/>
   *
   * タイムアウト時間には、0以上の値を設定してください。<br/>
   * ０は無制限を意味します。<br/>
   * マイナスの値を設定した場合には現在設定されているタイムアウト時間をそのまま利用します。<br/><br/>
   * 
   * ※注意　使用するJDBCドライバによっては有効にならないこともあります。<br/>
   * (クエリー実行前にjava.sql.Statement#setQueryTimeoutメソッドを呼び出しています。)<br/><br/>
   * 
   * ※注意　使用するJDBCドライバによっては有効にならないこともあります。<br/>
   * (クエリー実行前にjava.sql.Statement#setQueryTimeoutメソッドを呼び出しています。)<br/>
   * 
   * @param sec クエリーのタイムアウト時間（単位：秒） デフォルト：-1（現在の値をそのまま利用）
   * @see java.sql.Statement#setQueryTimeout
   */
  public void setQueryTimeout(int sec) {
    queryTimeout = sec;
  }

  /**
   * クエリーのタイムアウト時間を取得します。<br/>
   * 
   * @return クエリーのタイムアウト時間（単位：秒）
   */
  public int getQueryTimeout() {
    return queryTimeout;
  }

  /**
   * クエリーのタイムアウト時間を設定します。<br/><br/>
   *
   * ただし、0以上の場合のみです。<br/>
   *
   * @param queryTimeout タイムアウト時間（マイナス値なら設定しない）
   */
  void setQueryTimeoutToStatement(Statement st, int queryTimeout) throws java.sql.SQLException {
    if (queryTimeout >= 0) st.setQueryTimeout(queryTimeout);
  }
  
  /**
   * SQLServerTypeに応じて、AutoCommit, getPrimaryKeyの使用可否を決定します。
   * 
   * @param sqlServerType DBサーバのタイプ
   */
  void resetBySQLServerType(SQLServerType sqlServerType) {
    switch (sqlServerType) {
      case INFORMIX:
        this.setAutoCommitAvailable(false);
        this.getPrimaryKeyAvailable = true; // OK?
        break;
      case ODBC:
        this.setAutoCommitAvailable(false);
        this.getPrimaryKeyAvailable = false;
        break;
      default:
        this.setAutoCommitAvailable(true);
        this.getPrimaryKeyAvailable = true;
        break;
    }
  }
  
  private void setAutoCommitAvailable(boolean autoCommitAvailable) {
    this.autoCommitAvailable = autoCommitAvailable;
  }

  /**
   * 指定されたテーブルの主キー列を取得します。<br/>
   * 
   * @param  catalog    カタログ名
   * @param  schema     スキーマ名
   * @param  tableName  テーブル名
   * @return 主キーを構成するカラム名を取得のリスト
   * @exception SQLException SQL例外発生時
   * @exception DataException 例外発生時
   */
  public List<String> getPrimaryKeyColumns(String catalog, String schema, String tableName) throws SQLException, DataException {
    if (!this.getPrimaryKeyAvailable) {
      // getPrimaryKeyを利用できない場合はnullを返す。
      return null;
    }
        
    if (schema==null || Const.NULLSTRING.equals(schema)) {
      schema = jdbcConnection.getMetaData().getUserName();
    }
    String tables = tableName;
    String schemas = null;
    // searchMetaDataWithinSchemaプロパティがfalseの場合、スキーマパターンはnullとする
    if (isSearchMetaDataWithinSchema() && StringUtil.isNotNullString(schema)) {
      schemas = schema;
    }
      
    //とりあえずORACLEは大文字でないとヒットしない。。。
    //他のDBMSは調査すべし。
    if (isSearchMetaDataByUpperCase()) {
      if (schemas != null) schemas = schemas.toUpperCase();
      tables = tableName.toUpperCase();
    } else if (isSearchMetaDataByLowerCase()) {
      if (schemas != null) schemas = schemas.toLowerCase();
      tables = tableName.toLowerCase();
    }
      
    if (StringUtil.isNullString(schemas) || StringUtil.isNullString(catalog)) {
      catalog = null; 
    }
      
    java.util.ArrayList<String> result = new java.util.ArrayList<String>();
    ResultSet rs = jdbcConnection.getMetaData().getPrimaryKeys(catalog, schemas, tables);
    while (rs.next()) {
      result.add(rs.getString(4));
    }
    rs.close();
    return result;
  }

  /**
   * メタデータ検索時に、更新テーブル名を大文字に変換するかどうかを設定します。<br/>
   *
   * @param  newSearchMetaDataByUpperCase  大文字変換するかどうか
   */
  public void setSearchMetaDataByUpperCase(boolean newSearchMetaDataByUpperCase) {
    searchMetaDataByUpperCase = newSearchMetaDataByUpperCase;
  }

  /**
   * メタデータ検索時に、更新テーブル名を大文字に変換するかどうかを取得します。<br/>
   *
   * @return  大文字変換するかどうか
   */
  public boolean isSearchMetaDataByUpperCase() {
    return searchMetaDataByUpperCase;
  }

  /**
   * メタデータ検索時に、更新テーブル名を英小文字に変換するかどうかを設定します。<br/>
   *
   * @param  newSearchMetaDataByLowerCase  英小文字変換するかどうか
   */
  public void setSearchMetaDataByLowerCase(boolean newSearchMetaDataByLowerCase) {
    searchMetaDataByLowerCase = newSearchMetaDataByLowerCase;
  }

  /**
   * メタデータ検索時に、更新テーブル名を英小文字に変換するかどうかを取得します。<br/>
   *
   * @return  英小文字変換するかどうか
   */
  public boolean isSearchMetaDataByLowerCase() {
    return searchMetaDataByLowerCase;
  }

  /**
   * メタデータ検索時に、スキーマ名を使用するかどうかを設定します。<br/>
   *
   * @param  newSearchMetaDataWithinSchema  メタデータ検索時に、スキーマ名を使用するかどうか
   */
  public void setSearchMetaDataWithinSchema(boolean newSearchMetaDataWithinSchema) {
    searchMetaDataWithinSchema = newSearchMetaDataWithinSchema;
  }

  /**
   * メタデータ検索時に、スキーマ名を使用するかどうかを取得します。<br/>
   *
   * @return  メタデータ検索時に、スキーマ名を使用するかどうか
   */
  public boolean isSearchMetaDataWithinSchema() {
    return searchMetaDataWithinSchema;
  }

  /**
   * PreparedStatement内のパラメータにParamsのパラメータの値を割り当てます。<br/>
   *
   * @param  pstmt   PreparedStatementオブジェクト
   * @param  params DataMapオブジェクト
   * @exception  jp.littlesoft.sql.exception.LSSQLException  例外発生時
   * @exception  jp.littlesoft.data.DataException  例外発生時
   */
  public void bindParamValues(PreparedStatement pstmt, Params<?> params) throws LSSQLException, DataException{
    try {
      int[] paramPos = null;
      pstmt.clearParameters();
      for (int i=0; i<params.getColumnCount(); i++ ) {
        paramPos = params.getColumn(i).getParamPos();
        if (paramPos==null || paramPos.length==0  || params.getColumn(i).getParamType()==Column.ParamType.OUT) continue;
        for (int p=0; p<paramPos.length; p++) {
          if (params.isNull(i)) {
            if (params.getColumn(i).getType()==Column.Type.UNKNOWN) {
              pstmt.setNull(paramPos[p], Column.mapToSql(Column.Type.STRING));
            } else {
              pstmt.setNull(paramPos[p], params.getColumn(i).getSqlType());
            }
            continue;
          }
            //JDBC 3.0に対応しているJDBCドライバーならバインド変数の型を取得できる可能性がある。
            //Column.TypeがUNKNOWNなら、その型で設定してもよいかもしれない。
            switch (params.getColumn(i).getType()) {
            case BIGDECIMAL:
              pstmt.setBigDecimal(paramPos[p], params.getBigDecimal(i));
              break;
            case BOOLEAN:
              pstmt.setBoolean(paramPos[p], params.getBoolean(i));
              break;
            case BYTE:
              pstmt.setByte(paramPos[p], params.getByte(i));
              break;
            case BINARY:
              pstmt.setBytes(paramPos[p], params.getBinary(i));
              break;
            case LONGBINARY:
            case BLOB: // JDBC2.1の仕様によれば、BinaryStreamで処理できる？
              byte[] longBinary = params.getBinary(i);
              ByteArrayInputStream in = new ByteArrayInputStream(longBinary);
              pstmt.setBinaryStream(paramPos[p], in, longBinary.length);
              break;
            case DATE:
              pstmt.setDate(paramPos[p], params.getDate(i));
              break;
            case DOUBLE:
              pstmt.setDouble(paramPos[p], params.getDouble(i));
              break;
            case FLOAT:
              pstmt.setFloat(paramPos[p], params.getFloat(i));
              break;
            case INT:
              pstmt.setInt(paramPos[p], params.getInt(i));
              break;
            case LONG:
              pstmt.setLong(paramPos[p], params.getLong(i));
              break;
            case SHORT:
              pstmt.setShort(paramPos[p], params.getShort(i));
              break;
            case STRING:
              pstmt.setString(paramPos[p], DataUtil.replaceToDB(params.getString(i)));
              break;
            case TIME:
              pstmt.setTime(paramPos[p], params.getTime(i));
              break;
            case TIMESTAMP:
              pstmt.setTimestamp(paramPos[p], params.getTimestamp(i));
              break;
            case CLOB:
              // ここにはnullは来ないはず
              char[] charArray = params.getString(i).toCharArray();
              CharArrayReader car = new CharArrayReader(charArray);
              pstmt.setCharacterStream(paramPos[p], car, charArray.length);
              break;
            case JSONOBJECT:
              try {
                pstmt.setString(paramPos[p], LSJson.serialize(params.getObject(i),false));
              } catch (JSONException e) {
                throw new DataException(e);
              }
              break;

            /* 下記メソッドは未検討
            //JDBC 2.0 Array パラメータを設定します。
            setArray(int i, Array x)

            //指定されたパラメータを、指定されたバイト数を持つ指定された入力ストリームに設定します。
            setAsciiStream(int parameterIndex, InputStream x, int length)

            //指定されたパラメータを、指定された数のバイトを持つ指定された入力ストリームに設定します。
            setBinaryStream(int parameterIndex, InputStream x, int length)

            //JDBC 2.0 BLOB パラメータを設定します。
            setBlob(int i, Blob x)

            //JDBC 2.0 指定されたパラメータを、指定された文字数である指定された Reader オブジェクトに設定します。
            setCharacterStream(int parameterIndex, Reader reader, int length)

            //JDBC 2.0 CLOB パラメータを設定します。
            setClob(int i, Clob x)

            //JDBC 2.0 指定された Calendar オブジェクトを使用して、指定されたパラメータを java.sql.Date 値に設定します。
            setDate(int parameterIndex, Date x, Calendar cal)

            //JDBC 2.0 指定されたパラメータを SQL NULL に設定します。
            setNull(int paramIndex, int sqlType, String typeName)

            //パラメータの値を、オブジェクトを使用して設定します。
            setObject(int parameterIndex, Object x)

            //指定されたパラメータの値を、指定されたオブジェクトで設定します。
            setObject(int parameterIndex, Object x, int targetSqlType)

            //パラメータの値を、オブジェクトを使用して設定します。
            setObject(int parameterIndex, Object x, int targetSqlType, int scale)

            //JDBC 2.0 REF(<構造化型>) パラメータを設定します。
            setRef(int i, Ref x)

            //JDBC 2.0 指定された Calendar オブジェクトを使用して、指定されたパラメータを java.sql.Time 値に設定します。
            setTime(int parameterIndex, Time x, Calendar cal)

            //JDBC 2.0 指定された Calendar オブジェクトを使用して、指定されたパラメータを java.sql.Timestamp 値に設定します。
            setTimestamp(int parameterIndex, Timestamp x, Calendar cal)
            */
          }
        }
      }
    } catch (SQLException x) {
      throw new LSSQLException(x);
    }
  }

  /**
   * CallableStatement内のパラメータにDataMapのパラメータの値を割り当てます。<br/><br/>
   *
   * 以下の処理を行ないます。<br/>
   * ・in, inoutパラメータの値設定<br/>
   * ・inout, outパラメータの登録<br/>
   *
   * @param  cstmt    CallableStatementオブジェクト
   * @param  params  DataMapオブジェクト
   * @exception  jp.littlesoft.sql.exception.LSSQLException  例外発生時
   * @exception  jp.littlesoft.data.DataException  例外発生時
   */
  public void bindParamValues(CallableStatement cstmt, Params<?> params) throws LSSQLException, DataException {
    // INパラメータを割り当てる
    bindParamValues((PreparedStatement)cstmt, params);
    try {
      // OUTパラメータを割り当てる
      int[] paramPos = null;
      for (int i=0; i<params.getColumnCount(); i++ ) {
        paramPos = params.getColumn(i).getParamPos();
        if (paramPos==null || paramPos.length==0  || params.getColumn(i).getParamType()==Column.ParamType.IN) continue;
        for (int p=0; p<paramPos.length; p++) {
          switch (params.getColumn(i).getType()) {
            case BIGDECIMAL:
              cstmt.registerOutParameter(paramPos[p], params.getColumn(i).getSqlType(), params.getColumn(i).getScale());
              break;
            default:
              cstmt.registerOutParameter(paramPos[p], params.getColumn(i).getSqlType());
              break;
          }
        }
      }
    } catch (SQLException x) {
      throw new LSSQLException(x);
    }
  }

  /**
   * クエリー実行後のCallableStatementのout、inoutパラメータの値をParamsのパラメータに設定します。<br/>
   * 
   * @param  cstmt    CallableStatementオブジェクト
   * @param  params   Paramsオブジェクト
   * @exception  jp.littlesoft.sql.exception.LSSQLException  例外発生時
   * @exception  jp.littlesoft.data.DataException  例外発生時
   */  
  public void bindResultValues(CallableStatement cstmt, Params<?> params) throws LSSQLException, DataException {
    try {
      int[] paramPos = null;
      for (int i=0; i<params.getColumnCount(); i++ ) {
        paramPos = params.getColumn(i).getParamPos();
        if (paramPos==null || paramPos.length==0  || params.getColumn(i).getParamType()==Column.ParamType.IN) continue;
        for (int p=0; p<paramPos.length; p++) {
          switch(params.getColumn(i).getType()) {
            case STRING:
              String value = cstmt.getString(paramPos[p]);
              if (value!=null)
                params.setString(i, DataUtil.replaceFromDB(value));
              else
                params.setString(i, Const.NULLSTRING);  // nullの場合は ""を設定する。
              break;
            case BINARY:
            case LONGBINARY:
              params.setBinary(i, cstmt.getBytes(paramPos[p]));
              break;
            case BLOB:
              java.sql.Blob blob = cstmt.getBlob(paramPos[p]);
              long blobLen = blob.length();
              if (blobLen > (long) Integer.MAX_VALUE) {
                throw new InvalidColumnException("DATA_SIZE_ERROR",
                    new String[]{params.getColumn(i).getCaption()},
                    params.getColumn(i));
              }
              params.setBinary(i, blob.getBytes(1, (int) blobLen));
              break;
            case BYTE:
              params.setByte(i, cstmt.getByte(paramPos[p]));
              break;
            case SHORT:
              params.setShort(i, cstmt.getShort(paramPos[p]));
              break;
            case LONG:
              params.setLong(i, cstmt.getLong(paramPos[p]));
              break;
            case INT:
              params.setInt(i, cstmt.getInt(paramPos[p]));
              break;
            case FLOAT:
              params.setFloat(i, cstmt.getFloat(paramPos[p]));
              break;
            case DOUBLE:
              params.setDouble(i, cstmt.getDouble(paramPos[p]));
              break;
            case BOOLEAN:
              params.setBoolean(i, cstmt.getBoolean(paramPos[p]));
              break;
            case BIGDECIMAL:
              params.setBigDecimal(i, cstmt.getBigDecimal(paramPos[p]));
              break;
            case DATE:
              params.setDate(i, cstmt.getDate(paramPos[p]));
              break;
            case TIME:
              params.setTime(i, cstmt.getTime(paramPos[p]));
              break;
            case TIMESTAMP:
              params.setTimestamp(i, cstmt.getTimestamp(paramPos[p]));
              break;
            case JSONOBJECT:
              try {
                params.setObject(i, LSJson.deserialize(cstmt.getString(paramPos[p])));
              } catch (JSONException e) {
                throw new DataException(e);
              }
              break;
            case UNKNOWN:
            default:
              String _caption = params.getColumn(i).getCaption();
               throw new InvalidColumnException("INVALID_DATAFIELD_TYPE2",
                 new String[]{ _caption, params.getColumn(i).getType().toString()},
                 params.getColumn(i));
                 
          }
          if (cstmt.wasNull()) params.setNull(i);
        }
      }
    } catch (SQLException x) {
      throw new LSSQLException(x);
    }
  }
  
  /**
   * JDBCのトランザクションアイソレーションレベルを指定されたものに変更することを試みます。<br/><br/>
   *
   * 設定する値はjava.sql.Connectionで定義されてる定数を使用します。<br/>
   * ※JDBCコネクションが確立した状態でないと設定できません。<br/>
   *
   * newIsolationの値がマイナスの場合はトランザクションアイソレーションレベルの変更を行いません。<br/>
   *
   * @param  newIsolation  トランザクションアイソレーションレベル
   * @exception  jp.littlesoft.sql.NJSQLException  例外発生時
   * @see java.sql.Connection
   */
  public void setTransactionIsolation(int newIsolation) throws LSSQLException {
    if (newIsolation < 0)
      return;
    checkConnection();
    try {
      jdbcConnection.setTransactionIsolation(newIsolation);
    } catch (SQLException x) {
      throw new LSSQLException(x);
    }
  }
  
  /**
   * トランザクション遮断レベルを管理するクラス。<br/>
   * 
   * @author LittleSoft Corporation
   */
  public enum TransactionIsolationType {
    /**
     * トランザクションが不明であることを表す定数です。
     */
    UNKNOWN,
    /**
     * トランザクションがサポートされていないことを表す定数です。
     */
    TRANSACTION_NONE,
    /**
     * ダーティー読み込み、繰り返し不可の読み込み、およびファントム読み込みが起こることを表す定数です。
     */
    TRANSACTION_READ_UNCOMMITTED,
    /**
     * ダーティー読み込みは抑制され、繰り返し不可の読み込みおよびファントム読み込みが起こることを表す定数です。
     */
    TRANSACTION_READ_COMMITTED,
    /**
     * ダーティー読み込みおよび繰り返し不可の読み込みは抑制され、ファントム読み込みが起こることを表す定数です。
     */
    TRANSACTION_REPEATABLE_READ,
    /**
     * ダーティー読み込み、繰り返し不可の読み込み、およびファントム読み込みが抑制されることを表す定数です。
     */
    TRANSACTION_SERIALIZABLE    
  }
  
  /**
   * DBMSの種類を管理するクラス。<br/>
   * 
   * @author LittleSoft Corporation
   */
  public enum SQLServerType {
    
    /**
     * Oracleを表す定数です。
     */
    ORACLE,
    /**
     * PostgreSQLを表す定数です。
     */
    POSTRGESQL,
    /**
     * IBM DB2を表す定数です。
     */
    IBM_DB2,
    /**
     * Microsoft SQLを表す定数です。
     */
    MS_SQL,
    /**
     * MySQLを表す定数です。
     */
    MYSQL,
    /**
     * HSQLDBを表す定数です。
     */
    HSQLDB,
    /**
     * Sybaseを表す定数です。
     */
    SYBASE,
    /**
     * Informixを表す定数です。
     */
    INFORMIX,
    /**
     * HiRDBを表す定数です。
     */
    HIRDB,
    /**
     * ODBCを表す定数です。
     */
    ODBC,
    /**
     * DERBYを表す定数です。
     */
    DERBY,
    /**
     * H2を表す定数です。
     */
    H2,
    /**
     * SQL4Gを表す定数です。
     */
    LSJSON,
    /**
     * その他を表す定数です。
     */
    OTHER;

    /**
     * DBMS毎のJDBCドライバー名、DBMSの種類を示す定数の値を序数としてアクセスします。<br/>
     */
    public static final String[] DEFAULT_DIRVERS = {
      "Driver",
      "oracle.jdbc.OracleDriver",
      "org.postgresql.Driver",
      "COM.ibm.db2.jdbc.net.DB2Driver", // Type2 (8.1からはType4がある)
      "com.microsoft.jdbc.sqlserver.SQLServerDriver",
      "com.mysql.jdbc.Driver",
      "org.hsqldb.jdbcDriver",
      "com.sybase.jdbc3.jdbc.SybDriver",
      "com.informix.jdbc.IfxDriver",
      "JP.co.Hitachi.soft.DBPSV_Driver.JdbcDbpsvDriver",
      "sun.jdbc.odbc.JdbcOdbcDriver"
    };

    /**
     * DBMS毎のディフォルトJDBC URL、DBMSの種類を示す定数の値を序数としてアクセスします。<br/>
     */
    public static final String[] DEFAULT_URLS = {
      "jdbc:db://127.0.0.1/mydb",
      "jdbc:oracle:thin:@127.0.0.1:1521:orcl",
      "jdbc:postgresql://127.0.0.1:5432/mydb",
      "jdbc:db2://127.0.0.1:1433/mydb",
      "jdbc:microsoft:sqlserver://127.0.0.1:1433",
      "jdbc:mysql://127.0.0.1/mydb",
      //"jdbc:hsqldb:file:<filename>"
      "jdbc:hsqldb:hsql://localhost:9001",
      "jdbc:sybase:Tds:127.0.0.1:5000/mydb",
      "jdbc:informix-sqli://localhost:1526/mydb:INFORMIXSERVER=server",
      "jdbc:hitachi:dbplib://DB=HIRDB,DBID=22200,DBHOST=localhost",
      "jdbc:odbc:DATASOURCE"
    };

    /**
     * DBMSのJDBCドライバー名から種類を示す定数を取得します。<br/>
     * 
     * @param driverName JDBCドライバー名
     * @return DBMSの種類を示す定数
     */
    public static SQLServerType checkSQLServerType(String driverName) {
      if (driverName==null) return OTHER;
      driverName = driverName.toLowerCase();
      if (driverName.indexOf("oracle")>=0)     return ORACLE;
      if (driverName.indexOf("postgresql")>=0) return POSTRGESQL;
      if (driverName.indexOf("db2")>=0)        return IBM_DB2;
      if (driverName.indexOf("microsoft")>=0 && driverName.indexOf("sqlserver")>=0)
        return MS_SQL;
      if (driverName.indexOf("mysql")>=0)      return MYSQL;
      if (driverName.indexOf("hsqldb")>=0)     return HSQLDB;
      if (driverName.indexOf("sybase")>=0)     return SYBASE;
      if (driverName.indexOf("informix")>=0)   return INFORMIX;
      if (driverName.indexOf("hitachi")>=0 && driverName.indexOf("dbpsv")>=0) return HIRDB;
      if (driverName.indexOf("jdbcodbcdriver")>=0) return ODBC;
      if (driverName.indexOf("apache derby")>=0) return DERBY;
      if (driverName.indexOf("org.h2")>=0 || driverName.equals("h2")) return H2;
      if (driverName.indexOf("jp.littlesoft.sql4g")>=0) return LSJSON;
      return OTHER;
    }

    /**
     * DBMS毎にJDBCAdaptorのディフォルトプロパティを設定します。<br/><br/>
     * 
     * JDBCAdaptorのgetSQLServerType()に応じて、以下のプロパティを設定します。<br/>
     * <ul>
     *   <li>SearchMetaDataByUpperCase
     *   <li>SearchMetaDataByLowerCase
     *   <li>SearchMetaDataWithinSchema
     * </ul>
     *
     * @param sqlAdaptor SQLDataResolverオブジェクト
     * @see SQLDataResolverImpl#getSQLServerType
     */
    public static void setupDefault(SQLAdaptor sqlAdaptor) {
      switch (sqlAdaptor.getSQLServerType()) {
        case ORACLE:
          sqlAdaptor.setSearchMetaDataByUpperCase(true);
          sqlAdaptor.setSearchMetaDataByLowerCase(false);
          sqlAdaptor.setSearchMetaDataWithinSchema(true);
          break;
        case POSTRGESQL:
          sqlAdaptor.setSearchMetaDataByUpperCase(false);
          sqlAdaptor.setSearchMetaDataByLowerCase(true);
          sqlAdaptor.setSearchMetaDataWithinSchema(false);
          break;
        case IBM_DB2:
          sqlAdaptor.setSearchMetaDataByUpperCase(true);
          sqlAdaptor.setSearchMetaDataByLowerCase(false);
          sqlAdaptor.setSearchMetaDataWithinSchema(true);
          break;
        case MS_SQL:
          sqlAdaptor.setSearchMetaDataByUpperCase(false);
          sqlAdaptor.setSearchMetaDataByLowerCase(false);
          sqlAdaptor.setSearchMetaDataWithinSchema(true);
          break;
        case MYSQL:
          sqlAdaptor.setSearchMetaDataByUpperCase(false);
          sqlAdaptor.setSearchMetaDataByLowerCase(true);
          sqlAdaptor.setSearchMetaDataWithinSchema(false);
          break;
        case HSQLDB:
          sqlAdaptor.setSearchMetaDataByUpperCase(true);
          sqlAdaptor.setSearchMetaDataByLowerCase(false);
          sqlAdaptor.setSearchMetaDataWithinSchema(false);
          break;
        case SYBASE:
          sqlAdaptor.setSearchMetaDataByUpperCase(false);
          sqlAdaptor.setSearchMetaDataByLowerCase(false);
          sqlAdaptor.setSearchMetaDataWithinSchema(true);
          break;
        case INFORMIX:
          sqlAdaptor.setSearchMetaDataByUpperCase(false);
          sqlAdaptor.setSearchMetaDataByLowerCase(true);
          sqlAdaptor.setSearchMetaDataWithinSchema(false);
          break;
        case HIRDB:
          sqlAdaptor.setSearchMetaDataByUpperCase(true);
          sqlAdaptor.setSearchMetaDataByLowerCase(false);
          sqlAdaptor.setSearchMetaDataWithinSchema(true);
          break;
        case ODBC:
          sqlAdaptor.setSearchMetaDataByUpperCase(true);
          sqlAdaptor.setSearchMetaDataByLowerCase(false);
          sqlAdaptor.setSearchMetaDataWithinSchema(false);
          break;
        case DERBY:
          sqlAdaptor.setSearchMetaDataByUpperCase(true);
          sqlAdaptor.setSearchMetaDataByLowerCase(false);
          sqlAdaptor.setSearchMetaDataWithinSchema(true);
          break;
        case OTHER:
          sqlAdaptor.setSearchMetaDataByUpperCase(false);
          sqlAdaptor.setSearchMetaDataByLowerCase(false);
          sqlAdaptor.setSearchMetaDataWithinSchema(false);
          break;
      }
    }
  }
}
