/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.sql;

import java.io.Serializable;
import java.sql.Connection;
import java.util.ArrayList;
import java.util.HashMap;

import org.json.JSONException;

import jp.littlesoft.data.ITableDef;
import jp.littlesoft.data.Row;
import jp.littlesoft.data.Table;
import jp.littlesoft.data.TreeModel;
import jp.littlesoft.data.ITableDef.MetaData;
import jp.littlesoft.data.ITableDef.TableDef;
import jp.littlesoft.data.TreeModel.Node;
import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.soja.LSJson;
import jp.littlesoft.util.Const;

@SuppressWarnings("serial")
public class SQLTreeModel extends TreeModel {

  public String dbSchema = "";

  /**
   * @return dbSchema
   */
  public String getDbSchema() {
    return dbSchema;
  }

  /**
   * @param dbSchema セットする dbSchema
   */
  public void setDbSchema(String dbSchema) {
    this.dbSchema = dbSchema;
  }


  @TableDef(owner="SA", schema="PUBLIC", name="LS_SQLTREEMODEL_REC")
  static public class Record implements ITableDef {
    public Long id;
    public Long parentId;
    public String key;
    public String path;
    public String nodeName;
    public String link;
    public HashMap<String, Serializable> attribute;
    private Node node;
    public Boolean isLeaf = new Boolean(false);
    public Boolean isExpanded = new Boolean(false);
    @Override
    public MetaData<Record> createMetaData() {
      return new MetaData<Record>(){{
        addPrimaryKey($$.id);
        addIndex($$.key);
      }};
    }
  }
  @TableDef(owner="SA", schema="PUBLIC", name="LS_SQLTREEMODEL_SEQ")
  static public class Sequence implements ITableDef {
    public String id;
    public Long value;
    @Override
    public MetaData<Sequence> createMetaData() {
      return new MetaData<Sequence>(){{
        addPrimaryKey($$.id);
      }};
    }
  }

  public SQLTreeModel() {
    super();
    nodeList.add(this.root);
  }
  private ArrayList<Node> nodeList =  new ArrayList<Node>();
  private Table<Record> dataTable = null;
  public String nodeSeparator = "/";
  protected int selectedIndex = -1;


  public int indexOf(Node node) {
    return nodeList.indexOf(node);
  }

  @Override
  public void addNodeList(Node position, Node newNode) {
    int index = nodeList.indexOf(position);
    if (index == -1) {
      if (position instanceof Root) {
        nodeList.add(position);
        index = 0;
      } else {
        throw new RuntimeException("Can not find the specified destination node insertion.");
      }
    }
    if (index == nodeList.size()) {
      nodeList.add(newNode);
    } else {
      nodeList.add(index+1, newNode);
    }
  }

  @Override
  public void removeNodeList(Node node) {
    try {
      if (nodeList.contains(node)) {
        if (dataTable!=null) {
          try {
            dataTable.getRow(node.getOriginalRowIndex()).setDeletionReserved(true);
          } catch (ArrayIndexOutOfBoundsException x) {}
        }
      }
      nodeList.remove(node);
    } catch (DataException e) {
      throw new RuntimeException(e);
    }
  }
  private void setupDataTable(Connection conn) throws DataException {
    if (dataTable==null) {
      SQL<Record> sqlContext = new SQL<Record>() {{
        // 2010-01-05 mod by Kumada 同じ階層であればNodeから先に表示するようにするためソート順を修正
        //select().where($$.key).is(getKey()).end.orderBy($$.path);
        select().where($$.key).is(getKey()).end.orderBy($$.isLeaf, $$.path);
        // 2010-02-08 add by Kumada DBスキーマが個別に設定されている場合は、スキーマを指定
        if (!getDbSchema().equals(Const.NULLSTRING))
          getMetaData($$).setDBSchema(getDbSchema());
      }};
      dataTable = sqlContext.getActiveTable(conn);
      dataTable.setDataResolver(sqlContext.getResolver(conn));
      ((ISQLDataResolver)dataTable.getDataResolver()).setConnection(conn);
      String schemaName = sqlContext.getMetaData(sqlContext.$$).getDBSchema();
      String tableName = sqlContext.getMetaData(sqlContext.$$).getDBTableName(false);
      ((ISQLDataResolver)dataTable.getDataResolver()).setTargetDBSchema(schemaName);
      ((ISQLDataResolver)dataTable.getDataResolver()).setTargetDBTable(tableName);
      dataTable.getColumn("id").setNullable(true);
      //setNodeList(dataTable.visibleEntities());
    } else {
      ((ISQLDataProvider)dataTable.getDataProvider()).setConnection(conn);
      ((ISQLDataResolver)dataTable.getDataResolver()).setConnection(conn);
    }
  }

  public void activate(Connection conn) throws DataException {
    setupDataTable(conn);
    dataTable.activate();
    nodeList.clear();
    root.clearChildren();
    //nodeList.add(root);
    HashMap<Long, Node> nodeMap = new HashMap<Long, Node>();
    for (Row<Record> r : dataTable.visibleRows()) {
      Node node;
      if (r.entity.isLeaf) {
        node = new Leaf();
      } else if (r.entity.parentId==null) {
        String rootName = getRootName();
        //this.root = new Root(this);
        setRootName(rootName);
        node = root;
      } else {
        node = new Node();
      }
      node.setName(r.entity.nodeName);
      node.setLink(r.entity.link);
      node.setAttribute(r.entity.attribute);
      if (r.entity.isExpanded!=null && r.entity.isExpanded) {
        node.expand();
      }
      node.setOriginalRowIndex(dataTable.indexOf(r));
      nodeList.add(node);
      nodeMap.put(r.entity.id, node);
      if (r.entity.parentId!=null) {
        Node parent = nodeMap.get(r.entity.parentId);
        if (parent!=null) {
          parent.addChild(node, true);
        } else {
          throw new DataException("There is no specified parent data row. path[" + r.entity.path + "] parentId[" + r.entity.parentId + "]");
        }
      }
    }
  }

  public String getPathString(Node node) {
    if (node==null) return null;
    String temp = node.getName();
    Node parent = node.getParent();
    while(parent!=null) {
      temp = parent.getName() + nodeSeparator + temp;
      parent = parent.getParent();
    }
    return temp;
  }


  public void resolveChanges(Connection conn) throws DataException {
    setupDataTable(conn);
    if (!dataTable.isActive()) dataTable.activate();

    SQL<Sequence> updateSeq = new SQL<Sequence>(){{
      update().set(
          $($$.value).accept( $($$.value).add(1) )
      ).prepare();
      // 2010-02-08 add by Kumada DBスキーマが個別に設定されている場合は、スキーマを指定
      if (!getDbSchema().equals(Const.NULLSTRING))
        getMetaData($$).setDBSchema(getDbSchema());
    }};
    SQL<Sequence> insertSeq = new SQL<Sequence>(){{
      insert().values(
        $($$.id).accept(this.getClass().getName()),
        $($$.value).accept(1)
      ).prepare();
      // 2010-02-08 add by Kumada DBスキーマが個別に設定されている場合は、スキーマを指定
      if (!getDbSchema().equals(Const.NULLSTRING))
        getMetaData($$).setDBSchema(getDbSchema());
    }};
    SQL<Sequence> selectSeq = new SQL<Sequence>(){{
      select().prepare();
      // 2010-02-08 add by Kumada DBスキーマが個別に設定されている場合は、スキーマを指定
      if (!getDbSchema().equals(Const.NULLSTRING))
        getMetaData($$).setDBSchema(getDbSchema());
    }};
    Row<Record> row = null;
    for (Node node : nodeList) {
      if (node.getOriginalRowIndex()==-1) {
        row = dataTable.appendRow(new Record());
        node.setOriginalRowIndex(dataTable.getRowIndex());
        if (updateSeq.execute(conn, true, true)==0) {
          insertSeq.execute(conn, true, true);
        }
        row.entity.id = selectSeq.execute(conn).get(0).value;
      } else {
        row = dataTable.getRow(node.getOriginalRowIndex());
        row.startEditing();
      }
      row.entity.key = this.getKey();
      row.entity.path = getPathString(node);
      row.entity.nodeName = node.getName();
      row.entity.link = node.getLink();
      row.entity.attribute = node.getAttribute();
      row.entity.node = node;
      row.entity.isLeaf = new Boolean((node instanceof Leaf));
      row.entity.isExpanded = new Boolean(node.isExpanded());
      row.validate();
    }
//    dataTable.first();
//    while(dataTable.inBounds()) {
//      Row<Record> r = dataTable.getRow();
//      if (r.entity.id==null) {
//        if (updateSeq.execute(conn, true, true)==0) {
//          insertSeq.execute(conn, true, true);
//        }
//        r.startEditing();
//        r.entity.id = selectSeq.execute(conn).get(0).value;
//        r.validate();
//      }
//      dataTable.next();
//    }
    for (Row<Record> r : dataTable.visibleRows() ) {
      try {
        if (r.entity.node.getParent()!=null) {
          r.startEditing();
          int parentIndex = r.entity.node.getParent().getOriginalRowIndex();
          r.entity.parentId = dataTable.getRow(parentIndex).entity.id;
          r.validate();
        }
      } catch (NullPointerException e) {
        e.printStackTrace();
      }
    }

    dataTable.first();
    while(dataTable.inBounds()) {
      Row<Record> r = dataTable.getRow();
      if (r.isDeletionReserved()) {
        dataTable.deleteRow();
        continue;
      }
      dataTable.next();
    }
    dataTable.resolveChanges();
  }

  public ArrayList<Node> getNodeList() {
    return nodeList;
  }

  public Node getNodeById(long nodeId) {
    for (Node n : nodeList) {
      if (n.getId()==nodeId) {
        return n;
      }
    }
    return null;
  }

  public Node getNodeByName(String name) {
    for (Node n : nodeList) {
      if (name.equals(n.getName())) {
        return n;
      }
    }
    return null;
  }


  public Integer getSelectedIndex() {
    return selectedIndex;
  }

  public Node getSelectedNode() {
    if (selectedIndex>-1)
      return getNodeList().get(selectedIndex);
    else
      return null;
  }

  public void setSelectedNode(Node node) {
    selectedIndex =  getNodeList().indexOf(node);
  }


  public Integer getRecordId(Node node) {
    try {
      int index = node.getOriginalRowIndex();
      return dataTable.getRow(index).entity.id.intValue();
    } catch (Exception x) {
      return null;
    }

  }
}
