/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.sql;

import java.io.Serializable;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import jp.littlesoft.data.ITableDef;
import jp.littlesoft.data.Row;
import jp.littlesoft.data.exception.DataException;

/**
 * エンティティクラスをもとにデータベーススキーマを更新するクラスです.<br/>
 * このクラスの機能を利用できるのは、ITableDefインターフェースを実装したクラスです。</br>
 * @author littlesoft
 *
 */
public class SchemaUpdator implements Serializable {
  private static final long serialVersionUID = 1094593190795585591L;
  private UpdateSQL<?> ddl = new UpdateSQL<Row<?>>();
  private HashMap<Class<?>, ITableDef.MetaData<?>> metaDatas = new HashMap<Class<?>, ITableDef.MetaData<?>>();

  public SchemaUpdator(Connection conn) {
    try {
      ddl.setConnection(conn);
    } catch (DataException e) {
      throw new RuntimeException(e);
    }
  }
  
  private ITableDef.MetaData<?> getMeta(Class<? extends ITableDef> type) throws DataException {
    ITableDef.MetaData<?> meta = metaDatas.get(type);
    if (meta!=null) return meta;
    try {
      meta = type.newInstance().createMetaData();
      metaDatas.put(type, meta);
      return meta;
    } catch (Exception e) {
      throw new DataException(e);
    }
  }
  
  /**
   * typesで指定したエンティティクラスに対応するデータベーステーブルをドロップします.</br>
   * @param types
   * @return SchemaUpdator
   * @throws DataException
   */
  @SuppressWarnings("unchecked")
  public SchemaUpdator dropTables(Class<?>... types) throws DataException  {
    for (Class<?> type : types) {
      ITableDef.MetaData<?> meta = getMeta((Class<? extends ITableDef>) type);
      ddl.setSQL(meta.toDropTableSQL());
      try {
        ddl.executeUpdate(false, false);
      } catch (DataException x) {
        //x.printStackTrace();
        System.out.println("dropTable : " + x.getLocalizedMessage());
      }
    }
    return this;
  }   

  /**
   * typesで指定したエンティティクラスに対応するデータベーステーブルをCREATEします.</br>
   * @param types
   * @return SchemaUpdator
   * @throws DataException
   */
  @SuppressWarnings("unchecked")
  public SchemaUpdator createTable(Class<?>... types) throws DataException {
    for (Class<?> type : types) {
      ITableDef.MetaData<?> meta = getMeta((Class<? extends ITableDef>) type);
      createTable(meta);
    }
    return this;
  }

  private void createTable(ITableDef.MetaData<?> meta) throws DataException {
    ddl.setSQL(meta.toCreateTableSQL());
    ddl.executeUpdate(false, false);
    if (meta.getPrimaryKey() != null) {
      ddl.setSQL(meta.getPrimaryKey().toString());
      ddl.executeUpdate(false, false);
    }
    for (ITableDef.Index index : meta.getIndexes()) {
      ddl.setSQL(index.toString());
      ddl.executeUpdate(false, false);
    }
  }

  public SchemaUpdator createOrReplaceTable(Class<?>... types) throws DataException {
    checkSchema(types);
    dropTables(types);
    createTable(types);
    addForeginKeys(types);
    return this;
  }

  /**
   * typesで指定したエンティティクラスに対応するデータベーステーブルが存在しない場合CREATEし、既に存在する場合、変更箇所を検出し、ALTER文を生成・適用します.</br>
   * @param types
   * @return SchemaUpdator
   * @throws DataException
   */
  public SchemaUpdator createOrAlterTable(Class<?>... types) throws DataException {
    checkSchema(types);
    if (checkDifference(types)) {
      dropForeginKeys(types);
      addForeginKeys(types);
    }
    return this;
  }
  
  private void checkSchema(Class<?>... types) throws DataException {
    try {
      DatabaseMetaData dbMeta = ddl.getConnection().getMetaData();
      String userName = dbMeta.getUserName();
      ArrayList<String> schemaNames = new ArrayList<String>();
      ResultSet schemas = dbMeta.getSchemas();
      while(schemas.next()) {
        schemaNames.add(schemas.getString(JDBCTableMetaData.SCHEMA_NAME));
      }
      for (Class<?> type : types) {
        ITableDef.MetaData<?> newMeta = getMeta((Class<? extends ITableDef>)type);
        if (newMeta.getDBSchema()==null) continue;
        boolean exists = false;
        for (String schema : schemaNames) {
          if (schema.equals(newMeta.getDBSchema())) {
            exists = true;
            break;
          }
        }
        if (!exists) {
          String sql = "CREATE SCHEMA " + newMeta.getDBSchema() 
          //+ " AUTHORIZATION " + newMeta.getOwner();
          + " AUTHORIZATION " + userName;
          ddl.setSQL(sql);
          try{
            ddl.executeUpdate(true, false);
            schemaNames.add(newMeta.getDBSchema());
          } catch (jp.littlesoft.sql.exception.LSSQLException x) {
            throw new DataException("Failed to create schema:" + x.getLocalizedMessage());
          }
        }
      }
    } catch (SQLException e) {
      throw new DataException(e);
    }
    
  }
  
  /**
   * 
   * @param types
   * @throws DataException
   */
  private boolean checkDifference(Class<?>... types) throws DataException {
    boolean result = false; 
    try {
      ArrayList<String> alterSQLs = new ArrayList<String>();
      DatabaseMetaData dbMeta = ddl.getConnection().getMetaData();
      for (Class<?> type : types) {
        ITableDef.MetaData<?> newMeta = getMeta((Class<? extends ITableDef>)type);
        JDBCTableMetaData jdbcMeta;
        try {
          jdbcMeta = new JDBCTableMetaData(dbMeta, 
              newMeta.getDBCatalog(), newMeta.getDBSchema(), newMeta.getDBTableName(false));
          jdbcMeta.checkDifference(newMeta, alterSQLs);
        } catch (JDBCTableMetaData.TableNotFound x) {
          //存在しないテーブルなので新規作成します。
          createTable(newMeta);
          result = true;
          continue;
        }
      }
      for (String alterSQL : alterSQLs) {
        System.out.println("> " + alterSQL);
        ddl.setSQL(alterSQL);
        try {
          ddl.executeUpdate(false, false);
          result = true;
        } catch (DataException x) {
          //x.printStackTrace();
          System.out.println(x.getLocalizedMessage());
        }
      }
    } catch (SQLException e) {
      throw new DataException(e);
    }
    return result;
  }
  
  /**
   * typesで指定したエンティティクラスに対応するデータベーステーブルの主キーをドロップします.</br>
   * @param types
   * @return SchemaUpdator
   * @throws DataException
   */
  @SuppressWarnings("unchecked")
  public SchemaUpdator dropPrimaryKeys(Class<?>... types) throws DataException {
    for (Class<?> type : types) {
      ITableDef.MetaData<?> meta = getMeta((Class<? extends ITableDef>) type);
      dropPrimaryKeys(meta);
    }
    return this;
  }
  
  private void dropPrimaryKeys(ITableDef.MetaData<?> meta) throws DataException {
    if (meta.getPrimaryKey() != null) {
      ddl.setSQL(meta.getPrimaryKey().toDropString());
      try {
        ddl.executeUpdate(false, false);
      } catch (Exception x) {
        System.out.println("dropPrimaryKeys : " + x.getLocalizedMessage());
      }
    }
  }
  
  /**
   * typesで指定したエンティティクラスに対応するデータベーステーブルに主キーを追加します.</br>
   * @param types
   * @return SchemaUpdator
   * @throws DataException
   */
  @SuppressWarnings("unchecked")
  public SchemaUpdator addPrimaryKeys(Class<?>... types) throws DataException {
    for (Class<?> type : types) {
      ITableDef.MetaData<?> meta = getMeta((Class<? extends ITableDef>) type);
      addPrimaryKeys(meta);
    }
    return this;
  }
  
  private void addPrimaryKeys(ITableDef.MetaData<?> meta) throws DataException {
    if (meta.getPrimaryKey() != null) {
      ddl.setSQL(meta.getPrimaryKey().toString());
      ddl.executeUpdate(false, false);
    }
  }

  /**
   * typesで指定したエンティティクラスに対応するデータベーステーブルの外部キーをドロップします.</br>
   * @param types
   * @return
   * @throws DataException
   */
  @SuppressWarnings("unchecked")
  public SchemaUpdator dropForeginKeys(Class<?>... types) throws DataException {
    for (Class<?> type : types) {
      ITableDef.MetaData<?> meta = getMeta((Class<? extends ITableDef>) type);
      dropForeginKeys(meta);
    }
    return this;
  }
  
  private void dropForeginKeys(ITableDef.MetaData<?> meta) throws DataException {
    try {
      for (ITableDef.ForeginKey fk : meta.getForeginKeys()) {
        ddl.setSQL(fk.toDropString());
        ddl.executeUpdate(false, false);
      }
    } catch (Exception x) {
      System.out.println("dropForeginKeys : " + x.getLocalizedMessage());
    }
  }

  /**
   * typesで指定したエンティティクラスに対応するデータベーステーブルの外部キーを追加します.</br>
   * @param types
   * @return
   * @throws DataException
   */
  @SuppressWarnings("unchecked")
  public SchemaUpdator addForeginKeys(Class<?>... types) throws DataException {
    for (Class<?> type : types) {
      ITableDef.MetaData<?> meta = getMeta((Class<? extends ITableDef>) type);
      addForeginKeys(meta);
    }
    return this;
  }
  
  private void addForeginKeys(ITableDef.MetaData<?> meta) throws DataException {
    for (ITableDef.ForeginKey fk : meta.getForeginKeys()) {
      ddl.setSQL(fk.toString());
      ddl.executeUpdate(false, false);
    }
  }
  
  /**
   * typesで指定したエンティティクラスに対応するデータベーステーブルのインデクスをドロップします.</br>
   * @param types
   * @return SchemaUpdator
   * @throws DataException
   */
  @SuppressWarnings("unchecked")
  public SchemaUpdator dropIndexes(Class<?>... types) throws DataException {
    for (Class<?> type : types) {
      ITableDef.MetaData<?> meta = getMeta((Class<? extends ITableDef>) type);
      dropIndexes(meta);
    }
    return this;
  }

  private void dropIndexes(ITableDef.MetaData<?> meta) throws DataException {
    for (ITableDef.Index ix : meta.getIndexes()) {
      ddl.setSQL(ix.toDropString());
      try {
        ddl.executeUpdate(false, false);
      } catch (Exception x) {
        // x.printStackTrace();
        System.out.println("dropIndexes : " + x.getLocalizedMessage());
      }
    }
  }

  /**
   * typesで指定したエンティティクラスに対応するデータベーステーブルのインデクスをCREATEします.</br>
   * @param types
   * @return SchemaUpdator
   * @throws DataException
   */
  @SuppressWarnings("unchecked")
  public SchemaUpdator createIndexes(Class<?>... types) throws DataException {
    for (Class<?> type : types) {
      ITableDef.MetaData<?> meta = getMeta((Class<? extends ITableDef>) type);
      createIndexes(meta);
    }
    return this;
  }
  
  private void createIndexes(ITableDef.MetaData<?> meta ) throws DataException {
    for (ITableDef.Index ix : meta.getIndexes()) {
      ddl.setSQL(ix.toString());
      ddl.executeUpdate(false, false);
    }
  }
}
