/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.sql;

import java.io.Serializable;
import java.sql.*;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import jp.littlesoft.data.Params;
import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.sql.exception.LSSQLException;
import jp.littlesoft.sql.logger.SQLLogger;
import jp.littlesoft.sql.logger.SQLTimeLogger;
import jp.littlesoft.util.Const;
import jp.littlesoft.util.StringUtil;

/**
 * 更新SQL文実行クラス。<br/><br/>
 *
 * Update,Delete,Insert,ストアドプロシージャ,DDL文など結果を返さないSQLを実行することができます。<br/>
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
public class UpdateSQL<T extends Serializable> extends Params<T> {
  
  /**
   * 
   */
  private static final long serialVersionUID = 1L;

  private static final transient Logger logger = LogManager.getLogger(UpdateSQL.class.getName());

  /**
   * JDBCAdaptorオブジェクトを保持する内部変数です。<br/>
   */
  private SQLAdaptor sqlAdaptor;

  /**
   * 実行するSQL文を保持する内部変数です。<br/>
   */
  private String sql = null;

  /**
   * 検索キーなどのパラメタを解析した後のSQL文を保持する内部変数です。<br/>
   */
  private String jdbcSql = null;

  /**
   * JDBCステートメントを保持する内部変数です。<br/>
   */
  private transient Statement jdbcStatement = null;

  /**
   * コンストラクタ。<br/>
   */  
  public UpdateSQL(T...t) {
    super(t);
  }
  
  // for PHPApplication
  @SuppressWarnings("unchecked")
  public UpdateSQL() {
  }
  
  /**
   * JDBCのコネクションを再設定します。<br/>
   * 
   * @param conn JDBCのコネクション
   * @throws DataException
   */
  @SuppressWarnings("unchecked")
  public UpdateSQL(Connection conn) throws DataException {
    this.setConnection(conn);
  }
  
  /**
   * JDBCのコネクションを設定します。<br/>
   * 
   * @param conn JDBCのコネクション
   * @throws DataException
   */
  public void setConnection(Connection conn) throws DataException {
    if (this.getSQLAdaptor()==null) {
      this.setSQLAdaptor(new SQLAdaptor());
    }
    this.getSQLAdaptor().setConnection(conn);    
  }

  /**
   * JDBCのコネクションを取得します。<br/>
   * 
   * @return JDBCのコネクション
   * @throws DataException
   */
  public Connection getConnection() {
    if (this.getSQLAdaptor()==null) {
      return null;
    }
    return this.getSQLAdaptor().getConnection();    
  }
  
  /**
   * JDBCAdaptorオブジェクト取得します。<br/>
   *
   * @return  SQLDataResolver  JDBCAdaptorオブジェクト
   */
  public SQLAdaptor getSQLAdaptor() {
    return sqlAdaptor;
  }

  /**
   * JDBCAdaptorオブジェクトを設定します。<br/>
   *
   * @param  sqlAdaptor  JDBCAdaptorオブジェクト
   */  
  public void setSQLAdaptor(SQLAdaptor sqlAdaptor) {
    this.sqlAdaptor = sqlAdaptor;
  }

  /**
   * 実行するSQL文を設定します。<br/><br/>
   *
   * 実行するSQL文を文字列で設定します。<br/>
   * SQL文にはパラメータを含める事が可能です。パラメータの先頭には":"を指定します。<br/>
   * UpdateSQLはSQL分中のパラメータを解析し、必要なColumnオブジェクトを生成します。<br/>
   * 
   * <pre>
   * [例1]
   *  // UPDATE文を実行する。
   *  updateSql.setSQL("UPDATE EMP SET SAL = SAL + 10000 WHERE JOB = :JOB");
   *  updateSql.getColumn("JOB").setType(Column.Type.DOUBLE);
   *  updateSql.setDouble("JOB", 1022);
   * 
   * [例2]
   *  // ストアドプロシージャを呼び出す。
   *  updateSql.setSQL("{call proc1(:PARAM1, :RESULT)}");
   *  updateSql.getColumn("PARAM1").setType(Column.Type.STRING);
   *  updateSql.getColumn("RESULT").setType(Operation.INT);
   *  updateSql.getColumn("RESULT").setParamType(Column.ParamType.OUT);
   *  updateSql.setSTRING("PARAM1", "xxxx");
   * 
   * [例3]
   *  // ストアドファンクションを呼び出す。
   *  updateSql.setSQL("{:RESULT = call func1(:PARAM1)}");
   *  updateSql.getColumn("PARAM1").setType(Column.Type.STRING);
   *  updateSql.getColumn("RESULT").setType(Operation.INT);
   *  updateSql.getColumn("RESULT").setParamType(Column.ParamType.OUT);
   *  updateSql.setSTRING("PARAM1", "xxxx");
   * </pre>
   * 
   * ※前回の実行による影響を受けないようにするために、事前に定義していたパラメータセットは
   * 初期化しています。従って、事前にパラメタを定義しても無意味です。上記の例のようにsetSQL()
   * を実行した後に、パラメタの型を定義するようにしてください。<br/><br/>
   * 
   * ※前回実行したときのSQLと同じSQLが指定された場合は何もしません。<br/>
   * 
   * @param  newSQL  実行するSQL文
   * @exception  jp.littlesoft.data.DataException  例外発生時
   */  
  public void setSQL(String newSQL) throws DataException{
    if (newSQL!=null && newSQL.equals(sql)) return;
    freeStatement();
    sql = newSQL;
    jdbcSql = null;
    this.clearColumns();
    jdbcSql = this.parseSQL(sql);
  }

  /**
   * 実行するSQL文を取得します。<br/>
   *
   * @return  String  SQL文
   */
  public String getSQL() {
    if (this.sql == null) return Const.NULLSTRING;
    return this.sql;
  }

  /**
   * JDBCステートメントを取得します。<br/>
   * 
   * @return JDBCステートメント
   */  
  protected Statement getJDBCStatement() {
    return this.jdbcStatement;
  }

  /**
   * JDBCステートメントを設定します。<br/>
   * 
   * @param jdbcStatement JDBCステートメント
   */  
  protected void setJDBCStatement(Statement jdbcStatement) {
    this.jdbcStatement = jdbcStatement;
  }

  /**
   * SQL文を実行します。<br/><br/>
   *
   * パラメータpendingCommitは当メソッド内部でのコミットを保留するか？を指定します。<br/>
   * 当パラメータをtrueに指定し、更新SQL文を実行した場合は、明示的にcommit,rollbackを行う必要があります。<br/>
   * commit,rollbackはSQL実行時のJDBCAdaptorオブジェクトの下記のメソッドで行います。<br/>
   * ・updateSQL.getConnection().commit();<br/>
   * ・updateSQL.getConnection().rollback();<br/><br/>
   *
   * パラメータreuseStatementは当メソッド内部で作成されるPreparedStatementオブジェクトを
   * 再利用するか？を指定します。<br/>
   * 当パラメータをtrueに指定する目的は、同一更新SQLのパフォーマンスを向上させることです。<br/>
   * また当パラメータをtrueに指定した場合は、明示的にcloseメソッドを呼び出し、開かれたままのカーソルをクローズする必要があります。<br/>
   * 
   * @param pendingCommit  True：コミットを保留する、False：しない
   * @param reuseStatement  True：ステートメントを再利用する、False：しない
   * @return  更新された行の数
   * @exception  jp.littlesoft.data.DataException  例外発生時
   */   
  @SuppressWarnings("unchecked")
  public int executeUpdate(boolean pendingCommit, boolean reuseStatement) throws DataException {
    int updateCount = 0;
    if (StringUtil.isNullString(this.getSQL())) {
      throw new DataException(null, null, "SQL_IS_NULL", new String[]{this.getClass().getName()});
    }
    if (sqlAdaptor==null) {
      throw new DataException(null, null, "JDBC_ADAPTOR_IS_NOT_SET",
          new String[]{this.getClass().getName()});
    }
    try {
      // 更新ログ出力 (改行文字も出力される。)
      if (SQLLogger.isUpdateDebugEnabled()) SQLLogger.logForUpdate(sqlAdaptor.getClientId(true), sql, this);

      if (!sqlAdaptor.isConnecting()) sqlAdaptor.connect();
      sqlAdaptor.setAutoCommit(false);
      if (!reuseStatement) freeStatement();

      // 実行時間の測定開始
      long start = System.currentTimeMillis();

      boolean _isCallableStatement = false;
      if (this.getColumnCount()==0) {
        if (jdbcStatement==null)  jdbcStatement = sqlAdaptor.createStatement();
        updateCount = jdbcStatement.executeUpdate(jdbcSql);
      } else {
        // SQL中の "\/\*$ xxxx $\*\/" を sqlParameters.getValueAsString("xxxx") の値に置換する
        String jdbcSql = this.getJDBCSQL();
        _isCallableStatement = isCallableStatement(this.getSQL());
        if (_isCallableStatement) {
          // StoredProcedure
          if (jdbcStatement==null)  jdbcStatement = sqlAdaptor.prepareCall(jdbcSql);
          sqlAdaptor.bindParamValues((CallableStatement)jdbcStatement, this);
          updateCount = ((CallableStatement)jdbcStatement).executeUpdate();
        } else {
          if (jdbcStatement==null)  jdbcStatement = sqlAdaptor.preparedStatement(jdbcSql);
          sqlAdaptor.bindParamValues((PreparedStatement)jdbcStatement, this);
          updateCount = ((PreparedStatement)jdbcStatement).executeUpdate();
        }
      }
            
      if (_isCallableStatement) {
        // bind OUT Parameter
        sqlAdaptor.bindResultValues((CallableStatement)this.getJDBCStatement(), (Params)this);
      }

      // 実行時間のログ出力
      if (SQLTimeLogger.isUpdateDebugEnabled()) {
        long executeTime = System.currentTimeMillis() - start;
        String qaName = "";
        SQLTimeLogger.logForUpdate(SQLTimeLogger.EXECUTE_UPDATE, sqlAdaptor.getClientId(true), qaName, this.getClass().getName(), executeTime);
      }

      if (!pendingCommit) sqlAdaptor.getConnection().commit();
      return updateCount;

    } catch (Exception e) {
      try {
        if (!pendingCommit) sqlAdaptor.getConnection().rollback();
      } catch (SQLException s) {}
      if(e instanceof SQLException) throw new LSSQLException((SQLException)e);
      if(e instanceof DataException) throw (DataException)e;
      throw new DataException(e);
    } finally {
      if (!reuseStatement) close();
    }
  }

  /**
   * JDBCステートメントをクローズします。<br/>
   */
  public void close() {
    freeStatement();
  }

  /**
   * JDBCステートメントを解放します。<br/>
   */
  private void freeStatement() {
    if (jdbcStatement != null) {
      try {  jdbcStatement.close(); } catch (java.sql.SQLException x) {}
      jdbcStatement = null;
    }
  }

  /**
   * SQL文がストアドプロシージャ用かどうかを調べます。<br/>
   *
   * @param   sql  SQL文
   * @return  True：ストアドプロシージャ用、False：それ以外
   */
  private static final boolean isCallableStatement(String sql) {
    if (StringUtil.isNullString(sql)) return false;
    String temp1 = sql.trim().toLowerCase();
    if (temp1.startsWith("{") && (temp1.indexOf("call ")!=-1) )
      return true;
    else
      return false;
  }
  /**
   * パラメータのプレースホルダがJDBC形式に変換された後のSQL文を取得します。<br/><br/>
   *
   * ※注意　このとき、SQL中の &#47;*$ xxxx $*&#47; というコメント部分もTable#getParams().getValueAsString("xxxx")の値に置換されます。<br/>
   * 例：htmlTable.setSQL("SELECT AAA FROM TBL_X WHERE BBB = :BBB &#47;*$ CCC_IN $*&#47;");<br/>
   *     htmlTable.getParams().getValueAsString("CCC_IN") の値が "AND CCC IN ('X', 'Y')"<br/>
   * の場合<br/>
   * "SELECT AAA FROM TBL_X WHERE BBB = ? AND CCC IN ('X', 'Y')"<br/>
   * が結果として返されます。<br/>
   *
   * @return  JDBC形式変換後のSQL文
   */
  public String getJDBCSQL() {
    return getJDBCSQL(this.jdbcSql);
  }
//protected String getJDBCSQL() {
//if (this.jdbcSql == null) return Const.NULLSTRING;
//return this.jdbcSql;
//}

  /**
   * 検索キーなどのパラメタを解析した後のSQL文を取得します。<br/>
   * 
   * @return 検索キーなどのパラメタを解析した後のSQL文
   */
  
  public String getJDBCSQL(String sql) {

    if (StringUtil.isNullString(sql)) return Const.NULLSTRING;
    if (this.getColumnCount() == 0) return sql;

    //  "\/\*$ ～ $\*\/" が一つもなければ sql をそのまま返す。
    int firstIndex = sql.indexOf("/*$ ");
    if (firstIndex < 0)
      return sql;
    int lastIndex = sql.lastIndexOf(" $*/");
    if (lastIndex < firstIndex)
      return sql;

    if (logger.isDebugEnabled())
      logger.debug("getJdbcSQL():before replace = " + sql);

    // "\/\* xxxx \*\/" をparamSet.getValueAsString("xxxx")の文字列に置換する。
    // firstIndexは '\/\*' かもしれないので、頭から走査しなおす。
    StringBuffer strBuf = new StringBuffer();
    int beforeIndexE = 0;
    int i = -1;
    while (i < lastIndex - 1) {
      i++;
      char c = sql.charAt(i);
      // "'" ～ "'"は読み飛ばす。
      if (c == '\'') {
        i++;
        while (i < lastIndex) {
          if (sql.charAt(i) == '\'')
            break;
          i++;
        }
        continue;
      }

      // "\/\*$ "でなければ次を検索
      if (!"/*$ ".equals(sql.substring(i, i + 4))) {
        continue;
      }
      int indexS = i;

      // " $\*\/" があるか?
      // " $\*\/" がなければ終了（走査しても意味がない）
      int indexE = sql.indexOf(" $*/", indexS + 5);
      if (indexE < 0)
        break;

      // "\/\*$ ～ $\*\/"以前の部分を追加
      strBuf.append(sql.substring(beforeIndexE, indexS));
      beforeIndexE = indexE + 4;

      // "\/\* KEYWORD \*\/"のKEYWORD部分の取得
      String keyword = sql.substring(indexS + 4, indexE).trim();
      if (keyword.length() == 0) {
        // キーワードがない場合は何も追加しない。つまり""に置換される。
        continue;
      }

      // 同じ名前のParamがあるか？
      if (this.indexOfColumn(keyword) != -1) {
        // あれば置換する。
        try {
          String replacedWord = this.getValueAsString(keyword);
          strBuf.append(replacedWord);
        } catch (DataException ex) {}
      } else {
        // なければ \/\* ～ \*\/ はそのままとする。（ヒントの可能性もあるため）
        strBuf.append(sql.substring(indexS, indexE + 4));
      }

    }

    // "\/\* ～ \*\/"以降の部分を追加
    strBuf.append(sql.substring(beforeIndexE));

    if (logger.isDebugEnabled())
      logger.debug("getJdbcSQL():after  replace = " + strBuf.toString());

    return strBuf.toString();

  }
}
