/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.sql.logger;

import java.io.Serializable;

import jp.littlesoft.data.Column;
import jp.littlesoft.data.Params;
import jp.littlesoft.data.exception.DataException;
import jp.littlesoft.util.Const;

import org.apache.log4j.*;

/**
 * 実行されるSQL文のロギングを行うクラス。<br/><br/>
 * 
 * ログの出力自体は、Log4j管理下のLoggerクラスに委譲されます。<br/>
 * 本クラスは、実行されるSQL文及びパラメータを編集し、Loggerクラスにログ内容を送信します。<br/>
 * 標準で、jp.littlesoft.sql.SQLDataResolver, jp.littlesoft.sql.UpdateSQLで行われるSQL処理に組み込まれています。<br/>
 * ログの出力レベルを<B>DEBUG</B>に指定すると出力されます。<br/><br/>
 * 
 * log4j.xml設定ファイルにて、更新用ログのみ／照会用ログのみ／更新用ログ＋照会用ログ などの出力制御ができます。
 * 
 * <pre>
 * 設定例：
 * 
 *  &lt;!-- 更新用ログ＋照会用ログ --&gt;
 *  &lt;logger name=&quot;jp.littlesoft.sql.logger.SQLLogger&quot;&gt;
 *      &lt;level value =&quot;DEBUG&quot; /&gt;
 *  &lt;/logger>
 * 
 *  &lt;!-- 更新用ログのみ --&gt;
 *  &lt;logger name=&quot;jp.littlesoft.sql.logger.SQLLogger.update&quot;&gt;
 *      &lt;level value ="DEBUG" /&gt;
 *  &lt;/logger>
 *  
 *  &lt;!-- 照会用ログのみ --&gt;
 *  &lt;logger name=&quot;jp.littlesoft.sql.logger.SQLLogger.query&quot;&gt;
 *      &lt;level value ="DEBUG" /&gt;
 *  &lt;/logger&gt;
 *  
 * 出力例:
 * 
 *  2004-12-24 11:34:24,078 LS-SQL clientId: SELECT EMP_CD, EMP_NAME WHERE DEPT_CD = :DEPT_CD ORDER BY EMP_CD
 *  2004-12-24 11:34:24,078 LS-SQL clientId:   [1] DEPT_CD: sales
 * </pre>
 * 
 * @author  LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public final class SQLLogger implements Serializable {

  private static final String CATEGORY        = SQLLogger.class.getName();
  private static final String CATEGORY_UPDATE = CATEGORY+".update";
  private static final String CATEGORY_QUERY  = CATEGORY+".query";
	
  private static Logger logger = null;
  private static Logger updateLogger = null;
  private static Logger queryLogger  = null;
	
  private SQLLogger() {
  }
  
  /**
   * 更新用のログを出力します。<br/><br/>
   * 
   * 注意:ログレベルが、<b>DEBUG</b>以外の時は出力は無視されます。<br/>
   * 
   * @param clientId 処理要求元を判別するための識別子（null指定可）
   * @param sql      出力するSQL文字列
   * @param params   出力するSQLに付随するパラメータ（null指定可）
   */
  public static void logForUpdate(String clientId, String sql, Params<?> params) {
    if (isUpdateDebugEnabled())
      log(updateLogger, clientId, sql, params);
  }
  
  /**
   * 照会用のログを出力します。<br/><br/>
   * 
   * 注意:ログレベルが、<b>DEBUG</b>以外の時は出力は無視されます。<br/>
   * 
   * @param clientId 処理要求元を判別するための識別子（null指定可）
   * @param sql      出力するSQL文字列
   * @param params   出力するSQLに付随するパラメータ（null指定可）
   */
  public static void logForQuery(String clientId, String sql, Params<?> params) {
    if (isQueryDebugEnabled())
      log(queryLogger, clientId, sql, params);
  }

  private static void log(Logger logger, String clientId, String sql, Params<?> params) {
    // このメソッドの呼び出し前に isDebugEnabledを行っているので、ここでは行わない。
    try {
      log(logger, clientId, (sql!=null ? sql.trim() : Const.NULLSTRING));
      if(params==null) return;
      for (int i=0; i<params.getColumnCount(); i++ ) {
        int[] paramPos = params.getColumn(i).getParamPos();
        if (paramPos==null || paramPos.length==0  || params.getColumn(i).getParamType()==Column.ParamType.OUT) continue;
        for (int p=0; p<paramPos.length; p++) {
                StringBuffer _logBuf = new StringBuffer();
                _logBuf.append("  [").append(paramPos[p]).append("] ")
                       .append(params.getColumn(i).getName()).append(": ")
                       .append(params.getValueAsString(i));
            log(logger, clientId, _logBuf.toString());
        }
      }
      log(logger, clientId, Const.NULLSTRING);
    } catch (DataException e){ e.printStackTrace(); }
  }
      
  private static void log(Logger logger, String clientId, String msg){
    logger.debug(formatMessage(clientId, msg));
  }

  private static String formatMessage(String clientId, String msg){
    final StringBuffer _retBuf = new StringBuffer();
    //_retBuf.append(clientId);
    if(msg == null) {
      _retBuf.append(' ');
    } else {
      //_retBuf.append(": ").append(msg);  
      _retBuf.append(" ").append(msg);      
    }  
    return _retBuf.toString();
  }

  /**
   * データ照会用ログが出力可能かどうか調べます。<br/>
   *
   * @return True：出力可、False：出力不可
   */
  public static boolean isQueryDebugEnabled() {
    resetQueryLogger();
    if (queryLogger == null)
      return false;
    return queryLogger.isDebugEnabled();
  }

  /**
   * データ更新用ログが出力可能かどうか調べます。<br/>
   * 
   * @return True：出力可、False：出力不可
   */
  public static boolean isUpdateDebugEnabled() {
    resetUpdateLogger();
    if (updateLogger == null)
      return false;
    return updateLogger.isDebugEnabled();
  }

  private static void resetLogger() {
    if (logger == null)
      logger = LogManager.getLogger(CATEGORY);
  }

  private static void resetQueryLogger() {
    if (queryLogger == null) {
      queryLogger = LogManager.getLogger(CATEGORY_QUERY);
      if (queryLogger == null)
        queryLogger = logger;
    }
  }

  private static void resetUpdateLogger() {
    if (updateLogger == null) {
      updateLogger = LogManager.getLogger(CATEGORY_UPDATE);
      if (updateLogger == null)
        updateLogger = logger;
    }
  }
  
  static {
    resetLogger();
    resetQueryLogger();
    resetUpdateLogger();
  }

}
