/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.util;

import java.util.Map;

/**
 * Stringに関するユーティリティメソッドを集めたクラス。<br/>
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public final class StringUtil implements java.io.Serializable {
  
  /**
   * コンマを内部的に置換する際の文字列を意味する定数です。<br/>
   * <!-- 本当は &comma; にしたかった -->
   */
  public static final String COMMA_REPLACE = "&comma";

  /**
   * replaceConstStringメソッド内でクォート文字で囲まれた文字列を置換する際の
   * プレフィックスを意味する定数です。<br/>
   */
  public static final String CONST_STRING = "_CONSTSTR";

  private StringUtil() {
  }

  /**
   * Win改行対応です。<br/><br/>
   * 
   * \nを見つけたら、その前に\rを挿入します。<br/>
   * 
   * @param str \rを挿入前の文字列(!=null)
   * @return \rを挿入後の文字列
   */
  public static String insertCR(String str) {
    StringBuffer s = new StringBuffer();
    char c;
    for(int i=0;i<str.length();i++){
      c = str.charAt(i);
      if (c == '\n') {
        s.append('\r');
      }
      s.append(c);
    }
    return s.toString();
  }

  /**
   * Win改行対応です。<br/><br/>
   *
   * CR('\r')を除去した文字列を返します。<br/>
   * 
   * @param str \rを含む文字列(!=null)
   * @return \rを除去した文字列
   */
  public static String deleteCR(String str) {
    StringBuffer s = new StringBuffer();
    char c;
    for(int i=0;i<str.length();i++){
      c = str.charAt(i);
      if (c != '\r') s.append(c);
    }
    return s.toString();
  }


  /**
   * Win改行対応です。<br/><br/>
   *
   * CR('\r') と LF('\n') を除去した文字列を返します。<br/>
   * 
   * @param str \rを含む文字列(!=null)
   * @return \rを除去した文字列
   */
  public static String deleteCRLF(String str) {
    StringBuffer s = new StringBuffer();
    char c;
    for(int i=0;i<str.length();i++){
      c = str.charAt(i);
      if (c != '\r' && c != '\n') s.append(c);
    }
    return s.toString();
  }

  /**
   * Win改行対応です。<br/><br/>
   * 
   * CR('\r') と LF('\n') を半角空白に変換した文字列を返します。<br/>
   * 
   * @param str CR,LFを変換したい文字列
   * @return CR,LFを変換した文字列
   */
  public static String replaceCRLFtoSP(String str) {
    StringBuffer s = new StringBuffer();
    char c;
    for(int i=0;i<str.length();i++){
      c = str.charAt(i);
      if (c == '\r' || c == '\n') {
        s.append(' ');
      } else {
        s.append(c);
      }
    }
    return s.toString();
  }

  /**
   * strがダブルクォーテーションで囲まれている場合、それを除去した文字列を返します。<br/>
   * 
   * @param str 前後のダブルクォーテーションを除去したい文字列
   * @return strから前後のダブルクォーテーションを除去した文字列
   */
  public static String deleteFirstLastDblQuote(String str) {
    if (str == null)
      return null;
    int keyLength = str.length();
    if (keyLength > 1 && str.charAt(0) == Const.DOUBLE_QUOTE && str.charAt(keyLength-1) == Const.DOUBLE_QUOTE) {
      // "..." で囲まれていたら、"を除去する。
      str = str.substring(1, keyLength-1);
    }
    return str;
  }


  /**
   * 文字列が null または長さ0かどうかを調べます。<br/>
   *
   * @param str nullかどうかを判定したい文字列
   * @return True：null(または長さ0)、False：null(または長さ0)ではない
   */
  public static final boolean isNullString(String str) {
    if (str == null || str.length() == 0) {
      return true;
    } else {
      return false;
    }
  }

  /**
   * 文字列が null 以外で、長さ1以上かどうかを調べます。<br/>
   *
   * @param str nullかどうかを判定したい文字列
   * @return True：null(または長さ0)ではない、False：null(または長さ0)
   */
  public static final boolean isNotNullString(String str) {
    if (str == null || str.length() == 0) {
      return false;
    } else {
      return true;
    }
  }

  /**
   * 2つの文字列が同じかどうかを調べます。<br/>
   * 
   * str1 または str2 のいずれか、もしくは両方が null の場合は falseを返します。<br/>
   * 
   * @param str1 比較したい文字列
   * @param str2 比較したい文字列
   * @return True：2つの文字列が同じ、False：同じではない(またはどちらか一方または両方がnull)
   */
  public static final boolean isEqualString(String str1, String str2) {
    if (str1 != null && str2 != null && str1.equals(str2)) {
      return true;
    } else { 
      return false;
    }
  }

  /**
   * 2つの文字列が同じかどうかを調べます。<br/><br/>
   * 
   * str1, str2 の両方がnullの場合もtrueを返します。<br/><br/>
   * 
   * ※注 str1=null, str2="" の場合はfalseを返します。<br/>
   * 
   * @param str1 比較したい文字列
   * @param str2 比較したい文字列
   * @return True：2つの文字列が同じ(または両方がnull)、False：同じではない(またはどちらか一方がnull)
   */
  public static final boolean isEqualStringOrBothNull(String str1, String str2) {
    if (str1 == null) {
      if (str2 == null)
        return true;
      return false;
    } else if (str2 != null && str1.equals(str2)) {
      return true;
    } else { 
      return false;
    }
  }


  /**
   * 前後の空白や改行文字を除去した文字列を返します。<br/><br/>
   * 
   * 先頭の' ', タブ, 改行文字と<br/>
   * 末尾の' ', タブ, 改行文字を除去した文字列を返します。<br/><br/>
   * 
   * 例:<br/>
   * trimSpace(" \t  ABC  ") ===> "ABC" が結果となる。<br/>
   * 
   * @param str 変換したい文字列
   * @return 変換後の文字列
   */
  public static final String trimSpace(String str) {
    if (StringUtil.isNullString(str)) return "";
    
    // 先頭の空白を除去する
    int i=0;
    while (i<str.length()) {
      char c = str.charAt(i);
      if (c != ' ' && c != '\t' && c != '\r' && c != '\n') {
        break;
      }
      i++;
    }
    if (i==str.length()) return "";
    
    String wStr = null;
    if (i == 0) {
      wStr = str;
    } else {
      wStr = str.substring(i);
    }
    
    // 末尾の改行コードは除去する。
    i=wStr.length();
    if (i < 1) return "";
    
    // 末尾の空白を除去する。
    // もし空白がなくても空白を追加する。（改行文字の代わり）
    while (i>0) {
      char c = wStr.charAt(i-1);
      if (c != ' ' && c!= '\t' && c != '\r' && c != '\n') {
        break;
      }
      i--;
    }
    if (i < 1) return "";
    wStr = wStr.substring(0, i);

    return wStr;
  }


  /**
   * 文字列を int値として解析します。<br/><br/>
   *
   * 解析できなかった場合は defautValue を返します。<br/>
   *
   * @param str int値として解析したい文字列
   * @return 解析したint値
   */
  public static final int parseInt(String str, int defaultValue) {
    try {
      if (str == null || str.length() == 0) return defaultValue;
      return Integer.parseInt(str);
    } catch (Exception ex) {
      return defaultValue;
    }
  }


  /**
   * 文字列をBoolean値として解析します。<br/><br/>
   *
   * strがnullまたは長さ0の場合は defaultValue を返します。<br/><br/>
   *
   * strが長さ1以上で、"true"以外の場合(大文字、小文字は区別しない)は
   * falseを返します。<br/>
   *
   * @param str boolean値として解析したい文字列
   * @param defaultValue デフォルトのboolean値
   * @return 解析したboolean値
   */
  public static final boolean parseBoolean(String str, boolean defaultValue) {
    if (str == null || str.length() == 0) return defaultValue;
    return Boolean.valueOf(str).booleanValue();
  }

  /**
   * String中の文字参照を変換します。<br/><br/>
   *
   * '&lt;' =&gt; "&amp;lt;"<br/>
   * '&gt;' =&gt;" &amp;gt;"<br/>
   * '&amp;' =&gt; "&amp;amp;"<br/>
   * '&yen;'' =&gt; "&amp;apos;"<br/>
   * '&quot;' =&gt; "&amp;quot;"<br/>
   *
   * @param   str 変換前文字列
   * @return  変換後文字列
   */
  public static String replaceCharToESC(String str) {
    return replaceCharToESC(str, false, false);
  }

  /**
   * String中の文字参照を変換します。<br/><br/>
   *
   * '&lt;' =&gt; "&amp;lt;"<br/>
   * '&gt;' =&gt; "&amp;gt;"<br/>
   * '&amp;' =&gt; "&amp;amp;"<br/>
   * '&yen;'' =&gt; "&amp;apos;"<br/>
   * '&quot;' =&gt; "&amp;quot;"<br/>
   *
   * @param  str 変換前文字列
   * @param  cntlToSpc True：\t,\r,\n以外の0-31までの文字コードを半角空白に置換する、False：置換しない
   * @return  変換後文字列
   */
  public static String replaceCharToESC(String str, boolean cntlToSpc) {
    return replaceCharToESC(str, cntlToSpc, false);
  }
  
  /**
   * String中の文字参照を変換します。<br/><br/>
   *
   * '&lt;' =&gt; "&amp;lt;"<br/>
   * '&gt;' =&gt; "&amp;gt;"<br/>
   * '&amp;' =&gt; "&amp;amp;"<br/>
   * '&yen;'' =&gt;" &amp;apos;"<br/>
   * '&quot;' =&gt; "&amp;quot;"<br/>
   *
   * @param  str 変換前文字列
   * @param  cntlToSpc True：\t,\r,\n以外の0-31までの文字コードを半角空白に置換する、False：置換しない
   * @param   lfToCrlf '\n' を "\r\n"に変換するか?
   * @return  変換後文字列
   */
  public static String replaceCharToESC(String str, boolean cntlToSpc, boolean lfToCrlf) {
    if (str == null || str.length() == 0) {
      return str;
    }

    StringBuffer s = new StringBuffer();
    char c;
    for(int i=0;i<str.length();i++){
      c = str.charAt(i);
      switch (c) {
        case '<':
          s.append("&lt;");
          break;
        case '>':
          s.append("&gt;");
          break;
        case '&':
          s.append("&amp;");
          break;
        case '\'':
          s.append("&apos;");
          break;
        case '"':
          s.append("&quot;");
          break;
        default:
          if (!cntlToSpc || c >= 32) {
            if (lfToCrlf && c == '\n') {
              if (i == 0 || str.charAt(i-1) != '\r') {
                // '\n'を"\r\n"に変換する。
                s.append('\r');
              }
            }
            s.append(c);
          } else {
            if (c == '\n') {
              if (lfToCrlf && (i == 0 || str.charAt(i-1) != '\r')) {
                // '\n'を"\r\n"に変換する。
                s.append('\r');
              }              
              s.append(c);
            } else if (c == '\t' || c == '\r') {
              s.append(c);
            } else {
              // \t,\r,\n以外の0-31までの制御文字は空白として出力する。
              s.append(' ');
            }
          }
          break;
      }
    }
    return s.toString();
  }


  /**
   * String中の文字参照を変換します。<br/><br/>
   *
   * "&amp;lt;" =&gt; '&lt;'<br/>
   * "&amp;gt;" =&gt; '&gt;'<br/>
   * "&amp;nbsp;" =&gt; ' '<br/>
   * "&amp;amp;" =&gt; '&amp;'<br/>
   * "&amp;apos;" =&gt; '&yen;''<br/>
   * "&amp;quot;" =&gt; '&yen;&quot;'<br/>
   *
   * @param   str 変換前文字列
   * @return  変換後文字列
   */
  public static String replaceESCToChar(String str) {
    if (str == null) return str;
    int len = str.length();
    if (len == 0) return str;
    
    if (str.indexOf('&') < 0 || str.indexOf(';') < 0) return str;

    StringBuffer s = new StringBuffer();
    int lastIndex = 0;
    while (lastIndex < len) {
      int ampIndex = str.indexOf('&', lastIndex);
      if (ampIndex < 0 || ampIndex >= len-1) {
        s.append(str.substring(lastIndex));
        break;
      }

      int semicIndex = str.indexOf(';', ampIndex+1);
      if (semicIndex < 0) {
        s.append(str.substring(lastIndex));
        break;
      }

      // '&'の前までをコピーする。
      s.append(str.substring(lastIndex, ampIndex));

      // &...; のキーワード部分を取り出し
      String esc = str.substring(ampIndex+1, semicIndex);
      if ("lt".equals(esc)) {
        s.append('<');
      } else if ("gt".equals(esc)) {
        s.append('>');
      } else if ("nbsp".equals(esc)) {
        s.append(' ');
      } else if ("amp".equals(esc)) {
        s.append('&');
      } else if ("apos".equals(esc)) {
        s.append('\'');
      } else if ("quot".equals(esc)) {
        s.append('\"');
      } else {
        s.append(str.substring(ampIndex, semicIndex+1));
      }
      lastIndex = semicIndex+1;        
    }
    return s.toString();
  }

  /**
   * String中の文字をJavaScript用に変換します。<br/><br/>
   *
   * "&yen;'" =&gt; "&yen;&yen;&yen;'"<br/>
   * 
   * "&yen;&yen;" =&gt; "&yen;&yen;&yen;&yen;"<br/>
   *
   * @param str 変換前文字列
   * @param replaceYen True：'&yen;&yen;'を"&yen;&yen;&yen;&yen;"に変換する、False：変換しない
   * @return 変換後文字列
   */
  public static String replaceForJavaScript(String str, boolean replaceYen) {
    if (str == null) return Const.NULLSTRING;
    int len = str.length();
    if (len == 0) return str;
    
    //if (str.indexOf('\'') < 0 || str.indexOf('\\') < 0) return str;

    StringBuffer s = new StringBuffer();
    int i = 0;
    while (i < len) {
      char c = str.charAt(i);
      if (c == '\'') {
        s.append("\\\'");
      } else if (c == '\n') {
        s.append("\\n");
      } else if (c == '\r') {
        s.append("\\r");
      } else if (c == '\t') {
        s.append("\\t");
      } else if (replaceYen && c == '\\') {
        s.append("\\\\");
      } else {
        s.append(c);
      }
      i++;
    }
    return s.toString();
  }

  /**
   * 指定された文字の数を数えます。<br/><br/>
   *
   * 例: countCharacter('X', "abcXdefX") ==&lt; 2<br/>
   * 
   * @param c 個数を数えたい文字
   * @param str 検索対象の文字列
   * @return 指定された文字の数
   */
  public static int countCharacter(char c, String str) {
    if (str == null) return 0;
    int count = 0;
    int idx = 0;
    while (idx >= 0) {
      idx = str.indexOf(c, idx);
      if (idx < 0) break;
      idx++;
      count++;
    }

    return count;
  }

  /**
   *
   * 与えられた文字列をASCII文字としてバイト配列に変換します。<br/><br/>
   *
   * str.getBytes("8859_1"); と同じ動作ですが、高速に動作します。<br/><br/>
   * 
   * ※注１　strにASCII文字以外が含まれていた場合、正しい結果を返しません。<br/>
   * ※注２　strのヌルチェックは行っていません。事前にチェックしてください。<br/>
   *
   * <!-- Javaパフォーマンステクニック」の本に載っていたものの改良版です。 -->
   * 
   * @param str バイト配列に変換したい文字列（ヌルは禁止）
   * @return strをASCII文字としてバイト配列に変換した結果
   */
  public static byte[] getAsciiBytes(String str) {

    int size = str.length();
    byte[] byteBuf = new byte[size];
    char[] charBuf = str.toCharArray();
    for (int i=0; i<size; i++) {
      byteBuf[i] = (byte)charBuf[i];
    }
    return byteBuf;
  }

  /**
   * 文字列をクエリーに適したかたちに変換します。<br/>
   *
   * @param str  対象文字列
   * @param from 置換対象の部分文字列 (長さ1以上のString)
   * @param to   第２引数を置換する文字列
   * @return     置換後の文字列
   */
  public static String replace( String str, String from, String to ){
    if( str == null || str.length() <= 0 ){
      return str;
    }
    if (to == null)
      to = "";
    String temp = null;
    StringBuffer buf = new StringBuffer();
    int lpos = 0;
    int rpos = 0;
    int replen = from.length();
    if (replen == 0)
      throw new IllegalArgumentException("StringUtil.replace: from argument is invalid.");
    while(lpos < str.length()){
        rpos = str.indexOf( from, lpos );
        if( rpos < 0 ) break;
        if( lpos < rpos ){
            temp = str.substring( lpos, rpos );
        } else {
            temp = "";
        }
        buf.append(temp).append(to);
        lpos = rpos + replen;
    }
    if( lpos < str.length() ){
        buf.append( str.substring( lpos, str.length() ) );
    }
    return buf.toString();
  }

  /**
   * String型の配列を","で連結します。<br/><br/>
   * 
   * ※注: String中に含まれる","は"&amp;comma"に置換される。<br/>
   * 
   * @param array 連結したいStringの配列
   * @return 連結結果
   */
  public static String arrayToString(String[] array) {
    if (array == null || array.length == 0) return "";
    
    StringBuffer strBuf = new StringBuffer();
    for (int i=0; i<array.length; i++) {
      if (i>0) strBuf.append(Const.COMMA_CHAR);
      strBuf.append(replace(array[i], Const.COMMA_STRING, COMMA_REPLACE));
    }
    return strBuf.toString();
  }

  /**
   * boolean型の配列を","で連結します。<br/>
   * 
   * @param array 連結したいStringの配列
   * @return 連結結果
   */
  public static String arrayToString(boolean[] array) {
    if (array == null || array.length == 0) return "";
    
    StringBuffer strBuf = new StringBuffer();
    for (int i=0; i<array.length; i++) {
      if (i>0) strBuf.append(Const.COMMA_CHAR);
      strBuf.append(array[i]);
    }
    return strBuf.toString();
  }
  
  /**
   * 文字列から、指定した一文字の位置を返します。<br/>
   * 
   * @param readData 検索対象の文字列
   * @param startIndex 検索開始位置
   * @param endIndex 検索終了位置
   * @param ch 検索したい文字
   * @param passQuote True：クォート文字を読み飛ばす、False：読み飛ばさない
   * @return 検索文字の位置
   */ 
  public static int findChar(String readData, int startIndex, int endIndex, char ch, boolean passQuote) {
    if (readData == null) return -1;

    char c;
    int i = startIndex;
    while (i < endIndex) {
      c = readData.charAt(i);
      if (c == ch) {
        return i;
      } else if (passQuote && (c == '"' || c == '\'')) {
        // "または'を見つけたら終わりの"または'まで読み飛ばす。
        i = findChar(readData, i+1, endIndex, c, false);
        if (i < 0) return -1;
      }
      i++;
    }
    return -1;
  }

  /**
   * 文字列から、指定した文字の位置を返します。<br/><br/>
   * 
   * charsで指定した文字のいずれかを見つけたら、その位置を返します。<br/>
   * 
   * @param readData 検索対象の文字列
   * @param startIndex 検索開始位置
   * @param endIndex 検索終了位置
   * @param chars 検索したい文字の配列
   * @return 検索文字の位置
   */ 
  public static int findChars(String readData, int startIndex, int endIndex, char[] chars) {
    if (readData == null) return -1;

    char c;
    int i = startIndex;
    while (i < endIndex) {
      c = readData.charAt(i);
      if (getIndex(chars, c) >= 0) {
        return i;
      }
      i++;
    }
    return -1;
  }

  /**
   * 文字列から、空白文字の位置を返します。<br/>
   * 
   * @param readData 検索対象の文字列
   * @param startIndex 検索開始位置
   * @param endIndex 検索終了位置
   * @param passQuote True：クォート文字を読み飛ばす、False：読み飛ばさない
   * @return 空白文字の位置(見つからない場合は-1)
   */ 
  public static int findSpace(String readData, int startIndex, int endIndex, boolean passQuote) {
    if (readData == null) return -1;

    char c;
    int i = startIndex;
    while (i < endIndex) {
      c = readData.charAt(i);
      if (c == ' ' || c == '\n' || c == '\r' || c == '\t') {
        return i;
      } else if (passQuote && (c == '"' || c == '\'')) {
        // "または'を見つけたら終わりの"または'まで読み飛ばす。
        i = findChar(readData, i+1, endIndex, c, false);
        if (i < 0) return -1;
      }
      i++;
    }
    return -1;
  }

  /**
   * 文字列から、空白以外の文字の位置を返します。<br/>
   * 
   * @param readData 検索対象の文字列
   * @param startIndex 検索開始位置
   * @param endIndex 検索終了位置
   * @return 空白以外の文字の位置(見つからない場合は-1)
   */ 
  public static int findNotSpace(String readData, int startIndex, int endIndex) {
    if (readData == null) return -1;

    char c;
    int i = startIndex;
    while (i < endIndex) {
      c = readData.charAt(i);
      if (c != ' ' && c != '\n' && c != '\r' && c != '\t') {
        return i;
      }
      i++;
    }
    return -1;
  }

  /**
   * 文字列から、指定した文字列の位置を返します。<br/>
   * 
   * @param readData 検索対象の文字列
   * @param startIndex 検索開始位置
   * @param endIndex 検索終了位置
   * @param findStr 検索したい文字列
   * @param passQuote True：クォート文字を読み飛ばす、False：読み飛ばさない
   * @return 検索文字列の位置(見つからない場合は-1)
   */ 
  public static int findString(String readData, int startIndex, int endIndex, String findStr, boolean passQuote) {
    if (readData == null) return -1;
    if (findStr == null || findStr.length() == 0) return -1;

    int findStrLen = findStr.length();

    int index = readData.substring(startIndex, endIndex).indexOf(findStr);
    if (!passQuote) return index;
    if (index < 0) return -1;
    
    // 'や"を読み飛ばす場合は一時ずつチェックする必要がある。
    int end = readData.length() - findStrLen;
    int i = startIndex;    
    char c;
    while (i < end) {
      c = readData.charAt(i);
      if (c == '"' || c == '\'') {
        // "または'を見つけたら終わりの"または'まで読み飛ばす。
        i = findChar(readData, i+1, end, c, false);
        if (i < 0) return -1;
      } else if (readData.substring(i, i+findStrLen).equals(findStr)) {
        return i;
      }
      i++;
    }
    return -1;
  }

  /**
   * char配列の中の文字のインデックスを返します。<br/><br/>
   * 
   * ※注 charsの要素数が多い場合には、Arrays.sortおよびArrays.binarySearchを
   * 利用することを検討してください。<br/>
   * 
   * @param chars 検索対象の文字配列
   * @param c 検索したい文字
   * @return 検索文字のインデックス(見つからない場合は-1)
   */
  public static int getIndex(char[] chars, char c) {
    if (chars == null) return -1;
    for (int i=0; i<chars.length; i++) {
      if (chars[i] == c) {
        return i;
      }
    }
    return -1;
  }

  /**
   * 指定した長さに満たない場合、指定した文字で左側を埋めた文字列を返します。<br/><br/>
   * 
   * 例：
   * lpad("abc", 'x', 5) ==&gt; "xxabc"<br/>
   * 
   * @param str 元の文字列
   * @param pad 追加する文字
   * @param len 文字列の長さ
   * @return 指定した長さに満たない場合、指定した文字で左側を埋めた文字列
   */
  public static String lpad(String str, char pad, int len) {
    int l;
    if (str == null)
      l = 0;
    else
      l = str.length();
    if (l >= len) return str;

    StringBuffer strBuf = new StringBuffer();
    for (int i=l; i<len; i++) {
      strBuf.append(pad);    
    }
    if (l > 0)
      strBuf.append(str);
    return strBuf.toString();
  }

  /**
   * 文字列の左側からだけトリミング処理を行います。<br/>
   * 
   * @param str 処理対象文字列
   * @return 処理結果
   */
  public static String trimLeft(String str) {
    if(str==null) return null;
    String trimmed = str.trim();
    return str.substring(str.indexOf(trimmed));
  }
  
  /**
   * 文字列の右側からだけトリミング処理を行います。<br/>
   * 
   * @param str 処理対象文字列
   * @return 処理結果
   */
  public static String trimRight(String str) {
    if(str==null) return null;
    String trimmed = str.trim();
    return str.substring(0, str.indexOf(trimmed)) + trimmed;
  }
  
  /**
   * BOMよりCharsetを読み取ります。<br/>
   * 
   * @param datas 先頭の3バイト
   * @return Charset(特定できなかった場合は、"8859_1"
   */
  public static String getCharset(byte[] datas) {
    // 先頭のBOMから、Charsetを推測する。
    // UCS-4は考慮しない。
    if (datas != null && datas.length >= 3) {
      if (datas[0] == (byte)0xfe && datas[1] == (byte)0xff) {
        return "UTF-16BE";        
      } else if (datas[0] == (byte)0xff && datas[1] == (byte)0xfe) {
        return "UTF-16LE";
      } else if (datas[0] == (byte)0xef && datas[1] == (byte)0xbb && datas[2] == (byte)0xbf) {
        return "UTF-8";
      }
    }
    return "8859_1";
  }
  
  /**
   * 文字列中の quoteCharで囲まれた文字列を置換して、HashMapに格納します。<br/><br/>
   * 
   * 同時に余計な空白を除去している。<br/>
   * また、\"など\に続く文字の変換も行っている。<br/>
   * 例：
   * <pre>
   *   String result = StringUtil.replaceConstString("\"XXX\"+\"AAA\\\"BBB\"+\"CCC\"", '\"', map);
   *   --&gt; result = "_CONSTSTR0_+_CONSTSTR1_+_CONSTSTR2_" となる。
   *          mapには、key=_CONSTSTR0_, value="XXX"
   *                   key=_CONSTSTR1_, value="AAA\"BBB"
   *                   key=_CONSTSTR2_, value="CCC"
   *          が格納される。
   * </pre>
   * 
   * @param parseString 変換前の文字列
   * @param quoteChar QuoteChar
   * @param hashMap 置換した内容を格納するMap(key="_CONSTSTRn_"(n=0,1,2,...), value=元の文字列)
   * @return 変換後の文字列
   */
  @SuppressWarnings("unchecked")
  public static String replaceConstString(String parseString, char quoteChar, Map hashMap) {
    StringBuffer strBuf = new StringBuffer();
    int constCount = 0;
    int len = parseString.length();
    char c;
    int i = 0;
    while (i<len) {
      c = parseString.charAt(i);
      if (c == '\\') {
        strBuf.append(c);
        i++;
        c = parseString.charAt(i);
        strBuf.append(c);
      } else if (c == quoteChar) {
        // 終わりの " までを _CONSTSTRn_に置換する。(nは0,1,2,...)
        int j = i+1;
        boolean foundEnd = false;
        StringBuffer varValueBuf = new StringBuffer();
        while (j<len) {
          c = parseString.charAt(j);
          if (c == '\\') {
            // '\\'は読み飛ばす。
            j++;
            if (j<len) {
              c = parseString.charAt(j);
              varValueBuf.append(c);
            }
          } else if (c == quoteChar) {
            foundEnd = true;
            i = j;
            break;
          } else {
            varValueBuf.append(c);
          }
          j++;
        }
        if (foundEnd) {
          StringBuffer _varNameBuf = new StringBuffer();
          _varNameBuf.append(CONST_STRING);
          _varNameBuf.append(constCount);
          _varNameBuf.append('_');
          String varName = _varNameBuf.toString();
          hashMap.put(varName, varValueBuf.toString());
          strBuf.append(varName);
          constCount++;
        } else {
          // エラーの場合
          strBuf.append(quoteChar);          
        }
      } else if (c == ' ' || c == '\n' || c == '\r') {
      // 余計な空白や改行文字を読み飛ばす
      } else {
        strBuf.append(c);
      }
      i++;
    }
    return strBuf.toString();
  }

  /**
   * 文字列中の quoteCharのいずれかで囲まれた文字列を置換して、HashMapに格納します。<br/><br/>
   * 
   * 同時に余計な空白を除去している。<br/>
   * また、\"など\に続く文字の変換も行っている。<br/>
   * 例：
   * <pre>
   *   String result = StringUtil.replaceConstString("\"XXX\"+'AAA\\\"BBB'+\"CCC\"", new char[]{'\"', '\''}, map);
   *   --&gt; result = "_CONSTSTR0_+_CONSTSTR1_+_CONSTSTR2_" となる。
   *          mapには、key=_CONSTSTR0_, value="XXX"
   *                   key=_CONSTSTR1_, value="AAA\"BBB"
   *                   key=_CONSTSTR2_, value="CCC"
   *          が格納される。
   * </pre>
   * 
   * @param parseString 変換前の文字列
   * @param quoteChars QuoteCharの配列
   * @param passSpaces True：空白文字(' ','\r','\n','\t'を読み飛ばす、False：読み飛ばさない
   * @param hashMap 置換した内容を格納するMap(key="_CONSTSTRn_"(n=0,1,2,...), value=元の文字列)
   * @return 変換後の文字列
   */
	  @SuppressWarnings("unchecked")
    public static String replaceConstString(String parseString, char[] quoteChars, boolean passSpaces, Map hashMap) {
    StringBuffer strBuf = new StringBuffer();
    //String s = "\"aa\\\"BB\"";
    int constCount = 0;
    int len = parseString.length();
    char c;
    int i = 0;
    while (i<len) {
      c = parseString.charAt(i);
      if (c == '\\') {
        strBuf.append(c);
        i++;
        c = parseString.charAt(i);
        strBuf.append(c);
      } else {
        final int quoteIdx = getIndex(quoteChars, c);
        if (quoteIdx >= 0) {
          final char quoteChar = quoteChars[quoteIdx];
          // 終わりの quoteChar までを _CONSTSTRn_に置換する。(nは0,1,2,...)
          int j = i+1;
          boolean foundEnd = false;
          StringBuffer varValueBuf = new StringBuffer();
          while (j<len) {
            c = parseString.charAt(j);
            if (c == '\\') {
              // '\\'は読み飛ばす。
              j++;
              if (j<len) {
                c = parseString.charAt(j);
                varValueBuf.append(c);
              }
            } else if (c == quoteChar) {
              foundEnd = true;
              i = j;
              break;
            } else {
              varValueBuf.append(c);
            }
            j++;
          }
          if (foundEnd) {
            StringBuffer _varNameBuf = new StringBuffer();
            _varNameBuf.append(CONST_STRING);
            _varNameBuf.append(constCount);
            _varNameBuf.append('_');
            String varName = _varNameBuf.toString();
            hashMap.put(varName, varValueBuf.toString());
            strBuf.append(varName);
            constCount++;
          } else {
            // エラーの場合
            strBuf.append(quoteChar);          
          }
        } else if (passSpaces && (c == ' ' || c == '\n' || c == '\r' || c == '\t')) {
        // 余計な空白や改行文字を読み飛ばす
        } else {
        }
        i++;
      }
    }
    return strBuf.toString();
  }

}
