/* Copyright (c) 2006-2010 LittleSoft Corporation
 * 
 * LittleSoft J-Programming
 * 当プログラムの使用許諾については以下を参照してください。
 * http://www.littlesoft.jp/lsj/license.html
 */
package jp.littlesoft.web;

import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.math.BigDecimal;
import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.Date;

import org.apache.velocity.VelocityContext;

import jp.littlesoft.data.Code;
import jp.littlesoft.data.Column;
import jp.littlesoft.data.Row;
import jp.littlesoft.util.StringUtil;
import jp.littlesoft.util.DataUtil.XWareki;

/**
 * HTMLエンコードに関するユーティリティメソッドを集めたクラス。<br/>
 * 
 * @author LittleSoft Corporation
 * @version 1.3.0
 */
@SuppressWarnings("serial")
public class HTML implements Serializable {
  
  private final static DecimalFormat decimalFormat = new DecimalFormat();
  private final static XWareki wareki = new XWareki();
  private final static SimpleDateFormat simpleDateFormat = new SimpleDateFormat();
  
  /**
   * HTML オブジェクトを生成します。<br/>
   */
  private HTML() {}

  private static java.util.Random ramdom = new java.util.Random();
  
  public static String getRandomValue(Object object) {
    String value = object.getClass().getName() 
      + Integer.toHexString(object.hashCode())
      + Long.toHexString(System.currentTimeMillis()) + "z" 
      + Long.toHexString(ramdom.nextLong()) + "Z";
    return Integer.toHexString(value.hashCode());
  }

  /**
   * ページに埋め込むINPUTタグ(hidden)を生成します。<br/>
   * 
   * @param name タグ名
   * @param value 値
   * @return 出力タグ内容
   */
  public static String toHiddenParam(String name , String value){
    return "<input type=\"hidden\" name=\"" + name + "\" value=\"" + value  + "\">";
  }
  
  /**
   * 引数 value で指定された文字列中の文字参照を変換します。 <br/><br/>
   *
   * parseCharacterReferences(value, false, false)を呼び出しています。<br/>
   *
   * @param   value 変換前文字列
   * @return  変換後文字列
   * @see #escape(String, boolean, boolean)
   */
  public static String escape(String value) {
    return escape(value, false, false);
  }

  /**
   * 引数 value で指定された文字列中の文字参照を変換します。<br/><br/>
   *
   * parseCharacterReferences(value, replaceSpace, false)を呼び出しています。<br/>
   *
   * @param value 変換前文字列
   * @param replaceSpace True：空白を置換する、False：しない
   * @return 変換後文字列
   * @see #escape(String, boolean, boolean)
   */
  public static String escape(String value, boolean replaceSpace) {
    return escape(value, replaceSpace, false);
  }

  /**
   * 引数 value で指定された文字列中の文字参照を変換します。<br/><br/>
   * 
   * 変換を行うのは次の文字です。<br/>
   * <ul>
   *   <li>'<' --> '&amp;lt;'</li>
   *   <li>'>' --> '&amp;gt;'</li>
   *   <li>'&' --> '&amp;amp;'</li>
   *   <li>'"' --> '&amp;quot;'</li>
   *   <li>'&yen;'' --> '&amp;#39;'</li>
   *   <li>' ' --> '&amp;nbsp;' (replaceSpaceがtrueのときのみ)</li>
   *   <li>'&yen;r&yen;n' --> '&lt;BR&gt;' (replaceCRLFがtrueのときのみ)</li>
   * </ul><br/>
   * 
   * ※注　replaceSpaceがfalseの場合は、半角空白は変換しません。<br/> 
   * ※注　replaceCRLFがfalseの場合は、\nは変換しません。<br/>
   * 
   * @param   value 変換前文字列
   * @param   replaceSpace True：空白を置換する、False：しない
   * @param   replaceCRLF True：CRLFを置換する、False：しない
   * @return  変換後文字列
   */
  public static String escape(String value, boolean replaceSpace, boolean replaceCRLF) {
    if (value==null) return ""; 
    if ("".equals(value)) {
      return value;
    }
    if (value.indexOf('<') < 0 && value.indexOf('>') < 0
        && value.indexOf('&') < 0 && value.indexOf('"') < 0 && value.indexOf('\'') < 0 && value.indexOf(' ') < 0
        && value.indexOf('\n') < 0) {
      return value;
    }
    StringBuffer newValue = new StringBuffer();
    for(int i=0; i<value.length(); i++) {
      char code = value.charAt(i);
      if (code == '<') {
        newValue.append("&lt;");
      }
      else if (code == '>') {
        newValue.append("&gt;");
      }
      else if (code == '&') {
        newValue.append("&amp;");
      }
      else if (code == '"') {
        newValue.append("&quot;");
      }
      else if (code == '\'') {
        newValue.append("&#39;");
      }
      else if (replaceSpace && code == ' ') {
        newValue.append("&nbsp;");
      }
      else if (replaceCRLF && code == '\r') {
        newValue.append("<BR>");
        if (i < value.length()-1) {
          if (value.charAt(i+1) == '\n') i++;
        }
      } else if (replaceCRLF && code == '\n') {
        newValue.append("<BR>");
      }
      else {
        newValue.append(code);
      }
    }
    return newValue.toString();
  }
  
  /**
   * 引数 value で指定された文字列中の文字参照を変換します。<br/><br/>
   * 
   * 変換を行うのは次の文字です。<br/>
   * <ul>
   *   <li>'&yen;r&yen;n' --> '&lt;BR&gt;'</li>
   * </ul><br/>
   * 
   * @param   value 変換前文字列
   * @return  変換後文字列
   */
  public static String replaceCRLF(String value) {
    if (value==null) return ""; 
    if ("".equals(value)) {
      return value;
    }
    StringBuffer newValue = new StringBuffer();
    for(int i=0; i<value.length(); i++) {
      char code = value.charAt(i);
      if (code == '\r') {
        newValue.append("<BR>");
        if (i < value.length()-1) {
          if (value.charAt(i+1) == '\n') i++;
        }
      } else if (code == '\n') {
        newValue.append("<BR>");
      }
      else {
        newValue.append(code);
      }
    }
    return newValue.toString();
  }

  static String toMaxLength(Column column) {
    // MaxLength対応
    String m = "";
    switch (column.getType()) {
    // 文字列型カラムの場合
    case STRING:
      // 設定されている Length -> MaxLength
      if (column.getLength() > -1) {
        m = Integer.toString(column.getLength());
      }
      break;
    // 数値型カラムの場合
    case BYTE:
    case SHORT:
    case LONG:
    case INT:
    case FLOAT:
    case DOUBLE:
    case BIGDECIMAL:
      if (column.getPrecision() > -1) {
        String basis = "";
        // 整数部分、小数部分を「9」で埋めて、基準数値を生成
        int repeat = (column.getScale() > -1) ? column.getPrecision() - column.getScale() : column.getPrecision();
        while (basis.length() < repeat) {
          basis = basis.concat("9");
        }
        if (column.getScale() > 0) {
          basis = basis.concat(".");
          for (int i = 0; i < column.getScale(); i++) {
            basis = basis.concat("9");
          }
        }
        // フォーマットパターンが設定されている場合は、フォーマット
        String formatPattern = column.getFormatPattern();
        if (!StringUtil.isNullString(formatPattern)) {
          decimalFormat.applyPattern(formatPattern);
          basis = decimalFormat.format(new BigDecimal(basis));
        }
        // 「-」マイナス入力の分、+1 する
        if (column.isPermitMinusValue()) {
          m = Integer.toString(basis.length() + 1);
        } else {
          m = Integer.toString(basis.length());
        }
      }
      break;
    // 日付、時刻型カラムの場合
    case DATE:
    case TIME:
    case TIMESTAMP:
      String basis = "";
      String formatPattern = column.getFormatPattern();
      try {
        // 基準日時を生成
        simpleDateFormat.applyPattern(Column.DEFAULT_FORMAT_PATTERN_TIMESTAMP); //明示的に指定(ロケールが異なるとパースエラーになる為)
        Date basisDate = simpleDateFormat.parse("1970/01/01 00:00:00.000");
        // 和暦フォーマットパターンが設定されている場合
        if (XWareki.isWarekiFormatPattern(formatPattern)) {
          basis = wareki.format(basisDate, formatPattern);
        }
        // それ以外の場合
        else {
          // パターンが設定されていない場合、デフォルトのフォーマットパターンを適用
          if (StringUtil.isNullString(formatPattern)) {
            switch (column.getType()) {
            case TIMESTAMP:
              simpleDateFormat.applyPattern(Column.DEFAULT_FORMAT_PATTERN_TIMESTAMP);
              break;
            case TIME:
              simpleDateFormat.applyPattern(Column.DEFAULT_FORMAT_PATTERN_TIME);
              break;
            default:
              simpleDateFormat.applyPattern(Column.DEFAULT_FORMAT_PATTERN_DATE);
              break;
            }
          }
          // フォーマットパターンが設定されている場合
          else {
            simpleDateFormat.applyPattern(formatPattern);
          }
          // フォーマット
          basis = simpleDateFormat.format(basisDate);
        }
        m = Integer.toString(basis.length());
      } catch (Exception x) {
        throw new RuntimeException(x);
      }
      break;
    default:
    }
    return m.equals("") ? "" : "maxlength=\"" + m + "\"";
  }

  static public class FormTemplate {

    private WebDataTable<?> table; 
    private String style = "";
    
    public FormTemplate(WebDataTable<?> table) {
      this.table = table; 
    }
    
    public FormTemplate(WebDataTable<?> table, String style) {
      this.table = table; 
      this.style = style;
    }
    
    public String createSimple(int index) throws Exception {
      if (!table.isActive()) table.activate();
      if (table.getRowCount()==0) {
        table.appendRow();
      } else {
        table.moveTo(index);
      }
      Row<?> row = table.getRow();
      StringBuffer buff = new StringBuffer();
      buff.append("<table ");
      buff.append(style);
      buff.append(">");
      buff.append("<tbody>");
      for (Column c : table.getColumnArray()) {
        if (Column.InputType.NONE.equals(c.getDefaultInputEditor())){
          continue;
        }
        buff.append("<tr>");
        buff.append("<th>");
        buff.append(escape(c.getCaption()));
        buff.append("</th>");
        buff.append("<td>");
        String name = table.toRequestParameterName(index, c.getName());
        String value = escape(row.getValueAsString(c.getName()));
        buff.append(toDefaultInputEditor(c, name, value));
        buff.append("</td>");
        buff.append("</tr>");
      }
      buff.append("</tbody>");
      buff.append("</table>");
      return buff.toString();
    }
    
    private String internalMerge(VelocityContext context, String resource) throws Exception {
      StringVelocityTemplate t = new StringVelocityTemplate(resource);
      return t.getMergedString(context);
    }
  }
  
  static public String toDefaultInputEditor(Column column, String name, String value) {
    Column.InputType type = column.getDefaultInputEditor();
    if (type==null) return value;
    String style = column.getDefaultInputStyle();
    if (style==null) style = "";
    String disabled = (column.isReadOnly()) ? " disabled " : "";
    switch(type) {
    case TEXT:
      return "<input " + style + " type=\"text\" name=\"" + name + "\" value=\"" + value + 
             "\" " + toMaxLength(column) + disabled + "/>";
    case PASSWORD:
      return "<input " + style + " type=\"password\" name=\"" + name + "\" value=\"" + value + 
      "\" " + toMaxLength(column) + disabled + "/>";
    case FILE:
      return "<input " + style + " type=\"file\" name=\"" + name + "\" " + toMaxLength(column) + disabled + "/>";
    case RADIO:
      String result = "";
      Code code = null;
      try {
        Class<? extends Code> optionType = column.getOptionType();
        Constructor<?> cns = Class.forName(optionType.getName()).getConstructor();
        code = (Code)cns.newInstance();
        for (String key : code.keySet()) {
          result += "<input " + style + " type=\"radio\" name=\"" + name + "\" value=\"" + key + "\" " + disabled;
          if (key != null && key.equals(value)) {
            result += " checked ";
          }
          result += "/><label>" + code.get(key) + "</label>";
        }
        return result;
      } catch(Exception ex) {
        return "";
      }
    case BUTTON:
      return "<input " + style + " type=\"button\" name=\"" + name + "\" value=\"" + value + 
      "\" " + disabled + "/>";
    case TEXTAREA:
      return "<textarea " + style + " name=\"" + name + "\" " + disabled + ">" + value + "</textarea>";
    case CHECKBOX:
    case LIST:
    case HIDDEN:
    case COMBOBOX:
      return "<span style=\"color:red;\">このエディターは未だ実装していません<span>";
    }
    return "";
  }
}
